/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa;

import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Vector;

/**
 * Class loader which utilizes class repository scoped to a specific {@link Deployment}.
 * Requests for classes/resources are first delegated to parent classloader, i.e. classloader
 * retrieved via {@link #getParentClassLoader()}. If not found, repository scoped to deployment
 * is checked. If still not found, request is delegated to {@link SharedClassLoader}.
 * <p>
 * Note that in addition to dependencies defined by <code>Deployment</code> project, the
 * following constituents which MAY be located within <code>Deployment</code> file are visible
 * to this classloader:
 * <p>
 * <li>An optional classes.jar file located within the optional 'EOA-INF' directory.
 * <li>All jar files located within the optional 'EOA-INF/lib' directory.
 * 
 */
public class DeploymentClassLoader extends AbstractClassLoader {

    private SharedClassLoader sharedClassLoader;

    /**
     * Preferred Constructor.
     * @param urls
     * @param sharedClassLoader
     */
    public DeploymentClassLoader(String id, URL[] urls, SharedClassLoader sharedClassLoader) {
        super(id, urls);
        this.sharedClassLoader = sharedClassLoader;
    }

    /**
     * Alternative constructor which SHOULD NOT be used within a production environment, but
     * MAY be appropriate for testing.
     * @param parent
     */
    public DeploymentClassLoader(String id, ClassLoader parent) {
        super(id, new URL[] {}, parent);
    }

    /**
     * Alternative constructor, e.g. appropriate when java system classloader cannot be
     * configured by <code>System</code> with 'provided dependencies'.
     * 
     * @param parent
     * @param urls
     * @param sharedClassLoader
     */
    public DeploymentClassLoader(String id, URL[] urls, ClassLoader parent, SharedClassLoader sharedClassLoader) {
        super(id, urls, parent);
        this.sharedClassLoader = sharedClassLoader;
    }

    /**
     * Returns shared classloader.
     * 
     * @return shared classloader or <code>null</code> if not defined.
     */
    public SharedClassLoader getSharedClassLoader() {
        return sharedClassLoader;
    }

    /*
     * Used by ApplicationClassLoader to bypass delegation to shared classloader.
     */
    Enumeration<URL> getResourcesInternal(String name) throws IOException {
        return super.getResources(name);
    }

    /*
     * Used by ApplicationClassLoader to bypass delegation to shared classloader.
     */
    Class<?> loadClassInternal(String name) throws ClassNotFoundException {
        return super.loadClass(name);
    }

    /*
     * Used by ApplicationClassLoader to exclude shared urls.
     */
    URL[] getURLsInternal() {
        return super.getURLs();
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.ClassLoader#loadClass(java.lang.String)
     */
    @Override
    public Class<?> loadClass(String name) throws ClassNotFoundException {
        return this.loadClass(name, false);
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.ClassLoader#loadClass(java.lang.String, boolean)
     */
    @Override
    public synchronized Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {
        try {
            return super.loadClass(name, resolve);
        } catch (ClassNotFoundException cfe) {
            if (sharedClassLoader != null) {
                return sharedClassLoader.loadClass(name);
            }
            throw cfe;
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.ClassLoader#getResource(java.lang.String)
     * 
     * NOTE: we synchronize to avoid race condition where two threads attempt to load same
     * class simultaneously (method on super is synchronized, but we need to lock method on
     * super AND on shared loader.) if class not loaded both threads will invoke findClass. it
     * is then a race in terms of which one will load and define class and which one will throw
     * a class linkage error:
     * 
     * "attempted duplicate class definition for name ..."
     */
    @Override
    public synchronized URL getResource(String name) {
        Enumeration<URL> answer = null;
        try {
            answer = getResources(name);
        } catch (IOException ie) {
            throw new RuntimeException(ie);
        }
        if (answer.hasMoreElements()) {
            return answer.nextElement();
        }
        return null;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.ClassLoader#getResources(java.lang.String)
     * 
     * NOTE: we synchronize to avoid race condition where two threads attempt to load same
     * class simultaneously (method on super is synchronized, but we need to lock method on
     * super AND on shared loader.) if class not loaded both threads will invoke findClass. it
     * is then a race in terms of which one will load and define class and which one will throw
     * a class linkage error:
     * 
     * "attempted duplicate class definition for name ..."
     */
    @Override
    public synchronized Enumeration<URL> getResources(String name) throws IOException {

        Enumeration<URL> answer = super.getResources(name);
        if (!answer.hasMoreElements()) {
            if (sharedClassLoader != null) {
                answer = sharedClassLoader.getResources(name);
            }
        }

        if (!answer.hasMoreElements()) {
            return answer;
        }

        // fixes a bug within jar handler that returns a jar entry when searching
        // for meta-inf/services/foo and foo is a directory rather than a file. if
        // a dir entry, the returned stream is unusable. detect error by attempting
        // to close stream. as of jdk 1.5.0_08

        Vector<URL> validURLs = new Vector<URL>();
        while (answer.hasMoreElements()) {
            URL next = answer.nextElement();
            if (next.getProtocol().equals("jar")) {
                try {
                    next.openStream().close();
                    validURLs.add(next);
                } catch (Exception e) {
                }
            } else {
                validURLs.add(next);
            }
        }
        return validURLs.elements();

    }

    /*
     * (non-Javadoc)
     * 
     * @see java.net.URLClassLoader#getURLs()
     */
    @Override
    public URL[] getURLs() {
        if (sharedClassLoader != null) {
            URL[] a = sharedClassLoader.getURLs();
            if (a.length > 0) {
                URL[] b = super.getURLs();
                if (b.length > 0) {
                    URL[] combined = new URL[b.length + a.length];
                    System.arraycopy(b, 0, combined, 0, b.length);
                    System.arraycopy(a, 0, combined, b.length, a.length);
                    return combined;
                } else {
                    return a;
                }
            } else {
                return super.getURLs();
            }
        } else {
            return super.getURLs();
        }
    }

    @Override
    public String toString() {
        return "id: " + id + "@" + Integer.toHexString(hashCode());
    }

}
