/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa;

import java.util.Collections;
import java.util.List;
import java.util.Map;

/**
 * Models dependencies defined by a <code>Deployment</code>, i.e. dependencies upon which
 * <code>Extension</code> instances defined within deployment MAY depend in order to
 * implement their behavior.
 */
public abstract class DeploymentDependencies {

    private List<ProvidedDependency> providedDependencies;
    private List<SharedDependency> sharedDependencies;
    private List<ScopedDependency> scopedDependencies;
    private List<FeatureDependency> featureDependencies;
    private Map<String, OtherDependency> otherDependencies;
    
    public DeploymentDependencies() {
    }

    /**
     * Constructor
     * 
     * @param providedDependencies
     *        or an empty list if none defined
     * @param sharedDependencies
     *        or an empty list if none defined
     * @param scopedDependencies
     *        or an empty list if none defined
     * @param featureDependencies
     *        or an empty list if none defined
     * @param otherDependencies
     *        or an empty map if none defined
     */
    public DeploymentDependencies(
            List<ProvidedDependency> providedDependencies, List<SharedDependency> sharedDependencies,
            List<ScopedDependency> scopedDependencies, List<FeatureDependency> featureDependencies, Map<String, OtherDependency> otherDependencies
            ) {
        if (providedDependencies != null) {
            this.providedDependencies = Collections.unmodifiableList(providedDependencies);
        }
        if (sharedDependencies != null) {
            this.sharedDependencies = Collections.unmodifiableList(sharedDependencies);
        }
        if (scopedDependencies != null) {
            this.scopedDependencies = Collections.unmodifiableList(scopedDependencies);
        }
        if (featureDependencies != null) {
            this.featureDependencies = Collections.unmodifiableList(featureDependencies);
        }
        if (otherDependencies != null) {
            this.otherDependencies = Collections.unmodifiableMap(otherDependencies);
        }
    }

    /**
     * Identifies 'provided' dependencies upon which <code>Extension</code> instances defined
     * within deployment MAY depend in order to implement their behavior.
     * 
     * @return List of 'provided' dependencies
     */
    @SuppressWarnings("unchecked")
    public final List<ProvidedDependency> getProvidedDependencies() {
        if (providedDependencies == null) {
            return Collections.EMPTY_LIST;
        }
        return providedDependencies;
    }

    /**
     * Identifies 'shared' dependencies upon which <code>Extension</code> instances defined
     * within deployment MAY depend in order to implement their behavior.
     * 
     * @return List of 'shared' dependencies or an empty list if none defined
     * 
     * @see SystemDependencies#getSharedDependencies()
     */
    @SuppressWarnings("unchecked")
    public final List<SharedDependency> getSharedDependencies() {
        if (sharedDependencies == null) {
            return Collections.EMPTY_LIST;
        }
        return sharedDependencies;
    }

    /**
     * Identifies 'scoped' dependencies upon which <code>Extension</code> instances defined
     * within deployment MAY depend in order to implement their behavior.
     * 
     * @return List of scoped dependencies or an empty list if none defined
     */
    @SuppressWarnings("unchecked")
    public final List<ScopedDependency> getScopedDependencies() {
        if (scopedDependencies == null) {
            return Collections.EMPTY_LIST;
        }
        return scopedDependencies;
    }

    /**
     * Identifies <code>FeatureFactory</code> instances upon which <code>Extension</code>
     * instances defined within deployment MAY depend in order to implement their behavior.
     * 
     * @return List<FeatureDependency> or an empty list if no dependencies defined
     * 
     * @see SystemDependencies#getFactoryDependencies()
     */
    @SuppressWarnings("unchecked")
    public final List<FeatureDependency> getFeatureDependencies() {
        if (featureDependencies == null) {
            return Collections.EMPTY_LIST;
        }
        return featureDependencies;
    }
    
    /**
     * Identifies 'other' dependencies defined by deployment each of which is mapped to a
     * versionless ref URI. Note that this implies that more than one version of the same
     * 'other' dependency MUST NOT be defined.
     * 
     * @return Map of 'other' dependencies
     */
    @SuppressWarnings("unchecked")
    public final Map<String, OtherDependency> getOtherDependencies() {
        if (otherDependencies == null) {
            return Collections.EMPTY_MAP;
        }
        return otherDependencies;
    }


}
