/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import javax.xml.namespace.QName;
import javax.xml.transform.dom.DOMSource;

import org.bluestemsoftware.specification.eoa.Deployment;
import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ClasspathDeployment.BuiltInComponentsDeployment;
import org.bluestemsoftware.specification.eoa.component.ComponentReference.SchemaComponentReference;
import org.bluestemsoftware.specification.eoa.component.ComponentReference.SchemaPropertyReference;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.RootComponent.RuntimeProvidable;
import org.bluestemsoftware.specification.eoa.component.application.Application;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationRT;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.binding.rt.BindingRT;
import org.bluestemsoftware.specification.eoa.component.engine.Engine;
import org.bluestemsoftware.specification.eoa.component.intrface.Interface;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;
import org.bluestemsoftware.specification.eoa.component.message.rt.Message;
import org.bluestemsoftware.specification.eoa.component.policy.Policy;
import org.bluestemsoftware.specification.eoa.component.rt.ComponentRT;
import org.bluestemsoftware.specification.eoa.component.rt.ProviderReader;
import org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT;
import org.bluestemsoftware.specification.eoa.component.rt.RootComponentRT.SingletonRT;
import org.bluestemsoftware.specification.eoa.component.types.Schema;
import org.bluestemsoftware.specification.eoa.component.types.SchemaComponent.Type;
import org.bluestemsoftware.specification.eoa.component.types.rt.SchemaDocument;
import org.bluestemsoftware.specification.eoa.ext.ExecutableException;
import org.bluestemsoftware.specification.eoa.ext.ExecutableExtension;
import org.bluestemsoftware.specification.eoa.ext.ExtensionException;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactory;
import org.bluestemsoftware.specification.eoa.ext.ManageableExtension;
import org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLDocument;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLException;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLFactory;
import org.bluestemsoftware.specification.eoa.ext.message.MessageFactory;
import org.bluestemsoftware.specification.eoa.ext.schema.SchemaFactory;
import org.bluestemsoftware.specification.eoa.system.ManagementContext;
import org.bluestemsoftware.specification.eoa.system.SystemContext;
import org.bluestemsoftware.specification.eoa.system.System.Log;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

/**
 * Models a component deployment archive which contains one or more {@link Component} and/or
 * one or more {@link ComponentRT} definitions.
 */
public class ComponentDeployment extends Deployment {

    public static final String LINE_BREAK = System.getProperty("line.separator");

    protected ComponentManager componentManager;
    protected File descriptor;
    protected Map<FragmentIdentifier, RootComponent> components = new HashMap<FragmentIdentifier, RootComponent>();
    protected Map<ComponentType, Map<FragmentIdentifier, RootComponent>> componentsByType = new HashMap<ComponentType, Map<FragmentIdentifier, RootComponent>>();
    protected Map<ComponentType, Map<ComponentName, RootComponentRT>> providers = new HashMap<ComponentType, Map<ComponentName, RootComponentRT>>();

    protected static final ReentrantReadWriteLock readWriteLock = new ReentrantReadWriteLock();
    protected static final Lock readLock = readWriteLock.readLock();
    protected static final Lock writeLock = readWriteLock.writeLock();

    private String ref;
    private boolean isTransitive;
    private Boolean componentsDeployed = Boolean.FALSE;
    private Boolean providersDeployed = Boolean.FALSE;
    private Boolean operationStylesModeled = Boolean.FALSE;
    private boolean isError = false;

    /**
     * Constructor.
     * 
     * @param organizationID
     * @param artifactID
     * @param version
     * @param extension
     * @param file
     * @param deploymentDependencies
     * @param modules
     *        each of which is mapped to module extension ref, or an empty map if none defined
     */
    public ComponentDeployment(String organizationID, String artifactID, String version, String extension,
            File file, ComponentDependencies deploymentDependencies, boolean isTransitive) {
        super(organizationID, artifactID, version, extension, file, deploymentDependencies);
        this.isTransitive = isTransitive;
    }

    /**
     * Returns ref URI which serves as a unique identifier for this deployment. The value MUST
     * be unique within contet of <code>System</code> and SHOULD be globally unique. The
     * format of ref URI is:
     * <p>
     * organizationID/artifactID/version/extension
     * <p>
     * Note that ref uri for a component deployment is fully qualified ref, i.e. it includes
     * version (more than one version of same component deployment can be deployed within
     * context of system).
     */
    public String getRef() {
        if (ref == null) {
            ref = organizationID + "/" + artifactID + "/" + version + "/" + extension;
        }
        return ref;
    }

    /**
     * Gets descriptor file which MUST be located within deployment at the following path:
     * META-INF/eoa/deployment.xml
     * 
     * @return File
     */
    public File getDescriptor() {
        return descriptor;
    }

    /**
     * Indicates if <code>ComponentDeployment</code> is associated with a transitive
     * <code>ComponentDependency</code>, i.e. transitive from the perspective of
     * <code>System</code>.
     * 
     * @return true if deployment is associated with a transitive dependency, false if not. If
     *         true, <code>Engine</code> provider definitions, <code>Application</code>
     *         provider definitions and/or any 'orphaned' <code>Message</code> provider
     *         definitions (a message definition that is defined within a deployment other than
     *         deployment within which corresponding {@link InterfaceMessage} definition is
     *         defined), which MAY be defined within deployment MUST NOT be loaded.
     */
    public boolean isTransitive() {
        return isTransitive;
    }

    /**
     * Suspends all instances of {@link ExecutableExtension} which MAY exist within this
     * deployment.
     */
    public synchronized void suspendAll() throws ExecutableException {
        for (Map<ComponentName, RootComponentRT> typedProviders : providers.values()) {
            for (RootComponentRT runtimeProvider : typedProviders.values()) {
                if (runtimeProvider instanceof ExecutableExtension) {
                    ((ExecutableExtension)runtimeProvider).suspend();
                }
            }
        }
    }

    /**
     * Resumes all instances of {@link ExecutableExtension} which MAY exist within this
     * deployment.
     */
    public synchronized void resumeAll() throws ExecutableException {
        for (Map<ComponentName, RootComponentRT> typedProviders : providers.values()) {
            for (RootComponentRT runtimeProvider : typedProviders.values()) {
                if (runtimeProvider instanceof ExecutableExtension) {
                    ((ExecutableExtension)runtimeProvider).resume();
                }
            }
        }
    }

    /*
     * component manager overwrites file after unzipping file to temporary location
     */
    void setFile(File file) {
        this.file = file;
    }

    /*
     * set by component manager when deployment inserted
     */
    void setComponentManager(ComponentManager componentManager) {
        if (componentManager == null) {
            throw new IllegalArgumentException("componentManager null");
        }
        this.componentManager = componentManager;
    }

    ComponentContext getComponentContext() {
        return (ComponentContext)deploymentContext;
    }

    /*
     * (non-Javadoc)
     * 
     * Invoked by <code>ComponentManager</code> during load phase. Loads, i.e. reads
     * definitions.
     */
    void load(ComponentContext deploymentContext, Log log) throws DeploymentException {

        if (deploymentContext == null) {
            throw new IllegalArgumentException("deploymentContext == null");
        } else {
            this.deploymentContext = deploymentContext;
        }
        writeLock.lock();
        try {
            Document descriptorDocument = null;
            FileInputStream fis = null;
            try {
                descriptor = new File(file, DESCRIPTOR_PATH);
                if (!descriptor.exists()) {
                    throw new DeploymentException("Error loading deployment "
                            + getRef()
                            + ". Deployment missing required descriptor '"
                            + DESCRIPTOR_PATH
                            + "'.");
                }
                fis = new FileInputStream(descriptor);
                InputSource inputSource = new InputSource(fis);
                inputSource.setSystemId(descriptor.getAbsolutePath());
                descriptorDocument = new DescriptorParser().parse(inputSource);
            } finally {
                if (fis != null) {
                    try {
                        fis.close();
                    } catch (IOException ie) {
                        throw new DeploymentException("Error deploying component deployment "
                                + getRef()
                                + ". "
                                + ie);
                    }
                }
            }
            parseDescriptorElement(descriptor.getAbsolutePath(), descriptorDocument.getDocumentElement(), log);
        } catch (Throwable th) {
            if (!(th instanceof DeploymentException)) {
                StringBuffer sb = new StringBuffer();
                if (th instanceof DefinitionMismatchException) {
                    sb.append(th.getMessage());
                } else {
                    try {
                        ByteArrayOutputStream baos = new ByteArrayOutputStream();
                        th.printStackTrace(new PrintStream(baos, true, "UTF-8"));
                        sb.append(LINE_BREAK);
                        sb.append(new String(baos.toByteArray(), "UTF-8"));
                    } catch (Exception exception) {
                        sb.append(th.toString());
                    }
                }
                throw new DeploymentException(sb.toString());
            } else {
                throw (DeploymentException)th;
            }
        } finally {
            writeLock.unlock();
        }
    }

    /**
     * Indicates whether or not deployment failed to deploy.
     * @return
     */
    public boolean isError() {
        readLock.lock();
        try {
            return isError;
        } finally {
            readLock.unlock();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * Used by component manager to determine if static components are deployed.
     */
    Boolean componentsDeployed() {
        readLock.lock();
        try {
            return componentsDeployed;
        } finally {
            readLock.unlock();
        }
    }

    /**
     * Adds a component whose definition was parsed from descriptor by an instance of
     * <code>ComponentReader</code>.
     * 
     * @param component
     * @throws DefinitionMismatchException
     *         If component <code>null</code> or if component with matching fragment
     *         identifier already loaded and components are not equivalent.
     */
    public void addComponent(RootComponent component) {
        if (component == null) {
            throw new IllegalArgumentException("component null");
        }
        writeLock.lock();
        try {
            if (components.get(component.getFragmentIdentifier()) != null) {
                Component existing = components.get(component.getFragmentIdentifier());
                if (existing.equals(component)) {
                    return;
                }
                throw new DefinitionMismatchException("Definition for component "
                        + component.getFragmentIdentifier()
                        + " differs from a definition already added to deployment."
                        + LINE_BREAK
                        + LINE_BREAK
                        + "Existing:"
                        + LINE_BREAK
                        + existing.toString()
                        + LINE_BREAK
                        + LINE_BREAK
                        + "Proposed:"
                        + LINE_BREAK
                        + component.toString());
            }
            components.put(component.getFragmentIdentifier(), component);
            Map<FragmentIdentifier, RootComponent> typed = componentsByType.get(component.getComponentType());
            if (typed == null) {
                typed = new HashMap<FragmentIdentifier, RootComponent>();
                componentsByType.put(component.getComponentType(), typed);
            }
            typed.put(component.getFragmentIdentifier(), component);
        } finally {
            writeLock.unlock();
        }
    }

    /**
     * Adds a runtime provider definition whose definition was parsed from descriptor by an
     * instance of <code>ComponentReader</code>, i.e. provider definition was inlined within
     * IDL document, e.g. a schema document or a policy expression.
     * 
     * @param provider
     * @throws DeploymentException
     *         if corresponding component, i.e. instanceof {@link RuntimeProvidable}, not
     *         defined within this deployment.
     */
    public void addProvider(RootComponentRT runtimeProvider) throws DeploymentException {
        if (runtimeProvider == null) {
            throw new IllegalArgumentException("runtimeProvider null");
        }
        writeLock.lock();
        try {

            ComponentType componentType = runtimeProvider.getComponentType();
            ComponentName componentName = runtimeProvider.getName();

            Map<ComponentName, RootComponentRT> typedProviders = providers.get(componentType);
            if (typedProviders == null) {
                typedProviders = new HashMap<ComponentName, RootComponentRT>();
                providers.put(componentType, typedProviders);
            }
            if (typedProviders.get(componentName) != null) {
                RootComponentRT existing = typedProviders.get(componentName);
                if (existing.equals(runtimeProvider)) {
                    return;
                }
                throw new IllegalArgumentException("Definition for runtime component "
                        + runtimeProvider.getFragmentIdentifier()
                        + " differs from a definition already added to deployment."
                        + LINE_BREAK
                        + LINE_BREAK
                        + "Existing:"
                        + LINE_BREAK
                        + existing.toString()
                        + LINE_BREAK
                        + "Proposed:"
                        + LINE_BREAK
                        + runtimeProvider.toString());
            } else {
                typedProviders.put(componentName, runtimeProvider);
            }

            // we should only be adding schema documents and policy expressions via
            // this method, i.e. message providers are not read by component readers.
            // and ... component reader must add component before adding provider

            if (!(runtimeProvider instanceof SingletonRT)) {
                throw new DeploymentException("Expected instanceof " + SingletonRT.class.getName());
            }

            RootComponent rootComponent = getComponent(componentType, componentName);

            if (rootComponent == null) {
                throw new DeploymentException("Component definition for componentType "
                        + componentType
                        + " and componentName '"
                        + componentName
                        + "' is undefined.");
            }

            if (!(rootComponent instanceof RuntimeProvidable)) {
                throw new DeploymentException("Provider definition defined for "
                        + ((RootComponent)rootComponent).getFragmentIdentifier()
                        + ", but RootComponent not RuntimeProvidable");
            }

            ((RootComponent)rootComponent).setRuntimeProvider(runtimeProvider);

        } finally {
            writeLock.unlock();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * Invoked indirectly by instances of ComponentReference via ComponentContext.
     */
    <R extends RootComponent> R findReferencedComponent(ComponentReference<R> componentReference) throws DeploymentException {

        readLock.lock();
        try {

            ComponentName cn = componentReference.getReferencedComponentName();
            ComponentType ct = componentReference.getReferencedComponentType();

            FragmentIdentifier referencingID = componentReference.getFragmentIdentifier();

            // if component definitions have been deployed, then we are establishing
            // a reference made by a runtime component. we must deploy the dependent
            // provider definition before returning the runtime providable, i.e. the
            // referencing component assumes runtime provider definition is complete.
            // note that it is responsibility of provider implementation to either
            // allow or disallow circular references

            if (componentsDeployed == Boolean.TRUE) {
                FragmentIdentifier fid = FragmentIdentifier.valueOf(ct, cn);
                RootComponent rc = SystemContext.getContext().getSystem().getComponent(fid);
                if (rc == null) {
                    return null; // the component reference is invalid. return null
                }
                CurrentReference.setCurrentReference(referencingID.toString());
                if (rc.runtimeProvider != null) {
                    rc.runtimeProvider.deploy(rc.componentContext);
                } else {
                    // anonymous apps and engines have no runtime provider
                }
                return componentReference.getReferencedComponentClass().cast(rc);
            }

            // component definitions have not yet been deployed, i.e. we are establishing
            // a reference made by a component. check this deployment first to avoid circular
            // reference error thrown by component manager. if not found, then we check
            // system which lazy deploys deployment within which referenced component is
            // defined. if found within deployment, we force component to deploy itself if
            // not already deployed, i.e. so it's definition will be complete, i.e. the
            // referencing component assumes referenced component has been deployed.

            if (componentReference instanceof SchemaComponentReference) {

                String scn = ((SchemaComponentReference)componentReference).getSchemaComponentName();
                Type sct = ((SchemaComponentReference)componentReference).getSchemaComponentType();

                FragmentIdentifier fid = findSchema(cn.getNamespaceURI(), sct, scn);
                RootComponent referencedComponent = (Schema)components.get(fid);

                if (referencedComponent == null) {
                    CurrentReference.setCurrentReference(getRef() + "@" + referencingID.toString());
                    return componentManager.findReferencedComponent(componentReference);
                } else {
                    CurrentReference.setCurrentReference(referencingID.toString());
                    referencedComponent.deploy((ComponentContext)deploymentContext);
                    return componentReference.getReferencedComponentClass().cast(referencedComponent);
                }

            }

            if (componentReference instanceof SchemaPropertyReference) {

                String pn = ((SchemaPropertyReference)componentReference).getSchemaPropertyName();

                FragmentIdentifier fid = findSchema(cn.getNamespaceURI(), pn);
                RootComponent rc = (Schema)components.get(fid);

                if (rc == null) {
                    CurrentReference.setCurrentReference(getRef() + "@" + referencingID.toString());
                    return componentManager.findReferencedComponent(componentReference);
                } else {
                    CurrentReference.setCurrentReference(referencingID.toString());
                    rc.deploy((ComponentContext)deploymentContext);
                    return componentReference.getReferencedComponentClass().cast(rc);
                }

            }

            RootComponent rc = getComponent(ct, cn);

            if (rc == null) {
                CurrentReference.setCurrentReference(getRef() + "@" + referencingID.toString());
                return componentManager.findReferencedComponent(componentReference);
            } else {
                CurrentReference.setCurrentReference(referencingID.toString());
                rc.deploy((ComponentContext)deploymentContext);
                return componentReference.getReferencedComponentClass().cast(rc);
            }

        } finally {
            readLock.unlock();
        }

    }

    /**
     * Returns a component definition contained within this deployment.
     * 
     * @param fragmentIdentifier
     * @return component or <code>null</code> if not found
     */
    public RootComponent getComponent(FragmentIdentifier fragmentIdentifier) {
        readLock.lock();
        try {
            return components.get(fragmentIdentifier);
        } finally {
            readLock.unlock();
        }
    }

    /**
     * Returns a component definition contained within this deployment.
     * 
     * @param componentType
     * @param componentName
     * @return component or <code>null</code> if not found
     */
    public RootComponent getComponent(ComponentType componentType, QName componentName) {
        return getComponent(FragmentIdentifier.valueOf(componentType, componentName));
    }

    /**
     * Returns a runtime component definition contained within this deployment.
     * 
     * @param componentType
     * @param componentName
     * @return component or <code>null</code> if not found
     */
    public RootComponentRT getProvider(ComponentType componentType, QName componentName) {
        Map<ComponentName, RootComponentRT> typedProviders = providers.get(componentType);
        if (typedProviders != null) {
            return typedProviders.get(componentName);
        } else {
            return null;
        }
    }

    /*
     * Locates a schema contained within this deployment which defines indicated property.
     */
    FragmentIdentifier findSchema(String targetNamespace, String propertyName) {
        readLock.lock();
        try {
            Map<FragmentIdentifier, RootComponent> schemas = componentsByType.get(ComponentType.SCHEMA);
            if (schemas == null) {
                return null;
            }
            for (RootComponent temp : schemas.values()) {
                if (((Schema)temp).getName().getNamespaceURI().equals(targetNamespace)) {
                    if (((Schema)temp).getProperty(propertyName) != null) {
                        return temp.getFragmentIdentifier();
                    }
                }
            }
            return null;
        } finally {
            readLock.unlock();
        }
    }

    /*
     * Locates a schema contained within this deployment which defines indicated schema
     * component.
     */
    FragmentIdentifier findSchema(String targetNamespace, Type schemaComponentType, String schemaComponentName) {
        readLock.lock();
        try {
            Map<FragmentIdentifier, RootComponent> schemas = componentsByType.get(ComponentType.SCHEMA);
            if (schemas == null) {
                return null;
            }
            for (RootComponent temp : schemas.values()) {
                if (((Schema)temp).getName().getNamespaceURI().equals(targetNamespace)) {
                    if (((Schema)temp).getSchemaComponent(schemaComponentType, schemaComponentName) != null) {
                        return temp.getFragmentIdentifier();
                    }
                }
            }
            return null;
        } finally {
            readLock.unlock();
        }
    }

    /*
     * Invoked by component manager when searching for a referenced component, i.e. a reference
     * established by a component defined within a separate deployment.
     */
    <R extends RootComponent> R getComponent(ComponentReference<R> componentReference) {
        ComponentName cname = componentReference.getReferencedComponentName();
        ComponentType ctype = componentReference.getReferencedComponentType();
        FragmentIdentifier fid = FragmentIdentifier.valueOf(ctype, cname);
        return componentReference.getReferencedComponentClass().cast(getComponent(fid));
    }

    /**
     * Returns all component definitions contained within this deployment.
     * 
     * @return collection of components or an empty collection if none defined
     */
    public synchronized Collection<RootComponent> getComponents() {
        readLock.lock();
        try {
            return Collections.unmodifiableCollection(components.values());
        } finally {
            readLock.unlock();
        }
    }

    /**
     * Invoked by <code>ComponentManager</code> once load phase is complete. Signals that
     * contained components may now establish references to other components and validate
     * definition.
     * 
     * @throws DeploymentException
     */
    void deployComponents(Log log) {
        writeLock.lock();
        try {
            if (isError || componentsDeployed) {
                return;
            }
            if (componentsDeployed == null) {
                log.error(LINE_BREAK
                        + "Deployment "
                        + getRef()
                        + " is circularly referenced by "
                        + CurrentReference.getCurrentReference());
                isError = true;
                return;
            } else {
                componentsDeployed = null;
            }
            for (RootComponent component : components.values()) {
                try {
                    component.deploy((ComponentContext)deploymentContext);
                } catch (Throwable th) {
                    String errorMessage = null;
                    if (th instanceof DeploymentException) {
                        errorMessage = th.getMessage();
                    } else {
                        StringBuilder sb = new StringBuilder();
                        try {
                            ByteArrayOutputStream baos = new ByteArrayOutputStream();
                            th.printStackTrace(new PrintStream(baos, true, "UTF-8"));
                            sb.append(LINE_BREAK);
                            sb.append(new String(baos.toByteArray(), "UTF-8"));
                        } catch (Exception exception) {
                            sb.append(th.toString());
                        }
                        errorMessage = sb.toString();
                    }
                    log.error(LINE_BREAK
                            + "Error deploying deployment "
                            + getRef()
                            + ". "
                            + LINE_BREAK
                            + errorMessage);
                    isError = true;
                    return;
                }
            }
            componentsDeployed = Boolean.TRUE;
        } finally {
            writeLock.unlock();
        }
    }

    /**
     * Invoked by <code>ComponentManager</code>. Signals that component deployment phase is
     * complete and that runtime provider instances may now be deployed. Note that this is
     * performed separately because circular references made across deployments by static
     * components are disallowed. This restriction does not apply to runtime components.
     */
    void deployProviders(Log log, Set<ManagementContext> managementContexts) {
        writeLock.lock();
        try {
            if (providersDeployed || isError) {
                return;
            }
            // allow runtime components to establish references and perform validation
            for (Map<ComponentName, RootComponentRT> typedProviders : providers.values()) {
                for (RootComponentRT runtimeProvider : typedProviders.values()) {
                    try {
                        runtimeProvider.deploy((ComponentContext)deploymentContext);
                    } catch (Throwable th) {
                        StringBuilder sb = new StringBuilder();
                        if (th instanceof DeploymentException) {
                            sb.append(th.getMessage());
                            while (th.getCause() != null) {
                                th = th.getCause();
                                sb.append(LINE_BREAK);
                                sb.append("Caused by: ");
                                sb.append(th.getMessage());
                            }
                        } else {
                            try {
                                ByteArrayOutputStream baos = new ByteArrayOutputStream();
                                th.printStackTrace(new PrintStream(baos, true, "UTF-8"));
                                sb.append(LINE_BREAK);
                                sb.append(new String(baos.toByteArray(), "UTF-8"));
                            } catch (Exception exception) {
                                sb.append(th.toString());
                            }
                        }
                        log.error(LINE_BREAK
                                + "Error deploying deployment "
                                + getRef()
                                + ". "
                                + LINE_BREAK
                                + sb.toString());
                        isError = true;
                        return;
                    }
                }
            }
            // now we initialize runtime components that implement manageable extension
            for (Map<ComponentName, RootComponentRT> typedProviders : providers.values()) {
                for (RootComponentRT runtimeProvider : typedProviders.values()) {
                    ManageableExtension me = null;
                    if (runtimeProvider instanceof ManageableExtension) {
                        me = (ManageableExtension)runtimeProvider;
                    }
                    if (me != null) {
                        try {
                            me.init(managementContexts);
                        } catch (Throwable th) {
                            String errorMessage = null;
                            if (th instanceof ExtensionException) {
                                errorMessage = th.toString();
                            } else {
                                StringBuilder sb = new StringBuilder();
                                try {
                                    ByteArrayOutputStream baos = new ByteArrayOutputStream();
                                    th.printStackTrace(new PrintStream(baos, true, "UTF-8"));
                                    sb.append(LINE_BREAK);
                                    sb.append(new String(baos.toByteArray(), "UTF-8"));
                                } catch (Exception exception) {
                                    sb.append(th.toString());
                                }
                                errorMessage = sb.toString();
                            }
                            log.error(LINE_BREAK
                                    + "Error deploying deployment "
                                    + getRef()
                                    + ". "
                                    + LINE_BREAK
                                    + errorMessage);
                            isError = true;
                            return;
                        }

                    }
                }
            }
            providersDeployed = Boolean.TRUE;
        } finally {
            writeLock.unlock();
        }
    }

    /*
     * see comments on RootComponent#modelOperationStyles()
     */
    void modelOperationStyles(Log log) {

        writeLock.lock();
        try {
            if (operationStylesModeled || isError) {
                return;
            }
            for (RootComponent component : components.values()) {
                if (component instanceof Interface) {
                    try {
                        component.modelOperationStyles();
                    } catch (Throwable th) {
                        String errorMessage = null;
                        if (th instanceof DeploymentException) {
                            errorMessage = th.getMessage();
                        } else {
                            StringBuilder sb = new StringBuilder();
                            try {
                                ByteArrayOutputStream baos = new ByteArrayOutputStream();
                                th.printStackTrace(new PrintStream(baos, true, "UTF-8"));
                                sb.append(LINE_BREAK);
                                sb.append(new String(baos.toByteArray(), "UTF-8"));
                            } catch (Exception exception) {
                                sb.append(th.toString());
                            }
                            errorMessage = sb.toString();
                        }
                        log.error(LINE_BREAK
                                + "Error deploying deployment "
                                + getRef()
                                + ". "
                                + LINE_BREAK
                                + errorMessage);
                        isError = true;
                        return;
                    }
                }
            }
            operationStylesModeled = Boolean.TRUE;
        } finally {
            writeLock.unlock();
        }

    }

    /**
     * Invoked by <code>ComponentManager</code> when underlying deployment file is deleted.
     * Makes deployment, a candidate for garbage collection, once its components are no longer
     * referenced by components within other currently loaded deployments.
     * 
     * @throws DeploymentException
     *         if deployment contains an instance of {@link ExecutableExtension} that is not in
     *         a suspended state.
     */
    void unload() throws DeploymentException {

        writeLock.lock();
        try {

            for (Map<ComponentName, RootComponentRT> typedProviders : providers.values()) {
                for (RootComponentRT runtimeProvider : typedProviders.values()) {
                    if (runtimeProvider instanceof ManageableExtension) {
                        if (runtimeProvider instanceof ExecutableExtension) {
                            if (!((ExecutableExtension)runtimeProvider).isSuspended()) {
                                throw new DeploymentException("Error unloading component deployment "
                                        + getRef()
                                        + ". Runtime provider "
                                        + runtimeProvider.getFragmentIdentifier()
                                        + " MUST be suspended before corresponding RootComponent definition"
                                        + " can be unloaded.");
                            }
                        }
                        try {
                            ((ManageableExtension)runtimeProvider).destroy();
                        } catch (Throwable th) {
                            throw new DeploymentException("Caught throwable thrown by manageable extension "
                                    + runtimeProvider.getName()
                                    + " when attempting to destroy extension. "
                                    + th);
                        }
                    }
                }
            }

            // if component definitions are no longer referenced, coax jvm
            // into reclaiming them

            System.gc();

        } finally {
            writeLock.unlock();
        }

    }

    protected void parseDescriptorElement(String descriptorPath, Element descriptorElement, Log log) throws DeploymentException {

        // parse component definitions. component reader will add component and inlined
        // provider definitions to deployment via add methods

        Element componentsElement = getChildElement(descriptorElement, new QName(DESCRIPTOR_NS, "components"));
        if (componentsElement != null) {
            List<Element> definitions = getChildElements(componentsElement, new QName("*", "*"));
            for (Element definition : definitions) {
                String ns = definition.getNamespaceURI();
                IDLFactory idlFactory = SystemContext.getContext().getSystem().getIDLFactory(ns);
                if (idlFactory == null) {
                    throw new DeploymentException("IDLFactory '" + ns + "' is undefined.");
                }
                DOMSource source = new DOMSource(definition, descriptorPath);
                IDLDocument idlDocument;
                try {
                    idlDocument = idlFactory.readIDL((ComponentContext)deploymentContext, source);
                } catch (IDLException we) {
                    throw new DeploymentException("Error reading component definitions. " + we.getMessage());
                }
                ((ComponentReader)idlFactory).readComponents((ComponentContext)deploymentContext, idlDocument);
            }
        }

        // parse provider definitions

        Element providersElement = getChildElement(descriptorElement, new QName(DESCRIPTOR_NS, "providers"));

        if (providersElement != null) {

            List<Element> definitions = getChildElements(providersElement, new QName("*", "*"));

            for (Element definition : definitions) {

                // by convention, so that we can differentiate runtime component types
                // prior to locating reader (if deployment is transtive reader may
                // be undefined) we require that local name of provider element match
                // lower cased name of component type

                String localName = definition.getLocalName().toLowerCase();

                ComponentType componentType = null;

                if (localName.equals("message")) {
                    componentType = ComponentType.MESSAGE;
                } else if (localName.equals("schema")) {
                    componentType = ComponentType.SCHEMA;
                } else if (localName.equals("binding")) {
                    componentType = ComponentType.BINDING;
                } else if (localName.equals("application")) {
                    componentType = ComponentType.APPLICATION;
                } else if (localName.equals("policy")) {
                    componentType = ComponentType.POLICY;
                } else if (localName.equals("engine")) {
                    componentType = ComponentType.ENGINE;
                }

                if (componentType == null) {
                    throw new DeploymentException("Provider definition for "
                            + new QName(definition.getNamespaceURI(), definition.getLocalName())
                            + " is invalid. By convention, local name of element must be one of"
                            + " 'message', 'schema', 'binding', 'application',"
                            + " 'policy', or 'engine'");
                }

                // if this is a transitive deployment, we do not load application,
                // engine or orphaned message definitions

                if (isTransitive) {
                    if (componentType == ComponentType.ENGINE) {
                        if (log.isDebugEnabled()) {
                            log.debug("not loading runtime engine definition for transitive deployment "
                                    + getRef());
                        }
                        continue;
                    }
                    if (componentType == ComponentType.APPLICATION) {
                        if (log.isDebugEnabled()) {
                            log.debug("not loading runtime application definition for transitive deployment "
                                    + getRef());
                        }
                        continue;
                    }
                    if (componentType == ComponentType.MESSAGE) {
                        String prefixedName = definition.getAttribute("name");
                        if (prefixedName == null) {
                            throw new DeploymentException("Provider definition for "
                                    + new QName(definition.getNamespaceURI(), definition.getLocalName())
                                    + " is invalid. By convention, element"
                                    + " must define 'name' attribute which identifes name of message.");
                        }
                        QName name = ComponentName.valueOf(definition, prefixedName);
                        if (getComponent(componentType, name) == null) {
                            if (log.isDebugEnabled()) {
                                log.debug("not loading runtime message definition for transitive deployment "
                                        + getRef());
                            }
                            continue;
                        }
                    }
                }

                String pns = definition.getNamespaceURI();
                ProviderReader<?> providerReader = SystemContext.getContext().getSystem().getProviderReader(pns);
                if (providerReader == null) {
                    throw new DeploymentException("ProviderReader for namespace '" + pns + "' is undefined.");
                }

                DOMSource source = new DOMSource(definition, descriptorPath);
                Set<?> defs = providerReader.readProviders((ComponentContext)deploymentContext, source);

                for (Object object : defs) {

                    RootComponentRT provider = (RootComponentRT)object;

                    // because app provider does not have to extend from any core abstract
                    // class, we must perform some sanity checks here

                    if (provider instanceof ApplicationRT) {
                        if (provider.getName() == null) {
                            throw new DeploymentException("Application provider returned null componentName");
                        }
                        if (provider.getComponentType() == null) {
                            throw new DeploymentException("Application provider returned null componentType");
                        }
                        if (provider.getFragmentIdentifier() == null) {
                            throw new DeploymentException("Application provider returned null fragmentIdentifier");
                        }
                        if (((ApplicationRT)provider).getExtensionType() == null) {
                            throw new DeploymentException("Application provider returned null extension type");
                        }
                        if (((ApplicationRT)provider).getApplicationReferences() == null) {
                            throw new DeploymentException("Application provider returned null application refs");
                        }
                        if (((ApplicationRT)provider).getExtensionFactory() == null) {
                            throw new DeploymentException("Application provider returned null factory.");
                        }
                    }

                    ComponentName componentName = provider.getName();

                    if (provider.getComponentType() != componentType) {
                        throw new DeploymentException("Runtime definition for built-in component "
                                + componentName
                                + ", specifies componentType "
                                + provider.getComponentType()
                                + ". Expected "
                                + componentType);
                    }

                    if (componentName.getNamespaceURI().equals(RootComponent.EOA_COMPONENT_NS)) {
                        if (!(this instanceof BuiltInComponentsDeployment)) {
                            throw new DeploymentException("Runtime definition for built-in component "
                                    + componentName
                                    + ", cannot be overridden.");
                        }
                    }

                    Map<ComponentName, RootComponentRT> typedProviders = providers.get(componentType);
                    if (typedProviders == null) {
                        typedProviders = new HashMap<ComponentName, RootComponentRT>();
                        providers.put(componentType, typedProviders);
                    }

                    if (typedProviders.get(componentName) != null) {
                        throw new DeploymentException("Duplicate definition for Runtime provider "
                                + provider.getFragmentIdentifier()
                                + " defined within deployment "
                                + getRef()
                                + ".");
                    } else {

                        // if we're adding an application or engine provider, the
                        // spec limits one definition per deployment

                        if (componentType == ComponentType.APPLICATION) {
                            if (typedProviders.size() != 0) {
                                throw new DeploymentException("More than one runtime application provider"
                                        + " defined within deployment "
                                        + getRef()
                                        + ".");
                            }
                        }

                        if (componentType == ComponentType.ENGINE) {
                            if (typedProviders.size() != 0) {
                                throw new DeploymentException("More than one runtime engine provider"
                                        + " defined within deployment "
                                        + getRef()
                                        + ".");
                            }
                        }

                        typedProviders.put(componentName, provider);
                    }

                    RootComponent rootComponent = getComponent(componentType, componentName);

                    if (provider instanceof SingletonRT) {

                        // if provider is a singleton, the corresponding root component
                        // definition must be defined within same deployment as provider.
                        // we attempt the match up here. note that if provider is a schema
                        // provider, the component def is implied. this is a special case,
                        // i.e. all other root components must be defined via wsdl

                        if (rootComponent == null) {
                            if (provider instanceof SchemaDocument) {
                                ExtensionFactory xf = ((SchemaDocument)provider).getExtensionFactory();
                                rootComponent = ((SchemaFactory)xf).buildSchema(this, (SchemaDocument)provider);
                                addComponent(rootComponent);
                            } else {
                                throw new DeploymentException("Error loading runtime provider definition "
                                        + provider.getFragmentIdentifier()
                                        + ". Definition for corresponding RootComponent "
                                        + componentName
                                        + " is undefined.");
                            }
                        }

                        if (!(rootComponent instanceof RuntimeProvidable)) {
                            throw new DeploymentException("Provider definition defined for "
                                    + ((RootComponent)rootComponent).getFragmentIdentifier()
                                    + ", but RootComponent not RuntimeProvidable");
                        }

                        ((RootComponent)rootComponent).setRuntimeProvider(provider);

                    } else {

                        // it's a message provider. provider definition does not have to be in
                        // same deployment. but if it is, go ahead and do the match up

                        if (rootComponent != null) {
                            ((RootComponent)rootComponent).setRuntimeProvider(provider);
                        }

                    }

                }

            }

        }

        // now we need to handle instances of runtime providable with no associated
        // runtime component definition

        for (RootComponent rootComponent : getComponents()) {

            if (!(rootComponent instanceof RuntimeProvidable)) {
                continue;
            }

            RuntimeProvidable rp = (RuntimeProvidable)rootComponent;
            if (rootComponent.runtimeProvider != null) {
                continue;
            }

            // if runtime providable is instance of application or engine, then
            // we assume it's a remote partner definition. messages sent to this
            // system addressed to an engine with no provider will bounce, i.e.
            // no transport module definitions

            if (rp instanceof Application) {
                continue;
            }

            if (rp instanceof Engine) {
                continue;
            }

            // if runtime providable is an instance of schema, provider definition
            // was either inlined within wsdl, in which case schema factory added
            // provider via add provider method, or provider was defined standalone
            // in which case provider definition was read above. provider def should
            // never be null

            if (rp instanceof Schema) {
                throw new DeploymentException("Runtime provider for Schema component "
                        + rootComponent.getFragmentIdentifier()
                        + " is undefined.");
            }

            // if runtime providable is an instance of policy, provider definition
            // MUST be inlined within wsdl, i.e. standalone definitions are not
            // allowed (no target namespace to acquire). the provider def should
            // never be null

            if (rp instanceof Policy) {
                throw new DeploymentException("Runtime provider for Policy component "
                        + rootComponent.getFragmentIdentifier()
                        + " is undefined.");
            }

            // if runtime providable is instance of binding, then we retrieve an
            // arbitrary provider of the indicated type

            if (rp instanceof Binding) {
                String bt = ((Binding)rp).getBindingType();
                BindingFactory bf = null;
                for (BindingFactory tmp : SystemContext.getContext().getSystem().getBindingFactories()) {
                    if (tmp.getBindingType().equals(bt)) {
                        bf = tmp;
                        break;
                    }
                }
                if (bf == null) {
                    throw new DeploymentException("Factory which generates Bindings of wsdl type "
                            + bt
                            + " is undefined.");
                }
                BindingRT dp = bf.createDefaultBinding((ComponentContext)deploymentContext, (Binding)rp);
                internalAddProvider(dp);
                rootComponent.setRuntimeProvider(dp);
            }

            // runtime provider is always undefined for abstract messages, i.e.
            // message instances are not runtime singletons and constructor
            // does not perform the 'match up'. check to see if a provider has
            // already been defined. if not we create 'default' message provider
            // as required by spec

            if (rp instanceof InterfaceMessage) {
                InterfaceMessage am = (InterfaceMessage)rp;
                Message dp = (Message)getProvider(ComponentType.MESSAGE, am.getName());
                if (dp == null) {
                    MessageFactory mf = SystemContext.getContext().getSystem().getDefaultMessageFactory();
                    if (mf == null) {
                        throw new DeploymentException("'default' message factory is undefined");
                    }
                    dp = mf.createDefaultMessage(am);
                    internalAddProvider(dp);
                }
                rootComponent.setRuntimeProvider(dp);
            }

        }

    }

    @Override
    public String toString() {
        return getRef();
    }

    /**
     * Contains a thread local variable which maintains value of currently referencing instance
     * of <code>ComponentReference</code>. Used to generate a 'reference trail' to aid in
     * resolving circular references detected during deployment.
     */
    public static class CurrentReference {

        private static final ThreadLocal<String> currentReference = new ThreadLocal<String>() {
        };

        public static String getCurrentReference() {
            return currentReference.get();
        }

        public static void setCurrentReference(String referenceKey) {
            currentReference.set(referenceKey);
        }

    }

    /**
     * Thrown by deployment to indicate that a definition added to component differs from an
     * existing definition with same name and of the same type.
     */
    public static class DefinitionMismatchException extends RuntimeException {

        private static final long serialVersionUID = 1L;

        public DefinitionMismatchException() {
            super();
        }

        public DefinitionMismatchException(String message) {
            super(message);
        }

    }

    /*
     * Adds runtimeProviders without performing a comparison, i.e. throws an exception if one
     * already exists.
     */
    private void internalAddProvider(RootComponentRT runtimeProvider) throws DeploymentException {

        ComponentName cn = runtimeProvider.getName();
        ComponentType ct = runtimeProvider.getComponentType();

        Map<ComponentName, RootComponentRT> typedProviders = providers.get(ct);
        if (typedProviders == null) {
            typedProviders = new HashMap<ComponentName, RootComponentRT>();
            providers.put(ct, typedProviders);
        }

        if (typedProviders.get(cn) != null) {
            throw new DeploymentException("Definition for runtime provider "
                    + runtimeProvider.getFragmentIdentifier()
                    + " already defined.");
        } else {
            typedProviders.put(cn, runtimeProvider);
        }

    }

}
