/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component;

import java.io.ByteArrayOutputStream;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantReadWriteLock;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.Deployment;
import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.DeploymentManager;
import org.bluestemsoftware.specification.eoa.SharedClassLoader;
import org.bluestemsoftware.specification.eoa.component.ClasspathDeployment.BuiltInComponentsDeployment;
import org.bluestemsoftware.specification.eoa.component.ComponentReference.SchemaComponentReference;
import org.bluestemsoftware.specification.eoa.component.ComponentReference.SchemaPropertyReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.types.SchemaComponent.Type;
import org.bluestemsoftware.specification.eoa.ext.ExecutableException;
import org.bluestemsoftware.specification.eoa.system.ManagementContext;
import org.bluestemsoftware.specification.eoa.system.System.Log;

/**
 * Manages <code>RootComponent</code> deployments on behalf of <code>System</code>. This
 * class is threadsafe.
 */
public class ComponentManager extends DeploymentManager {

    private Log log;
    private Set<ManagementContext> managementContexts;
    private Map<String, ComponentDeployment> deploymentsByRef;
    private Map<FragmentIdentifier, ComponentDeployment> deploymentsByComponentFragmentID;

    private static final ReentrantReadWriteLock readWriteLock = new ReentrantReadWriteLock();
    private static final Lock readLock = readWriteLock.readLock();
    private static final Lock writeLock = readWriteLock.writeLock();

    static final String LINE_BREAK = System.getProperty("line.separator");

    /**
     * Construct component manager.
     * 
     * @param sharedClassLoader
     */
    public ComponentManager(Log log, SharedClassLoader sharedClassLoader, Set<ManagementContext> managementContexts) {
        if (log == null) {
            throw new IllegalArgumentException("log null");
        } else {
            this.log = log;
        }
        this.managementContexts = managementContexts;
        this.sharedClassLoader = sharedClassLoader;
        deploymentsByRef = new HashMap<String, ComponentDeployment>();
        deploymentsByComponentFragmentID = new HashMap<FragmentIdentifier, ComponentDeployment>();
    }

    /**
     * Retrieves <code>RootComponent</code> by fragmentID.
     * 
     * @param fragmentID
     * @return RootComponent or <code>null</code> if not found
     */
    public RootComponent getComponent(FragmentIdentifier fragmentID) {
        readLock.lock();
        try {
            ComponentDeployment deployment = deploymentsByComponentFragmentID.get(fragmentID);
            if (deployment != null) {
                return deployment.getComponent(fragmentID);
            } else {
                return null;
            }
        } finally {
            readLock.unlock();
        }
    }

    /**
     * Retrieves <code>RootComponent</code> by fragmentID.
     * 
     * @param fragmentID
     * @return RootComponent or <code>null</code> if not found
     */
    public ComponentDeployment getComponentDeployment(String deploymentRef) {
        readLock.lock();
        try {
            return deploymentsByRef.get(deploymentRef);
        } finally {
            readLock.unlock();
        }
    }

    /**
     * Retrieves <code>RootComponent</code> by type and qualified name.
     * 
     * @param componentType
     * @param componentName
     * @return RootComponent or <code>null</code> if not found
     */
    public RootComponent getComponent(ComponentType componentType, QName componentName) {
        if (componentType == null) {
            throw new IllegalArgumentException("componentType null");
        }
        if (componentName == null) {
            throw new IllegalArgumentException("componentName null");
        }
        return getComponent(FragmentIdentifier.valueOf(componentType, componentName));
    }

    /**
     * Invokes {@link ComponentDeployment#suspendAll()} on all contained deployments.
     * 
     * @throws ExecutableException
     */
    public synchronized void suspendAll() throws ExecutableException {
        for (ComponentDeployment temp : deploymentsByRef.values()) {
            temp.suspendAll();
        }
    }

    /**
     * Invokes {@link ComponentDeployment#resumeAll()} on all contained deployments.
     * 
     * @throws ExecutableException
     */
    public synchronized void resumeAll() throws ExecutableException {
        for (ComponentDeployment temp : deploymentsByRef.values()) {
            temp.resumeAll();
        }
    }

    /*
     * (non-Javadoc)
     * 
     * Finds referenced component by fragmentID, and deploys referenced deployment lazily, if
     * necessary, i.e. if deployment within which component is defined has not yet deployed its
     * component definitions, it is first deployed.
     */
    <R extends RootComponent> R findReferencedComponent(ComponentReference<R> componentReference) throws DeploymentException {
        readLock.lock();
        try {

            ComponentDeployment deployment = null;

            if (componentReference instanceof SchemaComponentReference) {
                SchemaComponentReference scr = (SchemaComponentReference)componentReference;
                String targetNamespace = scr.getReferencedComponentName().getNamespaceURI();
                Type sct = scr.getSchemaComponentType();
                String scn = scr.getSchemaComponentName();
                FragmentIdentifier fragmentID = null;
                for (ComponentDeployment temp : deploymentsByRef.values()) {
                    fragmentID = temp.findSchema(targetNamespace, sct, scn);
                    if (fragmentID != null) {
                        deployment = deploymentsByComponentFragmentID.get(fragmentID);
                        componentReference = newComponentReference(fragmentID, componentReference);
                        break;
                    }
                }
            } else if (componentReference instanceof SchemaPropertyReference) {
                SchemaPropertyReference spr = (SchemaPropertyReference)componentReference;
                String targetNamespace = spr.getReferencedComponentName().getNamespaceURI();
                String propertyName = spr.getSchemaPropertyName();
                FragmentIdentifier fragmentID = null;
                for (ComponentDeployment temp : deploymentsByRef.values()) {
                    fragmentID = temp.findSchema(targetNamespace, propertyName);
                    if (fragmentID != null) {
                        deployment = deploymentsByComponentFragmentID.get(fragmentID);
                        componentReference = newComponentReference(fragmentID, componentReference);
                        break;
                    }
                }
            } else {
                ComponentName cname = componentReference.getReferencedComponentName();
                ComponentType ctype = componentReference.getReferencedComponentType();
                FragmentIdentifier fragmentID = FragmentIdentifier.valueOf(ctype, cname);
                deployment = deploymentsByComponentFragmentID.get(fragmentID);
            }

            if (deployment != null) {
                if (deployment.componentsDeployed() != Boolean.TRUE) {
                    deployment.deployComponents(log);
                    if (deployment.isError()) {
                        return null;
                    }
                }
                return deployment.getComponent(componentReference);
            } else {
                return null;
            }

        } finally {
            readLock.unlock();
        }
    }

    /**
     * Used by <code>System</code> to 'insert' a deployment. Unpacks deployment archive to a
     * temporary location, parses deployment descriptor and loads component definition(s)
     * contained within deployment.
     * <p>
     * If a component with a matching fragment id has already been deployed, i.e. within
     * another deployment, the insert will fail. Consequently, if a component is defined
     * remotely, and is referenced by one or more components from another namespace, the
     * remotely defined component MUST be deployed via a wsdl include, i.e. as opposed to using
     * wsdl import statements within referencing component defintions.
     * <p>
     * Note that the insert is atomic, component definitions defined within deployment are
     * inserted as a group, i.e. if an error occurs, the insert fails for all components.
     * 
     * @param deployment
     * @param componentContext
     *        which SHOULD have been created by invoking
     *        {@link #createDeploymentContext(Deployment)}, e.g. unless context was created
     *        within context of a testing framework.
     */
    public void insert(ComponentContext componentContext) {
        if (componentContext == null) {
            throw new IllegalArgumentException("componentContext null");
        }
        ComponentDeployment deployment = componentContext.getDeployment();
        if (deployment == null) {
            throw new IllegalArgumentException("deployment null");
        } else {
            deployment.setComponentManager(this);
        }
        writeLock.lock();
        try {

            String ref = deployment.getRef();
            if (deployment instanceof BuiltInComponentsDeployment) {
                if (log.isDebugEnabled()) {
                    log.debug("loading deployment " + ref);
                }
            } else {
                log.info("loading deployment " + deployment.getArtifactID() + "/" + deployment.getVersion());
            }
            if (deploymentsByRef.containsKey(ref)) {
                log.error("Insert failed. Deployment " + ref + " already inserted.");
                return;
            }

            try {
                deployment.load(componentContext, log);
            } catch (Throwable th) {
                String errorMessage = null;
                if (th instanceof DeploymentException) {
                    errorMessage = th.toString();
                } else {
                    StringBuilder sb = new StringBuilder();
                    try {
                        ByteArrayOutputStream baos = new ByteArrayOutputStream();
                        th.printStackTrace(new PrintStream(baos, true, "UTF-8"));
                        sb.append(LINE_BREAK);
                        sb.append(new String(baos.toByteArray(), "UTF-8"));
                    } catch (Exception exception) {
                        sb.append(th.toString());
                    }
                    errorMessage = sb.toString();
                }
                log.error(LINE_BREAK
                        + "Error inserting deployment "
                        + deployment.getRef()
                        + ". "
                        + LINE_BREAK
                        + errorMessage);
                return;
            }

            // verify that component is not defined within another deployment. different
            // images of same component in different deployments would be unmanageable

            for (RootComponent component : deployment.getComponents()) {
                if (deploymentsByComponentFragmentID.get(component.getFragmentIdentifier()) != null) {
                    ComponentDeployment existing = deploymentsByComponentFragmentID.get(component
                            .getFragmentIdentifier());
                    log.error("Error inserting deployment "
                            + deployment.getRef()
                            + ". Definition for RootComponent "
                            + component.getFragmentIdentifier()
                            + " already deployed within deployment "
                            + existing.getRef()
                            + ".");
                    return;
                }
            }

            // if we make it this far, we can perform the actual insert, i.e. into our
            // collection(s)

            for (RootComponent component : deployment.getComponents()) {
                deploymentsByComponentFragmentID.put(component.getFragmentIdentifier(), deployment);
            }

            deploymentsByRef.put(ref, deployment);

        } finally {
            writeLock.unlock();
        }
    }

    /**
     * Updates an existing deployment by first performing a delete and then performing an
     * insert.
     * 
     * TODO: should update be atomic?
     * 
     * @param deployment
     * @throws DeploymentException
     */
    public void update(ComponentDeployment deployment) throws DeploymentException {
        if (deployment == null) {
            throw new IllegalArgumentException("deployment null");
        }
        if (deployment instanceof BuiltInComponentsDeployment) {
            throw new DeploymentException("Deployment contains 'built-in' component definition.");
        }
        log.info("updating deployment " + deployment.getRef());
        writeLock.lock();
        try {
            delete(deployment);
            insert(deployment.getComponentContext());
        } finally {
            writeLock.unlock();
        }
    }

    /**
     * Used by <code>System</code> to 'delete' a deployment. Note that the delete is atomic,
     * component definitions defined within deployment are deleted as a group, i.e. if an error
     * occurs, the delete fails for all components.
     * 
     * @param deployment
     * @throws DeploymentException
     *         If an error occurs deleting deployment.
     */
    public void delete(ComponentDeployment deployment) throws DeploymentException {
        if (deployment == null) {
            throw new IllegalArgumentException("deployment null");
        }
        if (deployment instanceof BuiltInComponentsDeployment) {
            throw new DeploymentException("Deployment contains 'built-in' component definition.");
        }
        if (log.isDebugEnabled()) {
            log.debug("deleting deployment " + deployment.getRef());
        }
        writeLock.lock();
        try {
            String ref = deployment.getRef();
            if (!deploymentsByRef.containsKey(ref)) {
                log.error(LINE_BREAK + "delete failed. Deployment ref " + ref + " not found.");
                return;
            }
            deployment.unload();
            for (RootComponent component : deployment.getComponents()) {
                deploymentsByComponentFragmentID.remove(component.getFragmentIdentifier());
            }
            deploymentsByRef.remove(ref);
        } finally {
            writeLock.unlock();
        }
    }

    /**
     * Signals the end of 'load phase' by <code>System</code>. An indication that all
     * deployments have been loaded and that 'deploy phase' and 'init' phase can now be
     * initiated.
     * 
     * @return flag indicating whether or not all deployments were succesfully deployed.
     */
    public boolean commit() {
        log.debug("commit begin");
        boolean success = true;
        writeLock.lock();
        try {
            Iterator<ComponentDeployment> itr = deploymentsByRef.values().iterator();
            while (itr.hasNext()) {
                ComponentDeployment deployment = itr.next();
                deployment.deployComponents(log);
                if (deployment.isError()) {
                    for (RootComponent component : deployment.getComponents()) {
                        deploymentsByComponentFragmentID.remove(component.getFragmentIdentifier());
                    }
                    itr.remove();
                    deploymentsByRef.remove(deployment.getRef());
                    success = false;
                }
            }
            itr = deploymentsByRef.values().iterator();
            while (itr.hasNext()) {
                ComponentDeployment deployment = itr.next();
                deployment.deployProviders(log, managementContexts);
                if (deployment.isError()) {
                    for (RootComponent component : deployment.getComponents()) {
                        deploymentsByComponentFragmentID.remove(component.getFragmentIdentifier());
                    }
                    itr.remove();
                    success = false;
                }
            }
            itr = deploymentsByRef.values().iterator();
            while (itr.hasNext()) {
                ComponentDeployment deployment = itr.next();
                deployment.modelOperationStyles(log);
                if (deployment.isError()) {
                    for (RootComponent component : deployment.getComponents()) {
                        deploymentsByComponentFragmentID.remove(component.getFragmentIdentifier());
                    }
                    itr.remove();
                    success = false;
                }
            }
        } finally {
            writeLock.unlock();
        }
        log.debug("commit end");
        return success;
    }

    /**
     * Invoked by <code>System</code> during shutdown. Attempts to unload all registered
     * deployments after first suspending all instances of <code>ExecutableComponent</code>.
     * 
     * @return boolean if component manager was stopped successfully. false if it was stopped
     *         with errors.
     */
    public boolean stop() {
        boolean success = true;
        writeLock.lock();
        try {
            log.info("deleting component deployments");
            Iterator<ComponentDeployment> itr = deploymentsByRef.values().iterator();
            while (itr.hasNext()) {
                ComponentDeployment deployment = itr.next();
                try {
                    if (log.isDebugEnabled()) {
                        log.debug("deleting deployment " + deployment.getRef());
                    }
                    deployment.unload();
                } catch (Throwable th) {
                    success = false;
                    log.error("Error deleting deployment. " + th);
                }
            }
            deploymentsByRef = null;
            deploymentsByComponentFragmentID = null;
        } finally {
            writeLock.unlock();
        }
        return success;
    }

    /*
     * creates a new componentReference which is a copy of target component reference except
     * with local name updated to actual local name of schema, i.e. it was unknown up until
     * this point, i.e. it's 'foo'. this object will be used to retrieve component from
     * deployment and then discarded
     */
    private <R extends RootComponent> ComponentReference<R> newComponentReference(FragmentIdentifier fid, ComponentReference<R> target) {
        String ln = fid.getPointerPart();
        ComponentName cn = new ComponentName(target.getReferencedComponentName().getNamespaceURI(), ln);
        ComponentType ct = target.getReferencedComponentType();
        Class<R> clazz = target.getReferencedComponentClass();
        return new ComponentReferenceInternal<R>(ct, cn, clazz);
    }

    private static class ComponentReferenceInternal<R extends RootComponent> implements ComponentReference<R> {

        private ComponentType componentType;
        private ComponentName componentName;
        private Class<R> componentClass;

        public ComponentReferenceInternal(ComponentType componentType, ComponentName componentName,
                Class<R> componentClass) {
            this.componentType = componentType;
            this.componentName = componentName;
            this.componentClass = componentClass;
        }

        public FragmentIdentifier getFragmentIdentifier() {
            return new FragmentIdentifier("foo", Scheme.EOA, "bar", "");
        }

        public R getReferencedComponent() {
            throw new UnsupportedOperationException();
        }

        public ComponentName getReferencedComponentName() {
            return componentName;
        }

        public ComponentType getReferencedComponentType() {
            return componentType;
        }

        public Class<R> getReferencedComponentClass() {
            return componentClass;
        }

    }

}
