/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.application;

import java.net.URI;
import java.net.URISyntaxException;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.intrface.Interface;

/**
 * Defines an <code>Application</code> 'role'. A role's requirements are defined by
 * <code>Interface</code> referenced by <code>Role</code>.
 */
public final class Role extends NestedComponent<Application, Application> implements ComponentReference<Interface> {

    private String name;
    private ComponentName interfaceName;
    private Interface intrface;

    /**
     * Constructs a Role component.
     * 
     * @param name
     * @param interfaceName
     */
    public Role(String name, ComponentName interfaceName) {
        if (interfaceName == null) {
            throw new IllegalArgumentException("interfaceName null");
        }
        if (name == null) {
            throw new IllegalArgumentException("name null");
        }
        this.interfaceName = interfaceName;
        try {
            new URI(name);
        } catch (URISyntaxException ue) {
            throw new IllegalArgumentException("Role name is not a valid URI. " + ue.getMessage());
        }
        this.name = name;
    }

    /**
     * Returns name of role which MAY be used by Messaging protocol, e.g. SOAP, to target
     * protocol specific headers to a specific <code>Service</code>.
     * <p>
     * Note that <code>null</code> is a valid value as long as it is unique within context of
     * <code>Application</code>. This is to accomodate non-compliant frameworks, i.e.
     * service based frameworks, which require protocol specific headers without role
     * identification.
     * 
     * @return
     */
    public String getName() {
        return name;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentName()
     */
    public ComponentName getReferencedComponentName() {
        return interfaceName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentType()
     */
    public ComponentType getReferencedComponentType() {
        return ComponentType.INTERFACE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponent()
     */
    public Interface getReferencedComponent() {
        return intrface;
    }

    final void establishReference() throws DeploymentException {
        if (intrface != null) {
            return;
        }
        intrface = getRootComponent().getComponentContext().findReferencedComponent(this);
        if (intrface == null) {
            throw new DeploymentException("ComponentReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. Referenced Interface "
                    + interfaceName
                    + " is undefined.");
        }
    }

    final void validate() throws DeploymentException {
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentClass()
     */
    public Class<Interface> getReferencedComponentClass() {
        return Interface.class;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append(name);
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "role",
                    pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(Application parent) {
        this.parent = parent;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(name);
        sb.append(System.getProperty("line.separator"));
        sb.append("interfaceName: ");
        sb.append(interfaceName);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(name) + hashCode(interfaceName);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof Role)) {
            return false;
        }

        if ((name == ((Role)obj).name || (name != null && name.equals(((Role)obj).name)))
                && (interfaceName == ((Role)obj).interfaceName || (interfaceName != null && interfaceName
                        .equals(((Role)obj).interfaceName)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}