/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.application.rt;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.application.Role;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;

/**
 * An abstract <code>RoleReference</code> implementation which MAY be extended by type
 * specific descendants as required.
 */
public abstract class AbstractRoleReference implements RoleReference {

    protected FragmentIdentifier fragmentIdentifier;
    protected String roleName;
    protected Role role;
    protected ElementPolicy elementPolicy;
    protected ApplicationReference parent;

    public AbstractRoleReference(String roleName, ElementPolicy elementPolicy) {
        if (roleName == null) {
            throw new IllegalArgumentException("roleName null");
        } else {
            this.roleName = roleName;
        }
        this.elementPolicy = elementPolicy;
    }

    /**
     * Identifies name of <code>Role</code> on referenced partner component.
     * 
     * @return name of referenced <code>Role</code>.
     */
    public String getRoleName() {
        return roleName;
    }

    /**
     * Gets <code>Role</code> defined on referenced partner component.
     * @return
     */
    public Role getRole() {
        return role;
    }

    /**
     * Gets <code>ElementPolicy</code> attached to this component.
     * 
     * @return elementPolicy or <code>null</code> if no policy attached
     */
    public ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getParent()
     */
    public ApplicationReference getParent() {
        return parent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getRootComponent()
     */
    public ApplicationRT getRootComponent() {
        return parent.getParent();
    }

    protected void establishReference() throws DeploymentException {
        if (role != null) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        role = parent.getReferencedComponent().getRole(roleName);
        if (role == null) {
            throw new DeploymentException("RoleReference "
                    + getFragmentIdentifier()
                    + " is invalid. Role "
                    + roleName
                    + " not defined on partner Application "
                    + parent.getReferencedComponent().getName());
        }
    }

    protected void validate() throws DeploymentException {
        // TODO: implement
    }

    protected void setParent(AbstractApplicationReference parent) {
        this.parent = (ApplicationReference)parent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.ComponentRT#getFragmentIdentifier()
     */
    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append("roleName");
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "applicationReference", pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("roleName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(roleName);
        return sb.toString();
    }

}
