/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.binding;

import java.util.Collection;
import java.util.Collections;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;

public abstract class BindingOperation extends NestedComponent<Binding, Binding> {

    protected QName name;
    protected BindingMessageReference inputMessageReference;
    protected BindingMessageReference outputMessageReference;
    protected Map<QName, ? extends BindingFaultReference> faultReferences;
    protected ElementPolicy elementPolicy;

    public BindingOperation(QName name, BindingMessageReference inputMessageReference,
            BindingMessageReference outputMessageReference,
            Map<QName, ? extends BindingFaultReference> faultReferences, ElementPolicy elementPolicy) {
        super();
        if (name == null) {
            throw new IllegalArgumentException("name null");
        } else {
            this.name = name;
        }
        if (inputMessageReference == null) {
            throw new IllegalArgumentException("inputMessageReference null");
        } else {
            this.inputMessageReference = inputMessageReference;
        }
        if (faultReferences == null) {
            throw new IllegalArgumentException("faultReferences null");
        } else {
            this.faultReferences = faultReferences;
        }
        this.outputMessageReference = outputMessageReference;
        this.elementPolicy = elementPolicy;
    }

    /**
     * Gets name of bound interface operation.
     * <p>
     * 
     * @return operationName Qualified name of bound operation. Namespace is same as interface
     *         namespace, unless operation was inherited from an interface extended by bound
     *         interface.
     */
    public QName getName() {
        return name;
    }

    /**
     * Retrieves object which contains a reference to a concrete message. Corresponding
     * InterfaceMessageReference direction is "In".
     * 
     * @return bindingMessageReference
     */
    public BindingMessageReference getInputMessageReference() {
        return inputMessageReference;
    }

    /**
     * Retrieves object which contains a reference to a concrete message. Corresponding
     * InterfaceMessageReference direction is "Out".
     * 
     * @return bindingMessageReference
     */
    public BindingMessageReference getOutputMessageReference() {
        return outputMessageReference;
    }

    /**
     * If bound interface operation uses IN_OUT, retrieves object containing reference to
     * metadata describing concrete message format associated with abstract fault thrown by
     * this operation. Fault is declared at binding granularity.
     * <p>
     * 
     * @param faultName
     *        Qualified name of fault
     * @return BindingFaultReference object, or null if no fault found matching faultName or
     *         operation uses IN_ONLY mep.
     */
    public BindingFaultReference getFaultReference(QName faultName) {
        if (faultReferences == null) {
            return null;
        }
        return faultReferences.get(faultName);
    }

    /**
     * If bound interface operation uses IN_OUT, retrieves objects each of which contains a
     * reference to metadata describing concrete message format associated with an abstract
     * fault thrown by this operatation. All fault references are returned.
     * <p>
     * 
     * @return Collection of BindingFaultReference objects. If no faults are defined for this
     *         operation or operation mep is IN_ONLY, an empty Collection is returned.
     */
    @SuppressWarnings("unchecked")
    public Collection<? extends BindingFaultReference> getFaultReferences() {
        if (faultReferences == null) {
            return Collections.EMPTY_LIST;
        }
        return faultReferences.values();
    }

    /**
     * Gets <code>ElementPolicy</code> attached to component.
     * 
     * @return ElementPolicy or <code>null</code> if no policy attached.
     */
    public ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    /**
     * Identity method which MUST be overriden by implementation. When an attempt is made to
     * import a <code>Binding</code> with the same name as a currently loaded
     * <code>Binding</code>, <code>Binding</code> components are compared for equality.
     * <p>
     * This method is used for error reporting and MUST detail the values which were used to
     * test for equalitiy.
     * 
     * @return String
     */
    public String toString() {
        return null;
    }

    /**
     * Identity method which MUST be overriden by implementation. When an attempt is made to
     * import a <code>Binding</code> with the same name as a currently loaded
     * <code>Binding</code>, <code>Binding</code> components are compared for equality.
     * 
     * @return String
     */
    public int hashCode() {
        return -1;
    }

    /**
     * Identity method which MUST be overriden by implementation. When an attempt is made to
     * import a <code>Binding</code> with the same name as a currently loaded
     * <code>Binding</code>, <code>Binding</code> components are compared for equality.
     * 
     * @return da
     */
    public boolean equals(Object obj) {
        return false;
    }

    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(rootName.getLocalPart());
            pointerPart.append("/");
            pointerPart.append("xmlns(ns=");
            pointerPart.append(name.getNamespaceURI());
            pointerPart.append(")");
            pointerPart.append("ns:" + name.getLocalPart());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "bindingOperation", pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(Binding parent) {
        this.parent = parent;
        if (inputMessageReference != null) {
            inputMessageReference.setParent(this);
        }
        if (outputMessageReference != null) {
            outputMessageReference.setParent(this);
        }
        if (faultReferences != null) {
            for (BindingFaultReference faultReference : faultReferences.values()) {
                faultReference.setParent(this);
            }
        }
    }
    
}