/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.engine;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.application.Application;

/**
 * References <code>Application</code> implemented by parent <code>Engine</code>.
 * <p>
 * Note in the future, this could be used to attach private policy which could be intersected
 * with public policy defined on reference <code>Application</code>, i.e. once policy with
 * subject APPLICATION is considered portable.
 */
public final class ApplicationReference extends NestedComponent<Engine, Engine> implements ComponentReference<Application> {

    private ComponentName applicationName;
    private Application application;

    public ApplicationReference(ComponentName applicationName) {
        this.applicationName = applicationName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentName()
     */
    public ComponentName getReferencedComponentName() {
        return applicationName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentType()
     */
    public ComponentType getReferencedComponentType() {
        return ComponentType.APPLICATION;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponent()
     */
    public Application getReferencedComponent() {
        return application;
    }

    protected void establishReference() throws DeploymentException {
        if (application != null) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        application = getRootComponent().getComponentContext().findReferencedComponent(this);
        if (application == null) {
            throw new DeploymentException("ComponentReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. Referenced Application "
                    + applicationName
                    + " is undefined.");
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    public FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            String pointerPart = rootName.getLocalPart();
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "applicationReference", pointerPart.toString());
        }
        return fragmentIdentifier;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentClass()
     */
    public Class<Application> getReferencedComponentClass() {
        return Application.class;
    }

    protected void setParent(Engine parent) {
        this.parent = parent;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("applicationName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(applicationName);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(applicationName);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof ApplicationReference)) {
            return false;
        }

        if ((applicationName == ((ApplicationReference)obj).applicationName || (applicationName != null && applicationName
                .equals(((ApplicationReference)obj).applicationName)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}
