/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.engine;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.RuntimeProvidable;
import org.bluestemsoftware.specification.eoa.component.application.Application;
import org.bluestemsoftware.specification.eoa.component.application.Role;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EngineRT;

/**
 * Models an application 'engine'. An <code>Engine</code> implements an
 * <code>Application</code>, i.e. as a <code>Service</code> implements an
 * <code>Interface</code>.
 */
public final class Engine extends RootComponent implements RuntimeProvidable {

    private ApplicationReference applicationReference;
    private Set<Actor> actors;
    private Map<String, Actor> actorsByRoleName = new HashMap<String, Actor>();
    private Map<QName, Actor> actorsByServiceName = new HashMap<QName, Actor>();
    private boolean referencesEstablished;

    /**
     * Constructs an <code>Engine</code> instance. Note that runtime provider, i.e.
     * <code>EngineRT</code> is set by <code>System</code> when component is deployed.
     * 
     * @param name
     * @param applicationReference
     * @param actors
     */
    public Engine(ComponentName name, ApplicationReference applicationReference, Set<Actor> actors) {

        if (name == null) {
            throw new IllegalArgumentException("name null");
        } else {
            this.name = name;
        }
        fragmentIdentifier = new FragmentIdentifier(name.getNamespaceURI(), Scheme.EOA, "engine", name
                .getLocalPart());
        if (applicationReference == null) {
            throw new IllegalArgumentException("applicationReference null");
        } else {
            this.applicationReference = applicationReference;
            applicationReference.setParent(this);
        }
        if (actors == null) {
            throw new IllegalArgumentException("actors null");
        } else {
            this.actors = Collections.unmodifiableSet(actors);
            for (Actor actor : actors) {
                actorsByRoleName.put(actor.getRoleName(), actor);
                actorsByServiceName.put(actor.getReferencedComponentName(), actor);
                actor.setParent(this);
            }
        }

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent.RuntimeProvidable#getRuntimeProvider()
     */
    public EngineRT getRuntimeProvider() {
        if (runtimeProvider == null || !runtimeProvider.isDeployed()) {
            throw new IllegalStateException("Provider definition not yet deployed.");
        }
        return (EngineRT)runtimeProvider;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#getComponentType()
     */
    @Override
    public ComponentType getComponentType() {
        return ComponentType.ENGINE;
    }
    
    /**
     * Gets referenced <code>Application</code> implemented by <code>Engine</code>
     * 
     * @return applicationReference
     */
    public ApplicationReference getApplicationReference() {
        return applicationReference;
    }

    /**
     * Retrieves a specific Actor component by service name.
     * 
     * @param serviceName
     *        name of <code>Service</code> which performs role.
     * 
     * @return Actor or <code>null</code> if undefined.
     */
    public Actor getActor(QName serviceName) {
        return actorsByServiceName.get(serviceName);
    }

    /**
     * Retrieves a specific Actor component by role name.
     * 
     * @param roleName
     *        name of <code>Role</code> performed by <code>Actor</code>.
     * 
     * @return Actor or <code>null</code> if undefined.
     */
    public Actor getActor(String roleName) {
        return actorsByRoleName.get(roleName);
    }

    /**
     * Retrieves all <code>Actor</code> components. Each <code>Actor</code> performs a
     * <code>Role</code> defined on <code>Application</code> implemented by
     * <code>Engine</code>.
     * 
     * @return Collection of <code>Actor</code> components or empty Collection if none
     *         defined
     */
    public Collection<Actor> getActors() {
        return actors;
    }

    @Override
    protected synchronized void establishReferences() throws DeploymentException {

        if (referencesEstablished) {
            return;
        }

        if (applicationReference != null) {
            applicationReference.establishReference();
        }

        for (Actor actor : actors) {
            actor.establishReference();
        }

        referencesEstablished = true;
    }

    @Override
    protected void validate() throws DeploymentException {

        // verify that an actor is defined for each role defined on referenced
        // application

        Application application = applicationReference.getReferencedComponent();

        for (Role role : application.getRoles()) {
            if (getActor(role.getName()) == null) {
                throw new DeploymentException("Engine "
                        + fragmentIdentifier
                        + " is invalid. "
                        + "An actor MUST be defined for role "
                        + role.getName()
                        + " defined on Application "
                        + applicationReference.getReferencedComponentName()
                        + " implemented by Engine.");
            }
        }

        // verify that each actor performs a role defined on referenced
        // application

        for (Actor actor : actors) {
            if (application.getRole(actor.getRoleName()) == null) {
                throw new DeploymentException("Engine "
                        + fragmentIdentifier
                        + " is invalid. Role "
                        + actor.getRoleName()
                        + " performed by actor "
                        + actor.getFragmentIdentifier()
                        + " is not defined on Application "
                        + applicationReference.getReferencedComponentName()
                        + " implemented by Engine.");
            }
        }

        // allow actor components to validate their definition

        for (Actor actor : actors) {
            actor.validate();
        }

    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(name);
        sb.append(System.getProperty("line.separator"));
        sb.append("applicationReference: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(applicationReference);
        sb.append(System.getProperty("line.separator"));
        sb.append("actors: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(actors);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(name) + hashCode(applicationReference) + hashCode(actors);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof Engine)) {
            return false;
        }

        if ((name == ((Engine)obj).name || (name != null && name.equals(((Engine)obj).name)))
                && (applicationReference == ((Engine)obj).applicationReference || (applicationReference != null && applicationReference
                        .equals(((Engine)obj).applicationReference)))
                && (actors == ((Engine)obj).actors || (actors != null && actors.equals(((Engine)obj).actors)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}