/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.engine.rt;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference.RequestActionReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference.ResponseActionReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceFaultReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceMessageReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation.MEP;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.OperationPolicy;
import org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT;
import org.bluestemsoftware.specification.eoa.component.service.EndpointOperation;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyException;

/**
 * References an EndpointOperation.
 */

public class EndpointOperationReference implements NestedComponentRT<EngineRT, EndpointReference> {

    private FragmentIdentifier fragmentIdentifier;
    private EndpointReference parent;
    private QName endpointOperationName;
    private EndpointOperation endpointOperation;
    private OperationPolicy effectivePolicy;
    private ElementPolicy elementPolicy;
    private Map<String, EndpointActionReference> actionReferences;
    private RequestActionReference requestAction;
    private ResponseActionReference responseAction;
    private Map<QName, ResponseActionReference> faultActions;

    public EndpointOperationReference(QName endpointOperationName, ElementPolicy elementPolicy,
            Map<String, EndpointActionReference> actionReferences) {
        if (endpointOperationName == null) {
            throw new IllegalArgumentException("endpointOperationName null");
        }
        if (actionReferences == null) {
            throw new IllegalArgumentException("actionReferences null");
        } else {
            this.actionReferences = actionReferences;
            for (EndpointActionReference actionReference : actionReferences.values()) {
                actionReference.setParent(this);
            }
        }
        this.endpointOperationName = endpointOperationName;
        if (elementPolicy != null) {
            this.elementPolicy = elementPolicy;
        }
    }

    /**
     * Identifies name of referenced <code>EndpointOperation</code> component.
     * 
     * @return QName name of referenced <code>EndpointOperation</code>.
     */
    public final QName getEndpointOperationName() {
        return endpointOperationName;
    }

    /**
     * Gets referenced <code>EndpointOperation</code> component.
     * 
     * @return EndpointOperation referenced <code>EndpointOperation</code>.
     */
    public final EndpointOperation getEndpointOperation() {
        return endpointOperation;
    }

    /**
     * Gets <code>ElementPolicy</code> attached to this component.
     * 
     * @return elementPolicy or <code>null</code> if no policy attached
     */
    public final ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    /**
     * Returns the 'effective' policy for this Component, i.e. returns element Policy attached
     * to this component.
     * 
     * @return EffectivePolicy effective policy with policy subject OPERATION or null if no
     *         policy defined.
     */
    public final OperationPolicy getEffectivePolicy() {
        return effectivePolicy;
    }

    /**
     * Gets a specific EndpointActionReference object.
     * 
     * @param action
     *        URI which identifies referenced action
     * 
     * @return EndpointActionReference
     */
    public final EndpointActionReference getActionReference(String action) {
        if (actionReferences == null) {
            return null;
        }
        return actionReferences.get(action);
    }

    /**
     * Gets all EndpointActionReference objects.
     * 
     * @return Collection of EndpointActionReference objects
     */
    @SuppressWarnings("unchecked")
    public final Collection<EndpointActionReference> getActionReferences() {
        if (actionReferences == null) {
            return Collections.EMPTY_LIST;
        }
        return Collections.unmodifiableCollection(actionReferences.values());
    }

    /**
     * Convenience method used to retrieve action which represents the initial message in
     * exchange.
     * 
     * @return request action, or <code>null</code> if no actions defined.
     */
    public final RequestActionReference getRequestAction() {
        if (actionReferences == null) {
            return null;
        }
        if (requestAction == null) {
            for (EndpointActionReference ear : actionReferences.values()) {
                if (ear.getReferencedComponent().getInterfaceAction().getDirection() == Direction.IN) {
                    requestAction = (RequestActionReference)ear;
                    break;
                }
            }
        }
        return requestAction;
    }

    /**
     * Convenience method used to retrieve action which represents normal response within the
     * message in exchange.
     * 
     * @return response action, or <code>null</code> if no actions defined or if MEP not
     *         'In-Out'.
     */
    public final ResponseActionReference getResponseAction() {
        if (actionReferences == null) {
            return null;
        }
        if (endpointOperation.getInterfaceOperation().getMessageExchangePattern() != MEP.IN_OUT) {
            return null;
        }
        if (responseAction == null) {
            for (EndpointActionReference ear : actionReferences.values()) {
                InterfaceAction ia = ear.getReferencedComponent().getInterfaceAction();
                if (ia.getDirection() == Direction.OUT) {
                    if (ia instanceof InterfaceMessageReference) {
                        responseAction = (ResponseActionReference)ear;
                        break;
                    }
                }
            }
        }
        return responseAction;
    }

    /**
     * Convenience method used to retrieve fault action which represents fault response within
     * message in exchange.
     * 
     * @param faultName
     * @return fault action, or <code>null</code> if no actions defined or if MEP not
     *         'In-Out' or 'Robust-In-Only', or if fault undefined.
     */
    public final ResponseActionReference getFaultAction(QName faultName) {
        if (faultName == null) {
            throw new IllegalArgumentException("faultName null");
        }
        if (actionReferences == null) {
            return null;
        }
        if (endpointOperation.getInterfaceOperation().getMessageExchangePattern() == MEP.IN_ONLY) {
            return null;
        }
        if (faultActions == null) {
            faultActions = new HashMap<QName, ResponseActionReference>();
        }
        ResponseActionReference faultAction = faultActions.get(faultName);
        if (faultAction == null) {
            for (EndpointActionReference ear : actionReferences.values()) {
                InterfaceAction ia = ear.getReferencedComponent().getInterfaceAction();
                if (ia.getDirection() == Direction.OUT) {
                    if (ia instanceof InterfaceFaultReference) {
                        InterfaceFaultReference ifr = (InterfaceFaultReference)ia;
                        if (faultName.equals(ifr.getFault().getName())){
                            faultAction = (ResponseActionReference)ear;
                            break;
                        }
                    }
                }
            }
        }
        if (faultAction != null) {
            faultActions.put(faultName, faultAction);
        }
        return faultAction;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getParent()
     */
    public final EndpointReference getParent() {
        return parent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getRootComponent()
     */
    public final EngineRT getRootComponent() {
        return parent.getRootComponent();
    }

    /*
     * Used by parent OperationReference component to add an 'implied' action reference
     * autogenerated during deployment
     */
    void addImpliedActionReference(EndpointActionReference ear) {
        ear.setParent(this);
        actionReferences.put(ear.getAction(), ear);
    }

    void validate() throws DeploymentException {
        for (EndpointActionReference actionReference : actionReferences.values()) {
            actionReference.validate();
        }
    }

    void establishReference() throws DeploymentException {
        if (endpointOperation != null) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        endpointOperation = parent.getEndpoint().getEndpointOperation(endpointOperationName);
        if (endpointOperation == null) {
            throw new DeploymentException("EndpointOperationReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. EndpointOperation '"
                    + endpointOperationName
                    + "', not defined on Endpoint '"
                    + parent.getEndpoint().getName()
                    + "'.");
        }
        if (actionReferences != null) {
            for (EndpointActionReference actionReference : actionReferences.values()) {
                actionReference.establishReference();
            }
        }
    }

    void deriveEffectivePolicy() throws PolicyException {
        if (elementPolicy != null) {
            effectivePolicy = new OperationPolicy(elementPolicy.getMergedExpression());
        }
        if (actionReferences != null) {
            for (EndpointActionReference actionReference : actionReferences.values()) {
                actionReference.deriveEffectivePolicy();
            }
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.ComponentRT#getFragmentIdentifier()
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append("xmlns(ns=");
            pointerPart.append(endpointOperationName.getNamespaceURI());
            pointerPart.append(")");
            pointerPart.append("ns:" + endpointOperationName.getLocalPart());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "endpointOperationReference", pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(EndpointReference parent) {
        this.parent = parent;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("endpointOperationName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(endpointOperationName);
        sb.append(System.getProperty("line.separator"));
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(endpointOperationName);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof EndpointOperationReference)) {
            return false;
        }

        if ((endpointOperationName == ((EndpointOperationReference)obj).endpointOperationName || (endpointOperationName != null && endpointOperationName
                .equals(((EndpointOperationReference)obj).endpointOperationName)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}