/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.engine.rt;

import java.util.Map;
import java.util.Set;

import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageModule;
import org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT;
import org.bluestemsoftware.specification.eoa.system.server.Feature;

/**
 * Implements a <code>Feature</code> on a per message basis. <code>FeatureModule</code>
 * instances are enabled/constrained via policy attached to component model hierarchy.
 * 
 * @see ApplicationModule
 * @see BindingModule
 * @see TransportModule
 */
public interface FeatureModule extends NestedComponentRT<EngineRT, EndpointActionReference> {

    /**
     * Gets <code>Feature</code> implemented by module.
     * 
     * @return Feature
     */
    public Feature getFeature();

    /**
     * Contains a <code>Set</code> of specifications, each of which models a version specific
     * instance of a specification which defines module behavior, e.g. a ws-* specification.
     * <p>
     * If {@link ActorReference#isMyService()} is true, represents alternatives 'advertised'
     * via public policy. The choice is made by client and is implied by infoset used within
     * inbound request.
     * <p>
     * If {@link ActorReference#isMyService()} is false, represents 'negotiated' alternatives,
     * i.e. public policy intersected with private policy defined by engine user. In which
     * case, the <code>Set</code> SHOULD contain a single alternative, i.e. a single instance
     * of <code>WSSpecification</code> which contains a single instance of {@link Policy}.
     * 
     * @return Map of supported versions, each of which is mapped to a namespace which
     *         identifies version.
     */
    public Map<String, WSSpecification> getSupportedVersions();

    /**
     * Models a specific version of a 'web service' specification, e.g. a ws-* specification.
     * Defines version specific constants, specification namespace, infoset and a set of
     * alternative, runtime configurations configured via ws-policy.
     * 
     * @see FeatureModule#getSupportedVersions()
     */
    public interface WSSpecification {

        /**
         * Gets namespace which identifies the specification and its version.
         */
        public String getNamespace();

        /**
         * Gets version specific message header infoset defined by specification which MAY be
         * required to encode/decode abstract message properties, i.e. in cases where module
         * operates within binding layer and binding module is not natively implemented by
         * binding.
         * 
         * @return infoset
         * @see BindingModule
         * @see MessageModule#getVersion()
         */
        public Infoset getInfoSet();

        /**
         * Defines runtime 'properties' set via ws-policy. Note that the returned
         * <code>Set</code> MAY contain more than one configuration, i.e. one for each
         * supported alternative.
         * 
         * @return
         */
        public Set<Policy> getPolicy();

        /**
         * Models version specific message header infoset defined by specification which MAY be
         * required to encode/decode abstract message properties, i.e. in cases where feature
         * is implemented within binding layer and feature is not natively implemented by
         * binding.
         * <p>
         * Note that interface MAY contain only static constants, i.e. no defined behavior.
         * 
         * @see BindingFeature
         */
        public interface Infoset {
        }

        /**
         * Models policy specificiation defined by the web service specification. Policy
         * defines an immutable set of variables set via via ws-policy.
         */
        public interface Policy {

            /**
             * Models version specific policy assertion infoset defined by specification used
             * to read/write policy expressions which enable/configure underlying feature.
             * 
             * @return
             */
            public Infoset getInfoset();

        }

    }

}
