/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.engine.rt;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.application.rt.ApplicationReference;
import org.bluestemsoftware.specification.eoa.component.application.rt.RoleReference;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.binding.UnderlyingProtocol;
import org.bluestemsoftware.specification.eoa.component.engine.Actor;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference.RequestActionReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference.ResponseActionReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceAction.Direction;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ServicePolicy;
import org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT;
import org.bluestemsoftware.specification.eoa.component.service.Endpoint;
import org.bluestemsoftware.specification.eoa.component.service.EndpointAction;
import org.bluestemsoftware.specification.eoa.component.service.EndpointOperation;
import org.bluestemsoftware.specification.eoa.component.service.Service;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.vm.VMTransportProtocol;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyException;

/**
 * References a Service component.
 * 
 */
public class ServiceReference implements NestedComponentRT<EngineRT, EngineReference>, ComponentReference<Service> {

    private FragmentIdentifier fragmentIdentifier;
    private EngineReference parent;
    private ComponentName serviceName;
    private Service service;
    private boolean isMyService;
    private ElementPolicy elementPolicy;
    private ServicePolicy effectivePolicy;
    private Map<String, EndpointReference> endpointReferences;
    private String roleName;
    private String consumerRoleName;

    /**
     * Constructs a <code>ServiceReference</code> component.
     * @param serviceName
     * @param isMyService
     * @param elementPolicy
     *        or <code>null</code> if none defined
     * @param endpointReferences
     *        which MAY be empty, i.e. either referenced Service defines no endpoints or no
     *        private policy defined with subject ENDPOINT, OPERATION or MESSAGE. In the latter
     *        case, <code>EndpointReference</code>, <code>EndpointOperationReference</code>
     *        and <code>EndpointActionReference</code> components will be auto-generated to
     *        mirror components on referenced Service. If not already specified, auto-generated
     *        <code>EndpointReference</code> components which reference an endpoint whose
     *        referenced binding employs an instance of {@link VMTransportProtocol} will be
     *        positioned first within the iteration order. If more than one endpoint employs a
     *        virtual transport, the ordering among those endpoints is unspecified. Note that
     *        the order of contained components implies endpoint preference.
     */
    public ServiceReference(ComponentName serviceName, boolean isMyService, ElementPolicy elementPolicy,
            LinkedHashMap<String, EndpointReference> endpointReferences) {
        if (serviceName == null) {
            throw new IllegalArgumentException("serviceName null");
        } else {
            this.serviceName = serviceName;
        }
        if (endpointReferences == null) {
            throw new IllegalArgumentException("endpointReferences null");
        } else {
            this.endpointReferences = endpointReferences;
            for (EndpointReference endpointReference : endpointReferences.values()) {
                endpointReference.setParent(this);
            }
        }
        this.isMyService = isMyService;
        this.elementPolicy = elementPolicy;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getParent()
     */
    public final EngineReference getParent() {
        return parent;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.NestedComponentRT#getRootComponent()
     */
    public final EngineRT getRootComponent() {
        return parent.getParent();
    }

    /**
     * Indicates if referenced Service component is owned, i.e. 'hosted' by root Engine
     * component.
     * 
     * @return boolean true if referenced Service is hosted by parent <code>Engine</code>,
     *         false if referenced Service is a 'partner' Service, i.e. Service is hosted on a
     *         partner Engine.
     */
    public final boolean isMyService() {
        return isMyService;
    }

    /**
     * Gets name of role which maps to a <code>Role</code> on <code>Application</code>
     * implemented by referenced <code>Engine</code>. Identifies 'role' performed by
     * <code>Service</code>.
     * 
     * @return role URI which MUST be unique within context of parent component.
     */
    public final String getRoleName() {
        return roleName;
    }

    /**
     * Returns role implemented by consumers of referenced <code>Service</code>. The value
     * retrieved from {@link #getRoleName()} is used as a base URI against which the value
     * 'consumer' is resolved to create the consumer roleName.
     * <p>
     * The uri MAY be used to target response headers to ultimate destination, i.e. to node
     * receiving response which MAY be returned by referenced service.
     * 
     * @return
     */
    public final String getConsumerRoleName() {
        if (consumerRoleName == null) {
            try {
                String temp = roleName.endsWith("/") ? roleName : roleName + "/";
                consumerRoleName = temp + "consumer";
            } catch (Exception ex) {
                throw new IllegalStateException("Error constructing consumer roleName. " + ex);
            }
        }
        return consumerRoleName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentName()
     */
    public final ComponentName getReferencedComponentName() {
        return serviceName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentType()
     */
    public final ComponentType getReferencedComponentType() {
        return ComponentType.SERVICE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponent()
     */
    public final Service getReferencedComponent() {
        return service;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentClass()
     */
    public final Class<Service> getReferencedComponentClass() {
        return Service.class;
    }

    /**
     * Gets <code>ElementPolicy</code> attached to this component.
     * 
     * @return elementPolicy or <code>null</code> if no policy attached
     */
    public final ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    /**
     * Returns the 'effective' policy for this Component, i.e. returns element Policy attached
     * to this component.
     * 
     * @return EffectivePolicy effective policy with policy subject SERVICE or null if no
     *         policy defined.
     */
    public final ServicePolicy getEffectivePolicy() {
        return effectivePolicy;
    }

    /**
     * Gets a specific EndpointReference object owned by this object.
     * 
     * @return EndpointReference
     */
    public final EndpointReference getEndpointReference(String endpointName) {
        return endpointReferences.get(endpointName);
    }

    /**
     * Convenience method which retrieves corresponding <code>RoleReference</code> defined on
     * application implemented by parent engine.
     * 
     * @return
     * @see EngineReference#getCorrespondingApplicationReference()
     */
    public final RoleReference getCorrespondingRoleReference() {
        ApplicationReference applicationReference = parent.getCorrespondingApplicationReference();
        return applicationReference.getRoleReference(roleName);
    }

    /**
     * Returns an iterator which iterates over all <code>EndpointReference</code> components
     * in the order in which they were defined. Note that order MAY be used to indicate
     * endpoint preference.
     * <p>
     * Note that if note explicitly defined, <code>EndpointReference</code> components will
     * be auto-generated to mirror components on referenced Service. When auto-generated,
     * <code>EndpointReference</code> components which reference an endpoint whose referenced
     * binding employs an instance of {@link VMTransportProtocol} are positioned first within
     * the iteration order. If more than one endpoint employs a virtual transport, the ordering
     * amongst those endpoints is unspecified.
     * 
     * @return Collection of EndpointReference objects
     */
    public final Iterator<EndpointReference> getEndpointReferences() {
        return endpointReferences.values().iterator();
    }

    void validate() throws DeploymentException {

        EngineReference engineReference = (EngineReference)parent;
        if (engineReference.isMyEngine()) {
            if (!isMyService) {
                throw new DeploymentException("ServiceReference "
                        + getFragmentIdentifier()
                        + " is invalid. Object flagged with 'myService' == true,"
                        + " but parent EngineReference component not flagged as 'myEngine'.");
            }
        }

        if (getCorrespondingRoleReference() == null) {
            throw new DeploymentException("RoleReference which references Role "
                    + roleName
                    + " is undefined on ApplicationReference "
                    + parent.getCorrespondingApplicationReference().getFragmentIdentifier()
                    + ".");
        }

        for (EndpointReference endpointReference : endpointReferences.values()) {
            endpointReference.validate();
        }

    }

    final void establishReference() throws DeploymentException {
        if (service != null) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        Actor actor = parent.getReferencedComponent().getActor(serviceName);
        if (actor == null) {
            throw new DeploymentException("ServiceReference "
                    + getFragmentIdentifier()
                    + " is invalid. No corresponding Actor defined on Engine "
                    + parent.getReferencedComponent().getName()
                    + " referenced by parent component matching service name "
                    + serviceName
                    + ".");
        } else {
            service = actor.getReferencedComponent();
            roleName = actor.getRoleName();
        }
        generateImpliedStructure();
        for (EndpointReference endpointReference : endpointReferences.values()) {
            endpointReference.establishReference();
        }
        orderEndpointsByPreference();
    }

    void deriveEffectivePolicy() throws PolicyException {
        if (elementPolicy != null) {
            effectivePolicy = new ServicePolicy(elementPolicy.getMergedExpression());
        }
        for (EndpointReference endpointReference : endpointReferences.values()) {
            endpointReference.deriveEffectivePolicy();
        }
    }

    void setParent(EngineReference parent) {
        this.parent = parent;
    }

    /**
     * Used to generate implied structure which MAY be overridden by type specific descendants
     * as necessary.
     * 
     * @param endpointOperationName
     * @param elementPolicy
     * @param actionReferences
     * @return
     */
    protected EndpointOperationReference createEndpointOperationReference(QName endpointOperationName, Map<String, EndpointActionReference> actionReferences) {
        return new EndpointOperationReference(endpointOperationName, null, actionReferences);
    }

    /**
     * Used to generate implied structure which MAY be overridden by type specific descendants
     * as necessary.
     * 
     * @param endpointName
     * @param elementPolicy
     * @param operationReferences
     * @return
     */
    protected EndpointReference createEndpointReference(String endpointName, Map<QName, EndpointOperationReference> operationReferences) {
        return new EndpointReference(endpointName, null, operationReferences);
    }

    /**
     * Used to generate implied structure which MAY be overridden by type specific descendants
     * as necessary.
     * 
     * @param action
     * @return
     */
    protected RequestActionReference createRequestActionReference(String action, ElementPolicy elementPolicy) {
        return new RequestActionReference(action, elementPolicy);
    }

    /**
     * Used to generate implied structure which MAY be overridden by type specific descendants
     * as necessary.
     * 
     * @param action
     * @return
     */
    protected ResponseActionReference createResponseActionReference(String action, ElementPolicy elementPolicy) {
        return new ResponseActionReference(action, elementPolicy);
    }

    protected void generateImpliedStructure() throws DeploymentException {

        for (Endpoint endpoint : service.getEndpoints()) {
            String endpointName = endpoint.getName();
            EndpointReference endpointReference = getEndpointReference(endpointName);
            if (endpointReference == null) {
                Map<QName, EndpointOperationReference> operationReferences = new HashMap<QName, EndpointOperationReference>();
                for (EndpointOperation endpointOperation : endpoint.getEndpointOperations()) {
                    QName operationName = endpointOperation.getName();
                    Map<String, EndpointActionReference> actionReferences = new HashMap<String, EndpointActionReference>();
                    for (EndpointAction endpointAction : endpointOperation.getEndpointActions()) {
                        String action = endpointAction.getAction();
                        Direction dir = endpointAction.getInterfaceAction().getDirection();
                        if (dir == Direction.IN) {
                            actionReferences.put(action, createRequestActionReference(action, null));
                        } else {
                            actionReferences.put(action, createResponseActionReference(action, null));
                        }
                    }
                    EndpointOperationReference eor = createEndpointOperationReference(operationName,
                            actionReferences);
                    operationReferences.put(operationName, eor);
                }
                endpointReference = createEndpointReference(endpointName, operationReferences);
                endpointReference.setAutoGenerated(true);
                endpointReference.setParent(this);
                endpointReferences.put(endpointReference.getEndpointName(), endpointReference);
            } else {
                for (EndpointOperation endpointOperation : endpoint.getEndpointOperations()) {
                    QName operationName = endpointOperation.getName();
                    EndpointOperationReference eor = endpointReference.getOperationReference(operationName);
                    if (eor == null) {
                        Map<String, EndpointActionReference> actionReferences = new HashMap<String, EndpointActionReference>();
                        for (EndpointAction endpointAction : endpointOperation.getEndpointActions()) {
                            String action = endpointAction.getAction();
                            Direction dir = endpointAction.getInterfaceAction().getDirection();
                            if (dir == Direction.IN) {
                                actionReferences.put(action, createRequestActionReference(action, null));
                            } else {
                                actionReferences.put(action, createResponseActionReference(action, null));
                            }
                        }
                        eor = createEndpointOperationReference(operationName, actionReferences);
                        endpointReference.addImpliedOperationReference(eor);
                    } else {
                        for (EndpointAction endpointAction : endpointOperation.getEndpointActions()) {
                            String action = endpointAction.getAction();
                            Direction dir = endpointAction.getInterfaceAction().getDirection();
                            EndpointActionReference ear = eor.getActionReference(action);
                            if (ear == null) {
                                if (dir == Direction.IN) {
                                    ear = createRequestActionReference(action, null);
                                } else {
                                    ear = createResponseActionReference(action, null);
                                }
                            } else {
                                if (dir == Direction.IN) {
                                    ear = new RequestActionReference(action, ear.getElementPolicy());
                                } else {
                                    ear = new ResponseActionReference(action, ear.getElementPolicy());
                                }
                            }
                            eor.addImpliedActionReference(ear);
                        }
                    }
                }
            }
        }

    }

    private void orderEndpointsByPreference() {

        // currently collection is ordered according to order in which epr's
        // were listed within provider definition with autogenerated
        // epr's inserted at end of list. we need to move autogenerated
        // epr's that employ virtual transport to top of list. note that
        // this can't be determined until after references established. if
        // more than one autogen'd epr employs virtual transport, the
        // ordering amongst these epr's is undefined

        List<EndpointReference> tempList = new ArrayList<EndpointReference>();
        for (EndpointReference epr : endpointReferences.values()) {
            if (epr.isAutoGenerated()) {
                Binding binding = epr.getEndpoint().getBindingReference().getReferencedComponent();
                UnderlyingProtocol underlyingProtocol = binding.getUnderlyingProtocol();
                if (underlyingProtocol instanceof VMTransportProtocol) {
                    EngineReference engineReference = epr.getParent().getParent();
                    if (engineReference.isMyEngine()) {
                        // we know our rt engine is defined within vm
                        tempList.add(0, epr);
                    } else {
                        try {
                            engineReference.getReferencedComponent().getRuntimeProvider();
                            tempList.add(0, epr);
                        } catch (IllegalStateException ie) {
                            // do not add. runtime engine is undefined,
                            // i.e. the endpoint is remote
                        }
                    }
                } else {
                    tempList.add(epr);
                }
            } else {
                tempList.add(epr);
            }
        }

        endpointReferences = new LinkedHashMap<String, EndpointReference>();
        for (EndpointReference epr : tempList) {
            endpointReferences.put(epr.getEndpointName(), epr);
        }
        endpointReferences = Collections.unmodifiableMap(endpointReferences);

    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.rt.ComponentRT#getFragmentIdentifier()
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append("xmlns(ns=");
            pointerPart.append(serviceName.getNamespaceURI());
            pointerPart.append(")");
            pointerPart.append("ns:" + serviceName.getLocalPart());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA,
                    "serviceReference", pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("serviceName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(serviceName);
        sb.append(System.getProperty("line.separator"));
        sb.append("isMyService: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(isMyService);
        sb.append(System.getProperty("line.separator"));
        sb.append("endpointReferences: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(endpointReferences);
        return sb.toString();
    }

}