/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.intrface;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentDeployment;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.system.SystemContext;

/**
 * Models an <code>Interface</code> component.
 */

public final class Interface extends RootComponent {

    private boolean referencesEstablished;
    private Set<QName> extendedInterfaceNames;
    private Set<Interface> extendedInterfaces;
    private Map<QName, InterfaceOperation> operations;
    private Map<QName, InterfaceOperation> declaredOperations;
    private Map<QName, InterfaceFault> faults;
    private Map<QName, InterfaceFault> declaredFaults;
    private Map<String, InterfaceAction> actions;
    private Set<String> styleDefaults;
    private ElementPolicy elementPolicy;

    public static final String LINE_BREAK = System.getProperty("line.separator");

    /**
     * Constructs an <code>Interface</code> component.
     * @param interfaceName
     * @param extendedInterfaceNames
     *        Set of qualifed names each of which identifies an <code>Interface</code>
     *        extended by this <code>Interface</code>, or an empty set if none.
     * @param declaredOperations
     *        <code>InterfaceOperation</code> components directly defined on this
     *        <code>Interface</code> or an empty map if none defined.
     * @param declaredFaults
     *        <code>InterfaceFault</code> components directly defined on this
     *        <code>Interface</code> or an empty map if none defined.
     * @param styleDefault
     *        operation style defaults or an empty set if none defined.
     * @param localPolicy
     *        Local <code>PolicyExpression</code> attached to this component.
     */
    public Interface(ComponentName interfaceName, Set<QName> extendedInterfaceNames,
            Map<QName, InterfaceOperation> declaredOperations, Map<QName, InterfaceFault> declaredFaults,
            Set<String> styleDefaults, ElementPolicy elementPolicy) {

        super();

        if (interfaceName == null) {
            throw new IllegalArgumentException("componentName null");
        } else {
            this.name = interfaceName;
        }
        fragmentIdentifier = new FragmentIdentifier(name.getNamespaceURI(), Scheme.EOA, "interface", name
                .getLocalPart());
        if (declaredOperations == null) {
            throw new IllegalArgumentException("declaredOperations null");
        } else {
            this.operations = Collections.unmodifiableMap(declaredOperations);
            this.declaredOperations = Collections.unmodifiableMap(declaredOperations);
        }
        if (extendedInterfaceNames == null) {
            throw new IllegalArgumentException("extendedInterfaceNames null");
        } else {
            this.extendedInterfaceNames = Collections.unmodifiableSet(extendedInterfaceNames);
        }
        if (declaredFaults == null) {
            throw new IllegalArgumentException("declaredFaults null");
        } else {
            this.faults = Collections.unmodifiableMap(declaredFaults);
            this.declaredFaults = Collections.unmodifiableMap(declaredFaults);
        }
        if (styleDefaults == null) {
            throw new IllegalArgumentException("defaultStyle null");
        } else {
            this.styleDefaults = Collections.unmodifiableSet(styleDefaults);
        }
        this.elementPolicy = elementPolicy;
        for (InterfaceOperation interfaceOperation : declaredOperations.values()) {
            interfaceOperation.setParent(this);
        }
        if (declaredFaults != null) {
            for (InterfaceFault interfaceFault : declaredFaults.values()) {
                interfaceFault.setParent(this);
            }
        }
    }

    /**
     * Gets name of <code>Interface</code> components extended by this <code>Interface</code>.
     * <p>
     * Note that Interface inheritance is NOT backwards compatible with WSDL 1.1. If
     * inheritance is used to define this component's structure, the component's definition
     * cannot be exported via WSDL 1.1.
     * 
     * @return Set of qualifed names each of which identifies an <code>Interface</code>
     *         extended by this <code>Interface</code>.
     */
    public final Set<QName> getExtendedInterfaceNames() {
        return extendedInterfaceNames;
    }

    /**
     * Gets <code>Interface</code> components extended by this <code>Interface</code>.
     * <p>
     * Note that Interface inheritance is NOT backwards compatible with WSDL 1.1. If
     * inheritance is used to define this component's structure, the component's definition
     * cannot be exported via WSDL 1.1.
     * 
     * @return Set of qualifed names each of which identifies an <code>Interface</code>
     *         extended by this <code>Interface</code>.
     */
    @SuppressWarnings("unchecked")
    public final Set<Interface> getExtendedInterfaces() {
        // TODO: iterate over extended interface names and process ancestors
        return Collections.EMPTY_SET;
    }

    /**
     * Defines default operation style which MAY provide additional constraints on the element
     * declarations used by all operations defined on this interface.
     * <p>
     * The constraints in no way affect the messages and faults exchanged. The style may
     * therefore be safely ignored in this context.
     * <p>
     * Note that this attribute is NOT backwards compatible with WSDL 1.1. If defined, this
     * attribute cannot be described via WSDL 1.1.
     * 
     * @return a Set of style enumerations or an empty Set if style is undefined.
     */
    public Set<String> getStyleDefaults() {
        return styleDefaults;
    }

    /**
     * Gets an <code>InterfaceOperation</code> defined on this Interface. The set of
     * operations examined include operations declared locally as well as those inherited from
     * extended Interface's, if any.
     * 
     * @param operationName
     *        Qualified name of operation.
     * 
     * @return interfaceOperation, or <code>null</code> if undefined.
     */
    public final InterfaceOperation getOperation(QName operationName) {
        return operations.get(operationName);
    }

    /**
     * Gets <code>InterfaceOperation</code> components defined on this <code>Interface</code>
     * as well those inherited from extended interfaces.
     * 
     * @return Collection of operations or an empty collection if none defined.
     */
    public final Collection<InterfaceOperation> getOperations() {
        return operations.values();
    }

    /**
     * Gets <code>InterfaceOperation</code> components directly defined on this
     * <code>Interface</code>.
     * 
     * @return Collection of operations or an empty collection if none defined.
     */
    public final Collection<InterfaceOperation> getDeclaredOperations() {
        return declaredOperations.values();
    }

    /**
     * Retrieves metadata describing a specific <code>InterfaceFault</code> defined within
     * <code>Interface</code> component's set of faults. The set of faults examined include
     * faults defined locally as well as those inherited from extended Interface's.
     * 
     * @param faultName
     *        Qualified name of fault
     * 
     * @return InterfaceFault object, or <code>null</code> if no fault found matching
     *         faultName.
     */
    public final InterfaceFault getFault(QName faultName) {
        return faults.get(faultName);
    }

    /**
     * Gets <code>InterfaceFault</code> components defined on this <code>Interface</code>
     * as well as those inherited from extended interfaces.
     * 
     * @return Collection of <code>InterfaceFault</code> components or an empty collection if
     *         none defined.
     */
    public final Collection<InterfaceFault> getFaults() {
        return faults.values();
    }

    /**
     * Gets <code>InterfaceFault</code> components directly defined on this
     * <code>Interface</code>.
     * 
     * @return Collection of <code>InterfaceFault</code> components or an empty collection if
     *         none defined.
     */
    public final Collection<InterfaceFault> getDeclaredFaults() {
        return declaredFaults.values();
    }

    /**
     * Retrieves <code>InterfaceAction</code> identified by a specific action URI. The
     * actions examined include actions defined locally as well as those inherited from
     * extended Interface components.
     * 
     * @return InterfaceAction or <code>null</code> if undefined.
     */
    public final InterfaceAction getInterfaceAction(String action) {
        return actions.get(action);
    }

    /**
     * Retrieves all <code>InterfaceActions</code>. The returned actions include actions
     * defined locally as well as those inherited from extended Interface components.
     * 
     * @return Collection of <code>InterfaceAction</code> components or an empty collection
     *         if none defined.
     */
    public final Collection<InterfaceAction> getInterfaceActions() {
        return actions.values();
    }

    /**
     * Gets <code>ElementPolicy</code> attached to component.
     * 
     * @return ElementPolicy or <code>null</code> if no policy attached.
     */
    public final ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#getComponentType()
     */
    @Override
    public ComponentType getComponentType() {
        return ComponentType.INTERFACE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#validate()
     */
    @Override
    protected final void validate() throws DeploymentException {
        for (InterfaceFault interfaceFault : declaredFaults.values()) {
            interfaceFault.validate();
        }
        for (InterfaceOperation operation : declaredOperations.values()) {
            operation.validate();
        }
        // TODO: validate myself
    }

    @Override
    protected synchronized void establishReferences() throws DeploymentException {

        if (referencesEstablished) {
            return;
        }

        for (QName ein : extendedInterfaceNames) {
            ComponentDeployment deployment = getComponentContext().getDeployment();
            Interface ei = (Interface)deployment.getComponent(ComponentType.INTERFACE, ein);
            if (ei == null) {
                ei = SystemContext.getContext().getSystem().getInterface(ein);
            }
            if (ei == null) {
                throw new DeploymentException("ComponentReference '"
                        + getFragmentIdentifier()
                        + "' is invalid. Extended interface "
                        + ein
                        + " is undefined.");
            } else {
                if (extendedInterfaces == null) {
                    extendedInterfaces = new HashSet<Interface>();
                }
                extendedInterfaces.add(ei);
            }
        }

        for (InterfaceFault interfaceFault : declaredFaults.values()) {
            interfaceFault.establishReference();
        }

        for (InterfaceOperation operation : declaredOperations.values()) {
            operation.establishReference();
        }

        // iterate over extended interfaces and clone the inherited components and
        // add to our structure. note that we use inherited interface's collection
        // of inherited components such that we inherit from interfaces directly
        // declared as extended as well as those implicitly extended, as required
        // by wsdl 20 spec

        if (extendedInterfaces != null) {

            for (Interface extendedInterface : extendedInterfaces) {
                for (InterfaceFault ifault : extendedInterface.getFaults()) {
                    if (faults.containsKey(ifault.getName())) {
                        InterfaceFault temp = faults.get(ifault.getName());
                        if (!temp.equals(ifault)) {
                            throw new DeploymentException("Inherited fault "
                                    + ifault.getName()
                                    + " clashes with existing fault of same name, but different structure."
                                    + LINE_BREAK
                                    + "Existing:"
                                    + LINE_BREAK
                                    + temp.toString()
                                    + LINE_BREAK
                                    + "Proposed:"
                                    + LINE_BREAK
                                    + ifault.toString()
                                    + LINE_BREAK);
                        }
                    } else {
                        InterfaceFault clone = null;
                        try {
                            clone = (InterfaceFault)ifault.clone();
                        } catch (CloneNotSupportedException fatchance) {
                        }
                        clone.setParent(this);
                        faults.put(clone.getName(), clone);
                    }
                }
                for (InterfaceOperation io : extendedInterface.getOperations()) {
                    if (operations.get(io.getName()) != null) {
                        InterfaceOperation temp = operations.get(io.getName());
                        if (!temp.equals(io)) {
                            throw new DeploymentException("Inherited operation "
                                    + io.getName()
                                    + " clashes with existing fault of same name, but different structure."
                                    + LINE_BREAK
                                    + "Existing:"
                                    + LINE_BREAK
                                    + temp.toString()
                                    + LINE_BREAK
                                    + "Proposed:"
                                    + LINE_BREAK
                                    + io.toString()
                                    + LINE_BREAK);
                        }
                    } else {
                        InterfaceOperation clone = null;
                        try {
                            clone = (InterfaceOperation)io.clone();
                        } catch (CloneNotSupportedException fatchance) {
                        }
                        clone.setParent(this);
                        operations.put(clone.getName(), clone);
                    }
                }
            }

        }

        // populate actions collection. note that we iterate over all operations,
        // including inherited operations

        actions = new HashMap<String, InterfaceAction>();
        for (InterfaceOperation interfaceOperation : operations.values()) {
            InterfaceMessageReference in = interfaceOperation.getInputMessageReference();
            if (in != null) {
                if (actions.containsKey(((InterfaceAction)in).getAction())) {
                    throw new IllegalArgumentException("Action URI '"
                            + ((InterfaceAction)in).getAction()
                            + "' not unique within context of Interface "
                            + name
                            + ".");
                }
                actions.put(((InterfaceAction)in).getAction(), in);
            }
            InterfaceMessageReference out = interfaceOperation.getOutputMessageReference();
            if (out != null) {
                if (actions.containsKey(((InterfaceAction)out).getAction())) {
                    throw new IllegalArgumentException("Action URI '"
                            + ((InterfaceAction)out).getAction()
                            + "' not unique within context of Interface "
                            + name
                            + ".");
                }
                actions.put(((InterfaceAction)out).getAction(), out);
            }
            for (InterfaceFaultReference faultRef : interfaceOperation.getFaultReferences()) {
                if (actions.containsKey(((InterfaceAction)faultRef).getAction())) {
                    throw new IllegalArgumentException("Action URI '"
                            + ((InterfaceAction)faultRef).getAction()
                            + "' not unique within context of Interface "
                            + name
                            + ".");
                }
                actions.put(((InterfaceAction)faultRef).getAction(), faultRef);
            }
            interfaceOperation.setParent(this);
        }

        referencesEstablished = true;

    }

    /*
     * see comments on super (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#modelOperationStyles()
     */
    @Override
    protected void modelOperationStyles() throws DeploymentException {
        for (InterfaceOperation io : getOperations()) {
            io.getInputMessageReference().modelApplicableStyles();
            if (io.getOutputMessageReference() != null) {
                io.getOutputMessageReference().modelApplicableStyles();
            }
        }
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(name);
        sb.append(System.getProperty("line.separator"));
        sb.append("extendedInterfaceNames: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(extendedInterfaceNames);
        sb.append(System.getProperty("line.separator"));
        sb.append("declaredOperations: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(declaredOperations);
        sb.append(System.getProperty("line.separator"));
        sb.append("declaredFaults: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(declaredFaults);
        sb.append(System.getProperty("line.separator"));
        return sb.toString();
    }

    @Override
    public final int hashCode() {
        return hashCode(name)
                + hashCode(extendedInterfaceNames)
                + hashCode(declaredOperations)
                + hashCode(declaredFaults);
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof Interface)) {
            return false;
        }

        if ((name == ((Interface)obj).name || (name != null && name.equals(((Interface)obj).name)))
                && (extendedInterfaceNames == ((Interface)obj).extendedInterfaceNames || (extendedInterfaceNames != null && extendedInterfaceNames
                        .equals(((Interface)obj).extendedInterfaceNames)))
                && (declaredOperations == ((Interface)obj).declaredOperations || (declaredOperations != null && declaredOperations
                        .equals(((Interface)obj).declaredOperations)))
                && (declaredFaults == ((Interface)obj).declaredFaults || (declaredFaults != null && declaredFaults
                        .equals(((Interface)obj).declaredFaults)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }
}