/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.intrface;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;

/**
 * Associates an abstract message type with an interface fault. Interface fault components are
 * declared on an Interface and describe a fault that has 'application level semantics', i.e.
 * that the client is expected to handle within application logic and potentially recover from.
 */

public final class InterfaceFault extends NestedComponent<Interface, Interface> implements ComponentReference<InterfaceMessage>, Cloneable {

    private QName name;
    private ComponentName interfaceMessageName;
    private InterfaceMessage interfaceMessage;
    private ElementPolicy elementPolicy;

    public InterfaceFault(QName name, ComponentName interfaceMessageName, ElementPolicy elementPolicy) {
        super();
        if (name == null) {
            throw new IllegalArgumentException("name null");
        }
        if (interfaceMessageName == null) {
            throw new IllegalArgumentException("interfaceMessageName null");
        }
        this.name = name;
        this.interfaceMessageName = interfaceMessageName;
        this.elementPolicy = elementPolicy;
    }

    /**
     * Retrieves qualified name of fault. Fault name is unique within context of parent
     * interface.
     * <p>
     * Note that despite having a qualified name, interface fault components cannot be
     * identified solely by their QName. If two interfaces defined within the same namespace
     * declare a fault with the same local name, the fault names will clash. Best practice
     * suggests that unless the faults are identical, local fault names should be unique within
     * an interface namespace.
     * 
     * @return faultName.
     */
    public final QName getName() {
        return name;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentName()
     */
    public ComponentName getReferencedComponentName() {
        return interfaceMessageName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentType()
     */
    public ComponentType getReferencedComponentType() {
        return ComponentType.MESSAGE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponent()
     */
    public InterfaceMessage getReferencedComponent() {
        return interfaceMessage;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentClass()
     */
    public Class<InterfaceMessage> getReferencedComponentClass() {
        return InterfaceMessage.class;
    }

    /**
     * Gets <code>ElementPolicy</code> attached to component.
     * 
     * @return ElementPolicy or <code>null</code> if no policy attached.
     */
    public final ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    final void establishReference() throws DeploymentException {
        if (interfaceMessage != null) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        interfaceMessage = getRootComponent().getComponentContext().findReferencedComponent(this);
        if (interfaceMessage == null) {
            throw new DeploymentException("ComponentReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. Referenced message "
                    + interfaceMessageName
                    + " is undefined.");
        }
    }

    final void validate() throws DeploymentException {
        // TODO: implement
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     * 
     * Note that we must account for fact that component may be inherited, i.e. namespace may
     * be different than namespace of parent interface.
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(rootName.getLocalPart());
            pointerPart.append("/");
            pointerPart.append("xmlns(ns=");
            pointerPart.append(name.getNamespaceURI());
            pointerPart.append(")");
            pointerPart.append("ns:" + name.getLocalPart());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "interfaceFault",
                    pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#clone()
     */
    @Override
    public Object clone() throws CloneNotSupportedException {
        InterfaceFault ifault = (InterfaceFault)super.clone();
        if (elementPolicy != null) {
            ifault.elementPolicy = (ElementPolicy)elementPolicy.clone();
        }
        ifault.parent = null;
        return ifault;
    }

    void setParent(Interface parent) {
        this.parent = parent;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(name);
        sb.append(System.getProperty("line.separator"));
        sb.append("interfaceMessageName: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(interfaceMessageName);
        return sb.toString();
    }

    @Override
    public final int hashCode() {
        return hashCode(name) + hashCode(interfaceMessageName);
    }

    @Override
    public final boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof InterfaceFault)) {
            return false;
        }

        if ((name == ((InterfaceFault)obj).name || (name != null && name.equals(((InterfaceFault)obj).name)))
                && (interfaceMessageName == ((InterfaceFault)obj).interfaceMessageName || (interfaceMessageName != null && interfaceMessageName
                        .equals(((InterfaceFault)obj).interfaceMessageName)))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}