/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.message.rt;

import java.io.IOException;
import java.io.InputStream;

import javax.xml.stream.XMLStreamReader;

import org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata.ApplicationData;
import org.w3c.dom.Document;
import org.xml.sax.SAXException;

public interface ContentFactory extends Document {

    /**
     * Imports <code>Content</code> which MAY have been generated by a different factory
     * implementation.
     * <p>
     * Avoids the overhead of possibly having to import the entire tree into memory using
     * {@link Document#importNode(org.w3c.dom.Node, boolean)}, i.e. by allowing implementation
     * to leverage the underlying {@link XMLStreamReader}, if available.
     * 
     * @param source
     *        to be imported. Note that source MUST NOT be used following importation.
     * @return imported content
     * @throws UnsupportedOperationException
     *         if reader not available on source
     * @see Content#getReader()
     */
    public Content importContent(Content source);

    /**
     * Creates a <code>Content</code> instance.
     * 
     * @param namespaceURI
     * @param qualifiedName
     * @return
     */
    public Content createContent(String namespaceURI, String qualifiedName);

    /**
     * Creates <code>ApplicationData</code> instance.
     * 
     * @param namespaceURI
     * @param qualifiedName
     * @return
     */
    public ApplicationData createApplicationData(String namespaceURI, String qualifiedName);

    /**
     * Parses input source into a <code>Content</code> instance. Compliant implementations
     * MUST use the default input serialization, i.e. {@link ContentSerializationXML}, and the
     * default encoding, i.e. UTF-8, when reading the input stream.
     * 
     * @param inputStream
     * @return Content content
     */
    public Content parseContent(InputStream inputStream) throws SAXException, IOException;

    /**
     * Parses <code>String</code> into a <code>Content</code> instance. Compliant implementations
     * MUST use the default input serialization, i.e. {@link ContentSerializationXML}, and the
     * default encoding, i.e. UTF-8, when parsing the <code>String</code>.
     * 
     * @param inputStream
     * @return Content content
     */
    public Content parseContent(String content) throws SAXException, IOException;

    /**
     * Parses input source into a <code>Content</code> instance using the indicated
     * serialization and the default character set encoding UTF-8.
     * 
     * @param inputStream
     * @param serialization
     * @return
     * @throws SAXException
     * @throws IOException
     */
    public Content parseContent(InputStream inputStream, ContentSerialization serialization) throws SAXException, IOException;

    /**
     * Parses input source into a <code>Content</code> instance using the indicated
     * serialization and character set encoding.
     * 
     * @param inputStream
     * @param serialization
     * @param encoding
     *        The character set encoding, e.g. UTF-8.
     * @return
     * @throws SAXException
     * @throws IOException
     */
    public Content parseContent(InputStream inputStream, ContentSerialization serialization, String encoding) throws SAXException, IOException;

    /**
     * Parses a <code>String</code> into a <code>Content</code> instance using the
     * indicated serialization and the default character set encoding UTF-8.
     * 
     * @param content
     * @param serialization
     * @return
     * @throws SAXException
     * @throws IOException
     */
    public Content parseContent(String content, ContentSerialization serialization) throws SAXException, IOException;

    /**
     * Parses <code>String</code> into a <code>Content</code> instance using the indicated
     * serialization and character set encoding.
     * 
     * @param content
     * @param serialization
     * @param encoding
     *        The character set encoding, e.g. UTF-8.
     * @return
     * @throws SAXException
     * @throws IOException
     */
    public Content parseContent(String content, ContentSerialization serialization, String encoding) throws SAXException, IOException;

    
}
