/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.policy.rt;

import java.util.HashSet;
import java.util.Set;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.ComponentReference;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.policy.Policy;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyException;

/**
 * Models 'element' policy as defined by the ws-policy specification. Contains policy
 * expressions attached to an 'element', i.e. a component definition.
 * <p>
 * Note that the means by which policy expressions MAY be attached to a component are language
 * specific and are manifested by type specific factories.
 */
public final class ElementPolicy implements Cloneable {

    private String policyType;
    private ComponentContext componentContext;
    private Set<PolicyExpression> resolvedExpressions;
    private Set<ComponentName> referencedExpressions;
    private PolicyExpression mergedExpression;

    /**
     * Constructs an <code>ElementPolicy</code> object.
     * 
     * @param componentContext
     *        which defines context within which component is deployed, i.e. component to which
     *        element policy was attached. Context is used to resolve referenced expressions,
     *        if any.
     * @param referencedExpressions
     *        names of referenced {@link PolicyExpression} component(s) which will be
     *        established when component to which policy is attached is deployed, or an empty
     *        set if none defined.
     * @param resolvedExpressions
     *        resolvable element policy, i.e. embedded policy elements and remotely defined
     *        expressions or an empty set if none defined. Note that if an expression is
     *        defined remotely, it MUST be resolved.
     */
    public ElementPolicy(ComponentContext componentContext, Set<PolicyExpression> resolvedExpressions,
            Set<ComponentName> referencedExpressions) {
        if (componentContext == null) {
            throw new IllegalArgumentException("componentContext null");
        } else {
            this.componentContext = componentContext;
        }
        if (resolvedExpressions == null) {
            throw new IllegalArgumentException("resolvedExpressions null");
        }
        if (referencedExpressions == null) {
            throw new IllegalArgumentException("referencedExpressions null");
        }
        if (resolvedExpressions.size() == 0 && referencedExpressions.size() == 0) {
            throw new IllegalArgumentException("Both resolvedExpressions AND referencedExpressions are empty.");
        }
        for (PolicyExpression policyExpression : resolvedExpressions) {
            if (policyType == null) {
                policyType = policyExpression.getType();
            } else {
                if (!policyType.equals(policyType)) {
                    throw new IllegalArgumentException("Expressions MUST be identically typed");
                }
            }
        }
        this.resolvedExpressions = resolvedExpressions;
        this.referencedExpressions = referencedExpressions;
    }

    /**
     * Indicates whether or not element policy expressions have been merged into a single
     * expression, i.e. {@link #getMergedExpression()} has been invoked.
     * 
     * @return true if merged, false if not
     */
    public boolean expressionsMerged() {
        return mergedExpression != null;
    }

    /**
     * Merges referenced expressions and resolved expressions into a single normalized
     * expression. Note that normalization forces resolution of policy references, if any.
     * 
     * @return
     */
    public PolicyExpression getMergedExpression() throws PolicyException {

        if (mergedExpression == null) {
            for (PolicyExpression resolvedExpression : resolvedExpressions) {
                if (mergedExpression == null) {
                    mergedExpression = resolvedExpression;
                } else {
                    mergedExpression = mergedExpression.mergeExpression(resolvedExpression);
                }
            }
            for (ComponentName pn : referencedExpressions) {
                Policy policy;
                try {
                    policy = componentContext.findReferencedComponent(new PolicyReferenceInternal(pn));
                } catch (DeploymentException de) {
                    throw new PolicyException(de.getMessage());
                }
                if (policy == null) {
                    throw new PolicyException("Element policy is invalid."
                            + " Referenced policy component "
                            + pn
                            + ", is undefined.");
                }
                PolicyExpression referencedExpression = policy.getRuntimeProvider().getPolicyExpression();
                if (policyType == null) {
                    policyType = referencedExpression.getType();
                } else {
                    if (!policyType.equals(referencedExpression.getType())) {
                        throw new IllegalArgumentException("Expressions MUST be identically typed");
                    }
                }
                if (mergedExpression == null) {
                    mergedExpression = referencedExpression.normalizeExpression();
                } else {
                    mergedExpression = mergedExpression.mergeExpression(referencedExpression);
                }
            }
            if (mergedExpression != null) {
                mergedExpression = mergedExpression.normalizeExpression();
            }
        }

        return mergedExpression;

    }

    /**
     * Combines element policy expressions defined on target object with policy expressions
     * defined on this object and returns reference to self
     * 
     * @param elementPolicy
     *        target attachment.
     * @return reference to self following combine
     * @throws IllegalArgumentException
     *         if element policy expressions merged on either this object or target object.
     */
    public ElementPolicy combine(ElementPolicy target) {
        if (target == null) {
            throw new IllegalArgumentException("elementPolicy null");
        }
        if (this.expressionsMerged()) {
            throw new IllegalArgumentException("expressions merged on this attachment");
        }
        if (target.expressionsMerged()) {
            throw new IllegalArgumentException("expressions merged on target attachment");
        }
        if (!componentContext.equals(target.componentContext)) {
            throw new IllegalArgumentException("Attachments sourced from different deployments");
        }
        this.resolvedExpressions.addAll(target.resolvedExpressions);
        this.referencedExpressions.addAll(target.referencedExpressions);
        return this;
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#clone()
     */
    @Override
    public Object clone() throws CloneNotSupportedException {
        
        if (expressionsMerged()) {
            throw new CloneNotSupportedException("expressions merged");
        }
        
        ElementPolicy clone = (ElementPolicy)super.clone();
        
        // component name is immutable, i.e. no need to clone
        clone.referencedExpressions = new HashSet<ComponentName>();
        clone.referencedExpressions.addAll(referencedExpressions);
        
        // policy expressions are mutable, i.e. we must copy
        clone.resolvedExpressions = new HashSet<PolicyExpression>();
        for(PolicyExpression policyExpression : resolvedExpressions) {
            clone.resolvedExpressions.add(policyExpression.copy());
        }
        
        return clone;
        
    }

    private static class PolicyReferenceInternal implements ComponentReference<Policy> {

        private ComponentName policyName;

        public PolicyReferenceInternal(ComponentName policyName) {
            this.policyName = policyName;
        }

        public FragmentIdentifier getFragmentIdentifier() {
            return new FragmentIdentifier("foo", Scheme.EOA, "bar", "");
        }

        public Policy getReferencedComponent() {
            throw new UnsupportedOperationException();
        }

        public ComponentName getReferencedComponentName() {
            return policyName;
        }

        public ComponentType getReferencedComponentType() {
            return ComponentType.POLICY;
        }

        public Class<Policy> getReferencedComponentClass() {
            return Policy.class;
        }

    }

}
