/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.policy.rt;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference;
import org.bluestemsoftware.specification.eoa.system.server.Feature;
import org.bluestemsoftware.specification.eoa.system.server.ModulableFeature;
import org.w3c.dom.Element;

/**
 * A <code>PolicyAssertion</code> identifies a {@link Feature} capability and/or requirement.
 * Assertions indicate 'domain-specific' semantics which are defined in separate,
 * domain-specific' specifications.
 * <p>
 * Note that <code>PolicyAssertion</code> MAY contain nested policy expression(s). Nested
 * expressions, if any, are not parsed beyond root <code>PolicyAssertion</code>. Nested
 * expressions, if any, MUST therefore be processed according to domain-specific rules, i.e.
 * specific to a particular type of {@link Feature}.
 */
public interface PolicyAssertion extends Element {

    /**
     * Assertions are strongly typed by the domain authors that define them. Type is qualified
     * name of underlying <code>Element</code> information item which represents the
     * assertion.
     * <p>
     * Note that an <code>PolicyAlternative</code> MAY contain more than one
     * <code>PolicyAssertion</code> with the same name. The methodology used to determine
     * aggregate behavior indicated by identically typed assertions is domain-specific, i.e.
     * particular to a specific {@link Feature} type.
     * 
     * @return qualified name of underlying <code>Element</code> information item.
     */
    public QName getType();

    /**
     * Value which MUST be set by an instance of {@link ModulableFeature} when an assertion
     * defined within its policy vocabulary namespace is processed.
     * 
     * @param isSupported
     *        true if entity satisfies requirement implied by assertion, false if not
     */
    public void setIsSupported(boolean isSupported);

    /**
     * Indicates whether or not an instance of {@link ModulableFeature} processed assertion and
     * satisfies the implied requirement.
     * <p>
     * Note that an instance of {@link EndpointActionReference} throws a
     * <code>DeploymentException</code> when an effective policy expression within scope is
     * unsupported, i.e. all assertions in at least one alternative MUST be supported.
     * 
     * @return true if supported, false if not.
     */
    public boolean isSupported();
    
    /**
     * Gets a <code>PolicyExpression</code> which MAY be nested within this assertion
     * instance.
     * 
     * @return PolicyExpression or <code>null</code> if no nested expression found.
     */
    public PolicyExpression getNestedExpression();

}
