/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.rt;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;
import org.bluestemsoftware.specification.eoa.component.RootComponent.RuntimeProvidable;

/**
 * Models a 'runtime' component, i.e. a component which provides a <code>RootComponent's</code>
 * runtime behavior.
 */
public interface RootComponentRT extends ComponentRT {

    /**
     * Returns qualified name of runtime component which MUST match name of
     * <code>RootComponent</code> for which runtime behavior is provided.
     * 
     * @return
     */
    public ComponentName getName();

    /**
     * Returns runtime component type which MUST match type of <code>RootComponent</code> for
     * which runtime behavior is provided.
     * 
     * @return
     */
    public ComponentType getComponentType();

    /**
     * Gets <code>ComponentContext</code> which models context within which runtime component
     * definition is deployed.
     * 
     * @return ComponentContext
     */
    public ComponentContext getComponentContext();
    
    /**
     * Returns <code>RootComponent</code> for which runtime behavior is provided.
     * 
     * @return RootComponent which implements RuntimeProvidable
     */
    public RuntimeProvidable getRuntimeProvidable();

    /**
     * Invoked by <code>System</code> as an indication that runtime components may establish
     * references to other components, if any, and validate composition.
     * 
     * @param componentContext
     * @throws DeploymentException
     */
    public void deploy(ComponentContext componentContext) throws DeploymentException;

    /**
     * Indicates whether <code>RootComponentRT</code> definition has been deployed, i.e.
     * references have been established and component definition has been validated. A
     * definition is considered to be undefined until it has been successfully deployed.
     * 
     * @return true if deployed, false if not.
     */
    public boolean isDeployed();

    /**
     * Marker interface which models a specific type of 'runtime' component, i.e. a component
     * for which exactly one instance of component is guaranteed to exist within context of
     * <code>System</code>.
     * <p>
     * Note that definition for a singleton runtime component MUST be defined within same
     * deployment as definition for corresponding <code>RootComponent</code>.
     */
    public interface SingletonRT<T extends RuntimeProvidable> extends RootComponentRT {
    }

}
