/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.service;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.intrface.Interface;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceFaultReference;
import org.bluestemsoftware.specification.eoa.component.intrface.InterfaceOperation;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.EndpointPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.PolicyExpression;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyException;

/**
 * Models an endpoint component. An endpoint defines a name, binding and executable address to
 * which messages may be sent.
 */

public class Endpoint extends NestedComponent<Service, Service> {

    private String name;
    private BindingReference bindingReference;
    private String address;
    private Map<QName, EndpointOperation> endpointOperations = new HashMap<QName, EndpointOperation>();
    private ElementPolicy elementPolicy;
    private EndpointPolicy effectivePolicy;
    private boolean effectivePolicyDerived;

    public Endpoint(String name, BindingReference bindingReference, String address,
            ElementPolicy elementPolicy) {
        if (name == null) {
            throw new IllegalArgumentException("name null");
        } else {
            this.name = name;
        }
        if (bindingReference == null) {
            throw new IllegalArgumentException("bindingReference null");
        } else {
            this.bindingReference = bindingReference;
            bindingReference.setParent(this);
        }
        if (address == null) {
            throw new IllegalArgumentException("address null");
        } else {
            this.address = address;
        }
        this.elementPolicy = elementPolicy;
    }

    /**
     * Endpoint name, a.k.a port name. Unique within context of a service component.
     * 
     * @return Endpoint name.
     */
    public final String getName() {
        return name;
    }

    /**
     * Returns referenced Binding component.
     * 
     * @return BindingReference
     */
    public final BindingReference getBindingReference() {
        return bindingReference;
    }

    /**
     * The network address used to invoke endpoint.
     * 
     * @return An address URI formatted as a String.
     */
    public final String getAddress() {
        return address;
    }

    /**
     * Gets object which describes an operation defined on this Endpoint
     * 
     * @param qualified
     *        name of operation
     * @return EndpointOperation
     */
    public final EndpointOperation getEndpointOperation(QName operationName) {
        return endpointOperations.get(operationName);
    }

    /**
     * Gets all operations defined on this Endpoint
     * 
     * @param qualified
     *        name of operation
     * @return Collection of endpoint operation objects
     */
    public final Collection<EndpointOperation> getEndpointOperations() {
        return endpointOperations.values();
    }

    /**
     * Gets <code>ElementPolicy</code> attached to component.
     * 
     * @return ElementPolicy or <code>null</code> if no policy attached.
     */
    public final ElementPolicy getElementPolicy() {
        return elementPolicy;
    }

    /**
     * Returns the 'effective' policy. Returns the merged result of element Policy attached to
     * this component, element policy attached to Binding utilized by this Endpoint and element
     * policy attached to Interface component implemented by parent Service component.
     * 
     * @return EffectivePolicy effective policy with policy subject ENDPOINT or
     *         <code>null</code> if no policy defined.
     */
    public final synchronized EndpointPolicy getEffectivePolicy() throws PolicyException {
        if (effectivePolicyDerived) {
            return effectivePolicy;
        }
        effectivePolicy = deriveEffectivePolicy();
        effectivePolicyDerived = true;
        return effectivePolicy;
    }

    /**
     * Instantiates EndpointOperation components, i.e. one for each operation defined on
     * Interface implemented by parent Service. Invoked after references are established.
     */
    void generateImpliedStructure() {
        Interface intrface = parent.getInterfaceReference().getReferencedComponent();
        for (InterfaceOperation io : intrface.getDeclaredOperations()) {
            Map<String, EndpointAction> endpointActions = new HashMap<String, EndpointAction>();
            EndpointOperation endpointOperation = new EndpointOperation(io.getName(), endpointActions);
            if (io.getInputMessageReference() != null) {
                EndpointAction endpointAction = new EndpointAction(io.getInputMessageReference().getAction());
                endpointActions.put(endpointAction.getAction(), endpointAction);
                endpointAction.setParent(endpointOperation);
            }
            if (io.getOutputMessageReference() != null) {
                EndpointAction endpointAction = new EndpointAction(io.getOutputMessageReference().getAction());
                endpointActions.put(endpointAction.getAction(), endpointAction);
                endpointAction.setParent(endpointOperation);
            }
            for (InterfaceFaultReference ifr : io.getFaultReferences()) {
                EndpointAction endpointAction = new EndpointAction(ifr.getAction());
                endpointActions.put(endpointAction.getAction(), endpointAction);
                endpointAction.setParent(endpointOperation);
            }
            endpointOperations.put(io.getName(), endpointOperation);
            endpointOperation.setParent(this);
        }
    }

    private EndpointPolicy deriveEffectivePolicy() throws PolicyException {
        PolicyExpression mergedExpression = null;
        if (elementPolicy != null) {
            PolicyExpression mlp = elementPolicy.getMergedExpression();
            mergedExpression = (mergedExpression == null) ? mlp : mergedExpression.mergeExpression(mlp);
        }
        Binding binding = bindingReference.getReferencedComponent();
        if (binding.getElementPolicy() != null) {
            PolicyExpression blp = binding.getElementPolicy().getMergedExpression();
            mergedExpression = (mergedExpression == null) ? blp : mergedExpression.mergeExpression(blp);
        }
        Interface intrface = parent.getInterfaceReference().getReferencedComponent();
        if (intrface.getElementPolicy() != null) {
            PolicyExpression ilp = intrface.getElementPolicy().getMergedExpression();
            mergedExpression = (mergedExpression == null) ? ilp : mergedExpression.mergeExpression(ilp);
        }
        if (mergedExpression != null) {
            return new EndpointPolicy(mergedExpression.normalizeExpression());
        } else {
            return null;
        }
    }

    void validate() throws DeploymentException {
        for (EndpointOperation endpointOperation : endpointOperations.values()) {
            endpointOperation.validate();
        }
        // verify that if binding defined at interface granularity, it binds to 
        // same interface implemented by parent service
        Binding binding = bindingReference.getReferencedComponent();
        if (!binding.isReusable()) {
            QName boundInterface = binding.getInterfaceReference().getReferencedComponentName();
            QName implementedInterface = parent.getInterfaceReference().getReferencedComponentName();
            if (!boundInterface.equals(implementedInterface)) {
                throw new DeploymentException("Endpoint "
                        + name
                        + " defined on service "
                        + parent.getName()
                        + " is invalid. Referenced binding "
                        + binding.getName()
                        + " binds interface "
                        + boundInterface
                        + " whereas parent service implements "
                        + implementedInterface
                        + ".");
            }
        }
        // TODO: validate myself
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            pointerPart.append("/");
            pointerPart.append(name);
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "endpoint",
                    pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(Service parent) {
        this.parent = parent;
    }

    void establishReference() throws DeploymentException {
        bindingReference.establishReference();
        generateImpliedStructure();
        for (EndpointOperation endpointOperation : endpointOperations.values()) {
            endpointOperation.establishReference();
        }
    }

    // *********************** Identity Methods ************************************

    @Override
    public final String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(name);
        sb.append(System.getProperty("line.separator"));
        sb.append("bindingReference: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(bindingReference);
        sb.append(System.getProperty("line.separator"));
        sb.append("address: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(address);
        sb.append(System.getProperty("line.separator"));
        sb.append("endpointOperations: ");
        sb.append(System.getProperty("line.separator"));
        sb.append(endpointOperations);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(name) + hashCode(bindingReference) + hashCode(address) + hashCode(endpointOperations);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof Endpoint)) {
            return false;
        }

        if ((name == ((Endpoint)obj).name || (name != null && name.equals(((Endpoint)obj).name)))
                && (bindingReference == ((Endpoint)obj).bindingReference || (bindingReference != null && bindingReference
                        .equals(((Endpoint)obj).bindingReference)))
                && (address == ((Endpoint)obj).address || (address != null && address
                        .equals(((Endpoint)obj).address)))
                && (endpointOperations == ((Endpoint)obj).endpointOperations || (endpointOperations != null && endpointOperations
                        .equals(((Endpoint)obj).endpointOperations)))

        ) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}