/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.types;

import java.util.Collections;
import java.util.Map;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.types.rt.ExpressionRT;

public final class FragmentExpression extends NestedComponent<Schema, SchemaFragment> {

    private String query;
    private Map<String, String> namespaces;
    private ExpressionRT runtimeExpression;

    public FragmentExpression(String query, Map<String, String> namespaces, ExpressionRT runtimeExpression) {
        if (query == null) {
            throw new IllegalArgumentException("query null");
        } else {
            this.query = query;
        }
        if (namespaces == null) {
            throw new IllegalArgumentException("namespaces null");
        } else {
            this.namespaces = Collections.unmodifiableMap(namespaces);
        }
        if (runtimeExpression == null) {
            throw new IllegalArgumentException("runtimeExpression null");
        } else {
            this.runtimeExpression = runtimeExpression;
        }
    }

    /**
     * Retrieves query used to extract an instance of referenced Property from runtime instance
     * of parent schema component.
     * <p>
     * If parent component is typed as an ELEMENT_DECLARATION, the query may be absolute or it
     * may be relative. The context <code>Node</code> from which query is evaluated MUST be a
     * synthetic Document Information Item which contains a single child Element Information
     * Item, i.e. an instance of parent element decl. As an example, to extract an embedded
     * instance of Property 'baz' from an instance of parent element decl 'foo' using XPath 1.0
     * as query language:
     * <p>
     * &ltfoo&gt&ltbar&gt&ltbaz/&gt&lt/bar&gt&lt/foo&gt
     * <p>
     * Requires absolute path: /foo/bar/baz Or relative path: ./bar/baz
     * <p>
     * If parent component is typed as a complex TYPE_DEFINITION, the query MUST be relative.
     * The context <code>Node</code> from which query is evaluated MUST be a synthetic
     * Document Information Item which contains a single child Element Information Item, i.e.
     * an anonymous element which contains an instance of parent type. As an example, assuming
     * parent type defines a sequence of two 'bar' elements, to extract the second embedded
     * instance of Property 'bar' from an instance of parent type:
     * <p>
     * &lt?&gt&ltbar/&gt&ltbar/&gt&lt/?&gt
     * <p>
     * Requires relative path: ./bar[2]
     * <p>
     * TODO: what's a use case for definining a fragment on a schema component typed as a
     * simple type and is this something that should be supported?
     * 
     * @return String a query which MUST conform to language identified by value retrieved via
     *         {@link #getQueryLanguage()}.
     */
    public String getQuery() {
        return query;
    }

    /**
     * Returns a Collection of namespace declarations, i.e. each entry maps a namespace prefix
     * to a namespace URI. The prefix serves as the key and is therefore unique within the
     * context of this expression.
     * <p>
     * Declared prefixes are used within the fragment query. Note that declared prefixes are in
     * no way related to prefixes which MAY occur within runtime instances of parent schema
     * component, e.g. an element information item.
     * 
     * @return Map Containing namespace declarations or an empty map if none required\defined.
     */
    public Map<String, String> getNamespaces() {
        return namespaces;
    }
    
    /**
     * Gets runtime manifestation of this <code>FragmentExpression</code>.
     * 
     * @return ExpressionRT
     */
    public ExpressionRT getRuntimeExpression() {
        return runtimeExpression;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.Component#getFragmentIdentifier()
     */
    public final FragmentIdentifier getFragmentIdentifier() {
        if (fragmentIdentifier == null) {
            QName rootName = getRootComponent().getName();
            StringBuffer pointerPart = new StringBuffer();
            pointerPart.append(parent.getFragmentIdentifier().getPointerPart());
            fragmentIdentifier = new FragmentIdentifier(rootName.getNamespaceURI(), Scheme.EOA, "fragmentExpression",
                    pointerPart.toString());
        }
        return fragmentIdentifier;
    }

    void setParent(SchemaFragment parent) {
        this.parent = parent;
    }

    void validate() throws DeploymentException {
    }

}
