/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.types;

import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.RootComponent.RuntimeProvidable;
import org.bluestemsoftware.specification.eoa.component.types.SchemaComponent.Type;
import org.bluestemsoftware.specification.eoa.component.types.rt.SchemaDocument;

/**
 * A deployable, 'top-level' component which wraps a runtime <code>Schema</code> instance,
 * i.e. a <code>SchemaDocument</code>.
 */
public final class Schema extends RootComponent implements RuntimeProvidable {

    private Map<Type, Map<String, SchemaComponent>> schemaComponents;
    private Map<String, SchemaProperty> properties;
    private Map<String, Set<SchemaImport>> schemaImports;
    private Set<SchemaInclude> schemaIncludes;

    /**
     * Constructs a <code>Schema</code> component. Note that runtime, i.e.
     * <code>SchemaDocument</code> is set by <code>System</code> when component is
     * deployed.
     * 
     * @param name
     *        Qualified name of <code>Schema</code>. By convention, namespace URI MUST be
     *        targetNamespace of <code>Schema</code> and local part MUST be value of id
     *        attribute defined on root element of schema document from which schema was
     *        parsed.
     * @param schemaComponents
     * @param properties
     * @param schemaImports
     * @param schemaIncludes
     */
    public Schema(ComponentName name, Map<Type, Map<String, SchemaComponent>> schemaComponents,
            Map<String, SchemaProperty> properties, Map<String, Set<SchemaImport>> schemaImports,
            Set<SchemaInclude> schemaIncludes) {
        if (name == null) {
            throw new IllegalArgumentException("name null");
        } else {
            this.name = name;
        }
        if (schemaComponents == null) {
            throw new IllegalArgumentException("schemaComponents null");
        } else {
            this.schemaComponents = Collections.unmodifiableMap(schemaComponents);
        }
        if (properties == null) {
            throw new IllegalArgumentException("properties null");
        } else {
            this.properties = Collections.unmodifiableMap(properties);
        }
        if (schemaImports == null) {
            throw new IllegalArgumentException("schemaImports null");
        } else {
            this.schemaImports = Collections.unmodifiableMap(schemaImports);
        }
        if (schemaIncludes == null) {
            throw new IllegalArgumentException("schemaIncludes null");
        } else {
            this.schemaIncludes = Collections.unmodifiableSet(schemaIncludes);
        }
        fragmentIdentifier = new FragmentIdentifier(name.getNamespaceURI(), Scheme.EOA, "schema", name
                .getLocalPart());
        for (SchemaComponent schemaComponent : getSchemaComponents()) {
            schemaComponent.setParent(this);
        }
        for (SchemaProperty property : getProperties()) {
            property.setParent(this);
        }
        for (Set<SchemaImport> set : this.schemaImports.values()) {
            for (SchemaImport schemaImport : set) {
                schemaImport.setParent(this);
            }
        }
        for (SchemaInclude schemaInclude : this.schemaIncludes) {
            schemaInclude.setParent(this);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent.RuntimeProvidable#getRuntimeProvider()
     */
    public SchemaDocument getRuntimeProvider() {
        if (runtimeProvider == null || !runtimeProvider.isDeployed()) {
            throw new IllegalStateException("Provider definition not yet deployed.");
        }
        return (SchemaDocument)runtimeProvider;
    }

    /**
     * Gets a <code>SchemaComponent</code> nested within this <code>Schema</code>.
     * Components nested within imported/included schema MUST be excluded.
     * 
     * @param componentType
     *        enumerated value
     * @param name
     *        name of component
     * @return schemaComponent or <code>null</code> if undefined
     */
    public SchemaComponent getSchemaComponent(Type componentType, String name) {
        Map<String, SchemaComponent> componentMap = schemaComponents.get(componentType);
        if (componentMap == null) {
            return null;
        }
        return componentMap.get(name);
    }

    /**
     * Gets all <code>SchemaComponents</code> nested within this <code>Schema</code>.
     * Components nested within imported/included schema MUST be excluded.
     * 
     * @return Set of <code>SchemaComponents</code>, or an empty Set if none defined.
     */
    public Set<SchemaComponent> getSchemaComponents() {
        Set<SchemaComponent> temp = new HashSet<SchemaComponent>();
        if (schemaComponents.get(Type.ELEMENT_DECLARATION) != null) {
            temp.addAll(schemaComponents.get(Type.ELEMENT_DECLARATION).values());
        }
        if (schemaComponents.get(Type.TYPE_DEFINITION) != null) {
            temp.addAll(schemaComponents.get(Type.TYPE_DEFINITION).values());
        }
        return temp;
    }

    /**
     * Gets a <code>Property</code> component nested within this <code>Schema</code>.
     * Components nested within imported/included schema MUST be excluded.
     * 
     * @param name
     *        name of property
     * @return property or <code>null</code> if undefined
     */
    public SchemaProperty getProperty(String name) {
        return properties.get(name);
    }

    /**
     * Gets all <code>Property</code> components nested within this <code>Schema</code>.
     * Components nested within imported/included schema MUST be excluded.
     * 
     * @return Collection of <code>Property</code> components, or an empty collection if none
     *         defined.
     */
    public Collection<SchemaProperty> getProperties() {
        return properties.values();
    }

    /**
     * Gets all <code>SchemaImport</code> components nested within this <code>Schema</code>
     * which import indicated namespace.
     * 
     * @return collection of <code>SchemaImport</code> components, or an empty collection if
     *         none defined matching indicated namespace.
     */
    @SuppressWarnings("unchecked")
    public Set<SchemaImport> getSchemaImports(String namespace) {
        Set<SchemaImport> set = schemaImports.get(namespace);
        if (set == null) {
            return Collections.EMPTY_SET;
        }
        return set;
    }

    /**
     * Gets all <code>SchemaInclude</code> components nested within this <code>Schema</code>
     * 
     * @return collection of <code>SchemaInclude</code> components, or an empty collection if
     *         none defined.
     */
    public Set<SchemaInclude> getSchemaIncludes() {
        return schemaIncludes;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#getComponentType()
     */
    @Override
    public ComponentType getComponentType() {
        return ComponentType.SCHEMA;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#establishReferences()
     */
    @Override
    protected synchronized void establishReferences() throws DeploymentException {
        for (Set<SchemaImport> set : this.schemaImports.values()) {
            for (SchemaImport schemaImport : set) {
                schemaImport.establishReference();
            }
        }
        for (SchemaInclude schemaInclude : getSchemaIncludes()) {
            schemaInclude.establishReference();
        }
        for (SchemaComponent schemaComponent : getSchemaComponents()) {
            Iterator<?> itr = schemaComponent.getFragments().iterator();
            while(itr.hasNext()) {
                SchemaFragment fragment = (SchemaFragment)itr.next();
                fragment.establishReference();
            }
        }
        for (SchemaProperty property : getProperties()) {
            property.establishReference();
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.RootComponent#validate()
     */
    @Override
    protected final void validate() throws DeploymentException {
        for (Set<SchemaImport> set : this.schemaImports.values()) {
            for (SchemaImport schemaImport : set) {
                schemaImport.validate();
            }
        }
        for (SchemaInclude schemaInclude : getSchemaIncludes()) {
            schemaInclude.validate();
        }
        for (SchemaComponent schemaComponent : getSchemaComponents()) {
            Iterator<?> itr = schemaComponent.getFragments().iterator();
            while(itr.hasNext()) {
                SchemaFragment fragment = (SchemaFragment)itr.next();
                fragment.validate();
            }
        }
        for (SchemaProperty property : getProperties()) {
            property.validate();
        }
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("name: ");
        sb.append(name);
        sb.append(System.getProperty("line.separator"));
        sb.append("schemaImports:");
        sb.append(System.getProperty("line.separator"));
        sb.append(schemaImports);
        sb.append(System.getProperty("line.separator"));
        sb.append("schemaIncludes:");
        sb.append(System.getProperty("line.separator"));
        sb.append(schemaIncludes);
        sb.append(System.getProperty("line.separator"));
        sb.append("schemaComponents:");
        sb.append(System.getProperty("line.separator"));
        sb.append(schemaComponents);
        sb.append(System.getProperty("line.separator"));
        sb.append("properties:");
        sb.append(System.getProperty("line.separator"));
        sb.append(properties);
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(name)
                + hashCode(schemaImports)
                + hashCode(schemaIncludes)
                + hashCode(schemaComponents)
                + hashCode(properties);
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof Schema)) {
            return false;
        }

        if ((name == ((Schema)obj).name || (name != null && name.equals(((Schema)obj).name)))
                && (schemaIncludes == ((Schema)obj).schemaIncludes || (schemaIncludes != null && schemaIncludes
                        .equals(((Schema)obj).schemaIncludes)))
                && (schemaImports == ((Schema)obj).schemaImports || (schemaImports != null && schemaImports
                        .equals(((Schema)obj).schemaImports)))
                && (schemaComponents == ((Schema)obj).schemaComponents || (schemaComponents != null && schemaComponents
                        .equals(((Schema)obj).schemaComponents)))
                && (properties == ((Schema)obj).properties || (properties != null && properties
                        .equals(((Schema)obj).properties)))

        ) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}
