/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.component.types;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentReference;
import org.bluestemsoftware.specification.eoa.component.NestedComponent;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentName;
import org.bluestemsoftware.specification.eoa.component.RootComponent.ComponentType;

/**
 * Abstract ancestor of <code>SchemaImport</code> and <code>SchemaInclude</code> each of
 * which references a <code>Schema</code> component.
 */
public abstract class SchemaReference extends NestedComponent<Schema, Schema> implements ComponentReference<Schema> {

    protected ComponentName schemaName;
    protected Schema schema;
    protected String location;

    public SchemaReference(ComponentName schemaName, String location) {
        if (schemaName == null) {
            throw new IllegalArgumentException("schemaName null");
        } else {
            this.schemaName = schemaName;
        }
        if (location == null) {
            throw new IllegalArgumentException("location null");
        } else {
            this.location = location;
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentName()
     */
    public ComponentName getReferencedComponentName() {
        return schemaName;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentType()
     */
    public ComponentType getReferencedComponentType() {
        return ComponentType.SCHEMA;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponentClass()
     */
    public Class<Schema> getReferencedComponentClass() {
        return Schema.class;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#getReferencedComponent()
     */
    /**
     * Gets location of referenced schema, i.e. value of schemaLocation attribute.
     * 
     * @return String or empty string if undefined
     */
    public String getSchemaLocation() {
        return location;
    }

    /**
     * Gets referenced <code>Schema</code>. Note that referenced schema MUST be same type,
     * i.e. language as parent schema.
     * 
     * @return Schema or <code>null</code> if reference not yet established.
     */
    public Schema getReferencedComponent() {
        return schema;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.ComponentReference#establishReference()
     */
    final void establishReference() throws DeploymentException {
        if (schema != null) {
            return;
        }
        if (parent == null) {
            throw new IllegalStateException("parent null");
        }
        Schema mySchema = getRootComponent();
        if (mySchema.getName().equals(schemaName)) {
            throw new DeploymentException("Illegal reference. ComponentReference "
                    + getFragmentIdentifier()
                    + " references parent Schema.");
        }
        schema = getRootComponent().getComponentContext().findReferencedComponent(this);
        if (schema == null) {
            throw new DeploymentException("ComponentReference '"
                    + getFragmentIdentifier()
                    + "' is invalid. Referenced Schema '"
                    + schemaName
                    + "' is undefined.");

        }
    }

    void setParent(Schema parent) {
        this.parent = parent;
    }

}
