/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext;


/**
 * A type specific factory for creating {@link Extension} objects. 
 */

public abstract class ExtensionFactory {

    protected ExtensionFactoryContext factoryContext;
    protected Provider provider;

    /**
     * Enumerates <code>ExtensionFactory</code> types.
     */
    public static enum Type {
        CONTAINER,
        SERVER,
        FEATURE,
        SCHEMA,
        EXPRESSION,
        CONNECTOR,
        ENGINE,
        APPLICATION,
        MESSAGE,
        BINDING,
        POLICY,
        IDL,
        MANAGEMENT
    }

    /**
     * Implemented by an <code>ExtensionFactory</code> service provider.
     * 
     * @see java.util.ServiceLoader
     */
    public interface Provider {

        /**
         * Gets 'short' name suitable for logging purposes, etc ...
         * 
         * @return A short identifier
         */
        public String spi_getName();

    }

    /**
     * Returns enum which identifies factory type.
     * 
     * @return Type
     */
    public abstract Type getType();

    /**
     * Returns reference URI which identifies type of <code>Extension</code> created by
     * factory.
     * 
     * @return ref URI which MUST be unique within context of {@link Type}.
     */
    public abstract String getExtensionType();

    /**
     * Gets 'short' name suitable for logging purposes, etc ...
     * 
     * @return A short identifier
     */
    public String getName() {
        if (factoryContext == null) {
            throw new IllegalStateException("Factory has no context.");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factoryContext.getClassLoader());
            return provider.spi_getName();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Sets <code>ExtensionFactory</code> provider. Set by <code>System</code> when factory
     * definition is deployed.
     * 
     * @param provider
     * @return this
     */
    void setProvider(Provider provider) {
        if (provider == null) {
            throw new IllegalArgumentException("provider null.");
        } else {
            this.provider = provider;
        }
    }

    /**
     * Gets underlying <code>ExtensionFactory</code> service provider.
     * 
     * @return service provider
     */
    public Provider getProvider() {
        return provider;
    }

    /**
     * Invoked when <code>ExtensionFactory</code> definition is deployed by <code>System</code>.
     * 
     * @param factoryContext
     * @throws IllegalStateException
     *         If provider not set.
     */
    void deploy(ExtensionFactoryContext factoryContext) {
        if (provider == null) {
            throw new IllegalStateException("Factory has no provider.");
        } else {
            this.factoryContext = factoryContext;
        }
    }

    /**
     * Gets <code>DeploymentContext</code> set when factory was deployed by system.
     * 
     * @return DeploymentContext
     */
    public ExtensionFactoryContext getFactoryContext() {
        return factoryContext;
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("type: ");
        sb.append(getType());
        sb.append(System.getProperty("line.separator"));
        sb.append("ref: ");
        sb.append(getExtensionType());
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(getType()) + hashCode(getExtensionType());
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof ExtensionFactory)) {
            return false;
        }

        if ((getType() == ((ExtensionFactory)obj).getType() || (getType() != null && getType().equals(
                ((ExtensionFactory)obj).getType())))
                && (getExtensionType() == ((ExtensionFactory)obj).getExtensionType() || (getExtensionType() != null && getExtensionType()
                        .equals(((ExtensionFactory)obj).getExtensionType())))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}