/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.lang.reflect.Modifier;
import java.util.Date;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.Dependency;
import org.bluestemsoftware.specification.eoa.Deployment;
import org.bluestemsoftware.specification.eoa.DeploymentClassLoader;
import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

/**
 * Models an extension factory deployment archive which contains exactly one {@link ExtensionFactory}
 * definition.
 */

public class ExtensionFactoryDeployment extends Deployment {

    private String ref;
    protected ExtensionFactory extensionFactory;

    private static final QName PROVIDER_ELEMENT = new QName(DESCRIPTOR_NS, "provider");
    private static final QName FACTORY_ELEMENT = new QName(DESCRIPTOR_NS, "factory");

    /**
     * Constructor.
     * 
     * @param organizationID
     * @param artifactID
     * @param version
     * @param extension
     * @param file
     * @param deploymentDependencies
     * @param modules
     *        each of which is mapped to module extension ref, or an empty map if none defined
     */
    public ExtensionFactoryDeployment(String organizationID, String artifactID, String version, String extension,
            File file, ExtensionFactoryDependencies deploymentDependencies) {
        super(organizationID, artifactID, version, extension, file, deploymentDependencies);
    }

    /**
     * Returns <code>ExtensionFactory</code> sourced from this deployment.
     * 
     * @return factory
     */
    public ExtensionFactory getExtensionFactory() {
        return extensionFactory;
    }

    /**
     * Returns URI which uniquely identifes this factory deployment within context of system.
     * URI is formatted as:
     * <p>
     * organizationID/artifactID/extension
     * <p>
     * Note that this URI matches the URI used to identify this deployment as a dependency,
     * i.e. {@link Dependency#getVersionlessRef()} (only one version of a factory deployment
     * can be deployed within context of system.)
     */
    @Override
    public String getRef() {
        if (ref == null) {
            ref = organizationID + "/" + artifactID + "/" + extension;
        }
        return ref;
    }

    /**
     * Invoked by <code>System</code> when factory is deployed.
     */
    public void deploy(ExtensionFactoryContext factoryContext) throws DeploymentException {

        if (deployed != null) {
            return;
        }

        if (factoryContext == null) {
            throw new IllegalArgumentException("factoryContext null");
        }

        Document descriptorDocument = null;
        FileInputStream fis = null;
        try {
            File descriptor = new File(file, DESCRIPTOR_PATH);
            if (!descriptor.exists()) {
                throw new DeploymentException("Error loading deployment "
                        + getRef()
                        + ". Deployment missing required descriptor '"
                        + DESCRIPTOR_PATH
                        + "'.");
            }
            fis = new FileInputStream(descriptor);
            InputSource inputSource = new InputSource(fis);
            inputSource.setSystemId(descriptor.getAbsolutePath());
            descriptorDocument = new DescriptorParser().parse(inputSource);
        } catch (Exception ex) {
            throw new DeploymentException("Error deploying extension factory deployment " + getRef() + ". " + ex);
        } finally {
            if (fis != null) {
                try {
                    fis.close();
                } catch (IOException ie) {
                    throw new DeploymentException("Error deploying extension factory deployment " + getRef() + ". " + ie);
                }
            }
        }

        String className = null;
        Element root = descriptorDocument.getDocumentElement();
        Element providerElement = getChildElement(root, PROVIDER_ELEMENT);
        if (providerElement == null) {
            throw new DeploymentException("Error deploying extension factory deployment "
                    + getRef()
                    + ". Descriptor missing required 'provider' element.");
        } else {
            Element factoryElement = getChildElement(providerElement, FACTORY_ELEMENT);
            className = getText(factoryElement);
        }

        Class<?> serviceProviderClass = null;
        try {
            DeploymentClassLoader dcl = factoryContext.getClassLoader();
            serviceProviderClass = dcl.loadClass(className);
        } catch (ClassNotFoundException ce) {
            throw new DeploymentException("Error deploying extension factory deployment "
                    + getRef()
                    + ". Error loading class "
                    + className
                    + ". "
                    + ce);
        }

        setExtensionFactory(serviceProviderClass);

        ExtensionFactory.Provider efp = null;
        try {
            efp = (ExtensionFactory.Provider)serviceProviderClass.newInstance();
        } catch (Exception ex) {
            throw new DeploymentException("Error deploying extension factory deployment "
                    + getRef()
                    + ". Error instantiating ExtensionFactory.Provider. "
                    + ex);
        }

        extensionFactory.setProvider(efp);
        extensionFactory.deploy(factoryContext);
        deployed = new Date();

    }

    /**
     * Protected method available to subclasses which provide a 'lightweight' deployment
     * fixture used for testing, i.e. where provider is instantiated directly rather than
     * discovered using service provider configuration file. Note that ExtensionFactory
     * instance MUST be set via {@link #setExtensionFactoryClass(Class)}.
     * 
     * @param factoryContext
     * @param provider
     */
    protected void deploy(ExtensionFactoryContext factoryContext, ExtensionFactory.Provider provider) throws DeploymentException {
        if (deployed != null) {
            return;
        }
        if (factoryContext == null) {
            throw new IllegalArgumentException("factoryContext null");
        }
        if (provider == null) {
            throw new IllegalArgumentException("provider not set");
        }
        if (extensionFactory == null) {
            throw new IllegalArgumentException("extensionFactory null");
        }
        extensionFactory.setProvider(provider);
        extensionFactory.deploy(factoryContext);
        deployed = new Date();
    }

    /**
     * Using reflection, identifies ExtensionFactory class, which MUST be enclosing class
     * within which service provider interface is defined. Discovered class is instantiated and
     * set on instance variable {@link #extensionFactory}.
     * 
     * @param serviceProviderClass
     * @return
     * @throws DeploymentException
     */
    protected void setExtensionFactory(Class<?> serviceProviderClass) throws DeploymentException {

        Class<?> serviceProviderInterface = null;
        for (Class<?> intrface : serviceProviderClass.getInterfaces()) {
            if (ExtensionFactory.Provider.class.isAssignableFrom(intrface)) {
                serviceProviderInterface = intrface;
                break;
            }
        }

        if (serviceProviderInterface == null) {
            throw new DeploymentException("Error deploying extension factory deployment "
                    + getRef()
                    + ". ExtensionFactory.Provider "
                    + serviceProviderClass.getName()
                    + " MUST implement an interface which is assignable from "
                    + ExtensionFactory.Provider.class
                    + ".");
        }

        Class<?> factoryClass = serviceProviderInterface.getDeclaringClass();

        if (factoryClass == null) {
            throw new DeploymentException("Error deploying extension factory deployment "
                    + getRef()
                    + ". Provider interface "
                    + serviceProviderInterface.getName()
                    + " implemented by provider class "
                    + serviceProviderClass.getName()
                    + " MUST, by convention, be declared within in an enclosing class which is assignable from "
                    + ExtensionFactory.class.getName()
                    + ". Enclosing class is null.");
        }

        if (factoryClass == null || !ExtensionFactory.class.isAssignableFrom(factoryClass)) {
            throw new DeploymentException("Error deploying extension factory deployment "
                    + getRef()
                    + ". Provider interface "
                    + serviceProviderInterface.getName()
                    + " implemented by provider class "
                    + serviceProviderClass.getName()
                    + " MUST, by convention, be declared within in an enclosing class which is assignable from "
                    + ExtensionFactory.class.getName()
                    + ". Enclosing class "
                    + factoryClass
                    + " is not assignable.");
        }

        int mods = factoryClass.getModifiers();
        if (Modifier.isAbstract(mods)) {
            throw new DeploymentException("Error deploying extension factory deployment "
                    + getRef()
                    + ". Provider interface "
                    + serviceProviderInterface.getName()
                    + " implemented by provider class "
                    + serviceProviderClass.getName()
                    + " MUST, by convention, be declared within a concrete instanceof "
                    + ExtensionFactory.class.getName()
                    + ". Enclosing class "
                    + factoryClass.getName()
                    + " is abstract.");
        }

        try {
            extensionFactory = (ExtensionFactory)factoryClass.newInstance();
        } catch (Exception ex) {
            throw new DeploymentException("Error deploying extension factory deployment "
                    + getRef()
                    + ". Error instantiating ExtensionFactory consumer "
                    + factoryClass.getName()
                    + ". "
                    + ex);
        }

    }
    
}
