/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.bluestemsoftware.specification.eoa.Deployment;
import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.DeploymentManager;
import org.bluestemsoftware.specification.eoa.SharedClassLoader;
import org.bluestemsoftware.specification.eoa.component.rt.ProviderReader;
import org.bluestemsoftware.specification.eoa.ext.application.ApplicationFactory;
import org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory;
import org.bluestemsoftware.specification.eoa.ext.connector.ConnectorFactory;
import org.bluestemsoftware.specification.eoa.ext.container.ContainerFactory;
import org.bluestemsoftware.specification.eoa.ext.engine.EngineFactory;
import org.bluestemsoftware.specification.eoa.ext.expression.ExpressionFactory;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureFactory;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLFactory;
import org.bluestemsoftware.specification.eoa.ext.management.ManagementContextFactory;
import org.bluestemsoftware.specification.eoa.ext.message.MessageFactory;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyFactory;
import org.bluestemsoftware.specification.eoa.ext.schema.SchemaFactory;
import org.bluestemsoftware.specification.eoa.ext.server.ServerFactory;

/**
 * Manages <code>ExtensionFactory</code> deployments on behalf of <code>System</code>.
 */
public class ExtensionFactoryManager extends DeploymentManager {

    protected Map<String, ExtensionFactoryDeployment> providerReaders;
    protected Map<String, ExtensionFactoryDeployment> extensionFactories;
    protected Map<String, ExtensionFactoryDeployment> applicationFactories;
    protected Map<String, ExtensionFactoryDeployment> containerFactories;
    protected Map<String, ExtensionFactoryDeployment> expressionFactories;
    protected Map<String, ExtensionFactoryDeployment> serverFactories;
    protected Map<String, ExtensionFactoryDeployment> engineFactories;
    protected Map<String, ExtensionFactoryDeployment> bindingFactories;
    protected Map<String, ExtensionFactoryDeployment> messageFactories;
    protected Map<String, ExtensionFactoryDeployment> schemaFactories;
    protected Map<String, Map<String, ExtensionFactoryDeployment>> connectorFactories;
    protected Map<String, ExtensionFactoryDeployment> idlFactories;
    protected Map<String, ExtensionFactoryDeployment> policyFactories;
    protected Map<String, ExtensionFactoryDeployment> managementContextFactories;
    protected Map<String, Map<String, ExtensionFactoryDeployment>> featureFactories;

    public ExtensionFactoryManager(SharedClassLoader sharedClassLoader) {
        this.sharedClassLoader = sharedClassLoader;
        providerReaders = new HashMap<String, ExtensionFactoryDeployment>();
        extensionFactories = new LinkedHashMap<String, ExtensionFactoryDeployment>();
        featureFactories = new HashMap<String, Map<String, ExtensionFactoryDeployment>>();
        applicationFactories = new HashMap<String, ExtensionFactoryDeployment>();
        containerFactories = new HashMap<String, ExtensionFactoryDeployment>();
        expressionFactories = new HashMap<String, ExtensionFactoryDeployment>();
        serverFactories = new HashMap<String, ExtensionFactoryDeployment>();
        engineFactories = new HashMap<String, ExtensionFactoryDeployment>();
        bindingFactories = new HashMap<String, ExtensionFactoryDeployment>();
        messageFactories = new HashMap<String, ExtensionFactoryDeployment>();
        schemaFactories = new HashMap<String, ExtensionFactoryDeployment>();
        connectorFactories = new HashMap<String, Map<String, ExtensionFactoryDeployment>>();
        idlFactories = new HashMap<String, ExtensionFactoryDeployment>();
        policyFactories = new HashMap<String, ExtensionFactoryDeployment>();
        managementContextFactories = new HashMap<String, ExtensionFactoryDeployment>();
    }

    public List<ExtensionFactory> getExtensionFactories() {
        List<ExtensionFactory> result = new ArrayList<ExtensionFactory>();
        for (ExtensionFactoryDeployment efd : extensionFactories.values()) {
            result.add(efd.getExtensionFactory());
        }
        return result;
    }

    public ProviderReader<?> getProviderReader(String providerNamespace) {
        ExtensionFactoryDeployment deployment = providerReaders.get(providerNamespace);
        if (deployment != null) {
            return (ProviderReader<?>)deployment.getExtensionFactory();
        } else {
            return null;
        }
    }

    public ExtensionFactory getExtensionFactory(String deploymentRef) {
        ExtensionFactoryDeployment deployment = extensionFactories.get(deploymentRef);
        if (deployment != null) {
            return deployment.getExtensionFactory();
        } else {
            return null;
        }
    }

    public ApplicationFactory getApplicationFactory(String extensionType) {
        ExtensionFactoryDeployment deployment = applicationFactories.get(extensionType);
        if (deployment != null) {
            return (ApplicationFactory)deployment.getExtensionFactory();
        } else {
            return null;
        }
    }

    public ExpressionFactory getExpressionFactory(String extensionType) {
        ExtensionFactoryDeployment deployment = expressionFactories.get(extensionType);
        if (deployment != null) {
            return (ExpressionFactory)deployment.getExtensionFactory();
        } else {
            return null;
        }
    }

    public MessageFactory getMessageFactory(String extensionType) {
        ExtensionFactoryDeployment deployment = messageFactories.get(extensionType);
        if (deployment != null) {
            return (MessageFactory)deployment.getExtensionFactory();
        } else {
            return null;
        }
    }

    public ContainerFactory getContainerFactory(String extensionType) {
        ExtensionFactoryDeployment deployment = containerFactories.get(extensionType);
        if (deployment != null) {
            return (ContainerFactory)deployment.getExtensionFactory();
        } else {
            return null;
        }
    }

    public ServerFactory getServerFactory(String extensionType) {
        ExtensionFactoryDeployment deployment = serverFactories.get(extensionType);
        if (deployment != null) {
            return (ServerFactory)deployment.getExtensionFactory();
        } else {
            return null;
        }
    }

    public EngineFactory getEngineFactory(String extensionType) {
        ExtensionFactoryDeployment deployment = engineFactories.get(extensionType);
        if (deployment != null) {
            return (EngineFactory)deployment.getExtensionFactory();
        } else {
            return null;
        }
    }

    public BindingFactory getBindingFactory(String extensionType) {
        ExtensionFactoryDeployment deployment = bindingFactories.get(extensionType);
        if (deployment != null) {
            return (BindingFactory)deployment.getExtensionFactory();
        } else {
            return null;
        }
    }
    
    

    public SchemaFactory getSchemaFactory(String extensionType) {
        ExtensionFactoryDeployment deployment = schemaFactories.get(extensionType);
        if (deployment != null) {
            return (SchemaFactory)deployment.getExtensionFactory();
        } else {
            return null;
        }
    }

    public ConnectorFactory getConnectorFactory(String extensionType, String impl) {
        Map<String, ExtensionFactoryDeployment> impls = connectorFactories.get(extensionType);
        if (impls != null) {
            ExtensionFactoryDeployment deployment = impls.get(impl);
            if (deployment != null) {
                return (ConnectorFactory)deployment.getExtensionFactory();
            } else {
                return null;
            }
        }
        return null;
    }

    public IDLFactory getIDLFactory(String extensionType) {
        ExtensionFactoryDeployment deployment = idlFactories.get(extensionType);
        if (deployment != null) {
            return (IDLFactory)deployment.getExtensionFactory();
        } else {
            return null;
        }
    }

    public PolicyFactory getPolicyFactory(String extensionType) {
        ExtensionFactoryDeployment deployment = policyFactories.get(extensionType);
        if (deployment != null) {
            return (PolicyFactory)deployment.getExtensionFactory();
        } else {
            return null;
        }
    }

    public Set<PolicyFactory> getPolicyFactories() {
        Set<PolicyFactory> result = new HashSet<PolicyFactory>();
        for (ExtensionFactoryDeployment deployment : policyFactories.values()) {
            result.add((PolicyFactory)deployment.getExtensionFactory());
        }
        return result;
    }

    public Set<SchemaFactory> getSchemaFactories() {
        Set<SchemaFactory> result = new HashSet<SchemaFactory>();
        for (ExtensionFactoryDeployment deployment : schemaFactories.values()) {
            result.add((SchemaFactory)deployment.getExtensionFactory());
        }
        return result;
    }
    
    public Set<BindingFactory> getBindingFactories() {
        Set<BindingFactory> result = new HashSet<BindingFactory>();
        for (ExtensionFactoryDeployment deployment : bindingFactories.values()) {
            result.add((BindingFactory)deployment.getExtensionFactory());
        }
        return result;
    }
    
    public Set<ManagementContextFactory> getManagementContextFactories() {
        Set<ManagementContextFactory> result = new HashSet<ManagementContextFactory>();
        for (ExtensionFactoryDeployment deployment : managementContextFactories.values()) {
            result.add((ManagementContextFactory)deployment.getExtensionFactory());
        }
        return result;
    }

    public FeatureFactory getFeatureFactory(String featureType, String impl) {

        Map<String, ExtensionFactoryDeployment> impls = featureFactories.get(featureType);
        if (impls != null) {

            ExtensionFactoryDeployment deployment = null;

            if (impl == null) {
                // return arbitrary impl as required by spec
                deployment = impls.values().iterator().next();
            } else {
                // return requested feature implementation
                deployment = impls.get(impl);
            }

            if (deployment != null) {
                return (FeatureFactory)deployment.getExtensionFactory();
            } else {
                return null;
            }

        }

        return null;
    }

    /**
     * Deploys an <code>ExtensionFactory</code> deployment. Note that if an attempt is made
     * to deploy a factory that has already been deployed, e.g. a different impl and\or
     * version, deployment is ignored and method returns quietly.
     * 
     * @param factoryContext
     *        which SHOULD have been created by invoking
     *        {@link #createDeploymentContext(Deployment)}, e.g. unless context was created
     *        within context of a testing framework.
     * @throws DeploymentException
     */
    public void deploy(ExtensionFactoryContext factoryContext) throws DeploymentException {

        if (factoryContext == null) {
            throw new IllegalArgumentException("factoryContext null");
        }

        ExtensionFactoryDeployment deployment = factoryContext.getDeployment();

        if (deployment == null) {
            throw new IllegalArgumentException("deployment null");
        }

        if (extensionFactories.containsKey(deployment.getRef())) {
            return;
        } else {
            extensionFactories.put(deployment.getRef(), deployment);
        }

        deployment.deploy(factoryContext);

        ExtensionFactory extensionFactory = deployment.getExtensionFactory();

        if (extensionFactory instanceof ProviderReader) {
            if (providerReaders.containsKey(extensionFactory.getExtensionType())) {
                throw new DeploymentException("Namespace '"
                        + extensionFactory.getExtensionType()
                        + "' claimed by ProviderReader defined within deployment "
                        + deployment.getRef()
                        + " is not unique. Namespace is already"
                        + " claimed by reader defined within deployment "
                        + providerReaders.get(extensionFactory.getExtensionType()).getRef()
                        + ".");
            } else {
                providerReaders.put(extensionFactory.getExtensionType(), deployment);
            }
        }

        if (extensionFactory.getType() == ExtensionFactory.Type.FEATURE) {
            FeatureFactory featureFactory;
            try {
                featureFactory = (FeatureFactory)extensionFactory;
            } catch (ClassCastException ce) {
                throw new DeploymentException("Expected instanceof "
                        + FeatureFactory.class.getName()
                        + ". Found "
                        + extensionFactory.getClass().getName());
            }
            String factoryType = featureFactory.getExtensionType();
            Map<String, ExtensionFactoryDeployment> featureMap = featureFactories.get(factoryType);
            if (featureMap == null) {
                featureMap = new HashMap<String, ExtensionFactoryDeployment>();
                featureFactories.put(factoryType, featureMap);
            }
            if (!featureMap.containsKey(featureFactory.getFeatureImpl())) {
                featureMap.put(featureFactory.getFeatureImpl(), deployment);
            } else {
                return;
            }
        }

        if (extensionFactory.getType() == ExtensionFactory.Type.APPLICATION) {
            ApplicationFactory applicationFactory;
            try {
                applicationFactory = (ApplicationFactory)extensionFactory;
            } catch (ClassCastException ce) {
                throw new DeploymentException("Expected instanceof "
                        + ApplicationFactory.class.getName()
                        + ". Found "
                        + extensionFactory.getClass().getName());
            }
            if (!applicationFactories.containsKey(applicationFactory.getExtensionType())) {
                applicationFactories.put(applicationFactory.getExtensionType(), deployment);
            } else {
                return;
            }
        }

        if (extensionFactory.getType() == ExtensionFactory.Type.EXPRESSION) {
            ExpressionFactory expressionFactory;
            try {
                expressionFactory = (ExpressionFactory)extensionFactory;
            } catch (ClassCastException ce) {
                throw new DeploymentException("Expected instanceof "
                        + ExpressionFactory.class.getName()
                        + ". Found "
                        + extensionFactory.getClass().getName());
            }
            if (!expressionFactories.containsKey(expressionFactory.getExtensionType())) {
                expressionFactories.put(expressionFactory.getExtensionType(), deployment);
            } else {
                return;
            }
        }

        if (extensionFactory.getType() == ExtensionFactory.Type.MESSAGE) {
            MessageFactory messageFactory;
            try {
                messageFactory = (MessageFactory)extensionFactory;
            } catch (ClassCastException ce) {
                throw new DeploymentException("Expected instanceof "
                        + MessageFactory.class.getName()
                        + ". Found "
                        + extensionFactory.getClass().getName());
            }
            if (!messageFactories.containsKey(messageFactory.getExtensionType())) {
                messageFactories.put(messageFactory.getExtensionType(), deployment);
            } else {
                return;
            }
        }

        if (extensionFactory.getType() == ExtensionFactory.Type.CONTAINER) {
            ContainerFactory containerFactory;
            try {
                containerFactory = (ContainerFactory)extensionFactory;
            } catch (ClassCastException ce) {
                throw new DeploymentException("Expected instanceof "
                        + ContainerFactory.class.getName()
                        + ". Found "
                        + extensionFactory.getClass().getName());
            }
            if (!containerFactories.containsKey(containerFactory.getExtensionType())) {
                containerFactories.put(containerFactory.getExtensionType(), deployment);
            } else {
                return;
            }
        }

        if (extensionFactory.getType() == ExtensionFactory.Type.SERVER) {
            ServerFactory serverFactory;
            try {
                serverFactory = (ServerFactory)extensionFactory;
            } catch (ClassCastException ce) {
                throw new DeploymentException("Expected instanceof "
                        + ServerFactory.class.getName()
                        + ". Found "
                        + extensionFactory.getClass().getName());
            }
            if (!serverFactories.containsKey(serverFactory.getExtensionType())) {
                serverFactories.put(serverFactory.getExtensionType(), deployment);
            } else {
                return;
            }
        }

        if (extensionFactory.getType() == ExtensionFactory.Type.ENGINE) {
            EngineFactory engineFactory;
            try {
                engineFactory = (EngineFactory)extensionFactory;
            } catch (ClassCastException ce) {
                throw new DeploymentException("Expected instanceof "
                        + EngineFactory.class.getName()
                        + ". Found "
                        + extensionFactory.getClass().getName());
            }
            if (!engineFactories.containsKey(engineFactory.getExtensionType())) {
                engineFactories.put(engineFactory.getExtensionType(), deployment);
            } else {
                return;
            }
        }

        if (extensionFactory.getType() == ExtensionFactory.Type.BINDING) {
            BindingFactory bindingFactory;
            try {
                bindingFactory = (BindingFactory)extensionFactory;
            } catch (ClassCastException ce) {
                throw new DeploymentException("Expected instanceof "
                        + BindingFactory.class.getName()
                        + ". Found "
                        + extensionFactory.getClass().getName());
            }
            if (!bindingFactories.containsKey(bindingFactory.getExtensionType())) {
                bindingFactories.put(bindingFactory.getExtensionType(), deployment);
            } else {
                return;
            }
        }

        if (extensionFactory.getType() == ExtensionFactory.Type.SCHEMA) {
            SchemaFactory schemaFactory;
            try {
                schemaFactory = (SchemaFactory)extensionFactory;
            } catch (ClassCastException ce) {
                throw new DeploymentException("Expected instanceof "
                        + SchemaFactory.class.getName()
                        + ". Found "
                        + extensionFactory.getClass().getName());
            }
            if (!schemaFactories.containsKey(schemaFactory.getExtensionType())) {
                schemaFactories.put(schemaFactory.getExtensionType(), deployment);
            } else {
                return;
            }
        }

        if (extensionFactory.getType() == ExtensionFactory.Type.CONNECTOR) {
            ConnectorFactory connectorFactory;
            try {
                connectorFactory = (ConnectorFactory)extensionFactory;
            } catch (ClassCastException ce) {
                throw new DeploymentException("Expected instanceof "
                        + ConnectorFactory.class.getName()
                        + ". Found "
                        + extensionFactory.getClass().getName());
            }
            String extensionType = connectorFactory.getExtensionType();
            String impl = connectorFactory.getConnectorImpl();
            Map<String, ExtensionFactoryDeployment> impls = connectorFactories.get(extensionType);
            if (impls == null) {
                impls = new HashMap<String, ExtensionFactoryDeployment>();
                connectorFactories.put(extensionType, impls);
            }
            if (!impls.containsKey(impl)) {
                impls.put(impl, deployment);
            } else {
                return;
            }
        }

        if (extensionFactory.getType() == ExtensionFactory.Type.IDL) {
            IDLFactory idlFactory;
            try {
                idlFactory = (IDLFactory)extensionFactory;
            } catch (ClassCastException ce) {
                throw new DeploymentException("Expected instanceof "
                        + IDLFactory.class.getName()
                        + ". Found "
                        + extensionFactory.getClass().getName());
            }
            if (!idlFactories.containsKey(idlFactory.getExtensionType())) {
                idlFactories.put(idlFactory.getExtensionType(), deployment);
            } else {
                return;
            }
        }

        if (extensionFactory.getType() == ExtensionFactory.Type.POLICY) {
            PolicyFactory policyFactory;
            try {
                policyFactory = (PolicyFactory)extensionFactory;
            } catch (ClassCastException ce) {
                throw new DeploymentException("Expected instanceof "
                        + PolicyFactory.class.getName()
                        + ". Found "
                        + extensionFactory.getClass().getName());
            }
            if (!policyFactories.containsKey(policyFactory.getExtensionType())) {
                policyFactories.put(policyFactory.getExtensionType(), deployment);
            } else {
                return;
            }
        }
        
        if (extensionFactory.getType() == ExtensionFactory.Type.MANAGEMENT) {
            ManagementContextFactory managementContextFactory;
            try {
                managementContextFactory = (ManagementContextFactory)extensionFactory;
            } catch (ClassCastException ce) {
                throw new DeploymentException("Expected instanceof "
                        + ManagementContextFactory.class.getName()
                        + ". Found "
                        + extensionFactory.getClass().getName());
            }
            if (!managementContextFactories.containsKey(managementContextFactory.getExtensionType())) {
                managementContextFactories.put(managementContextFactory.getExtensionType(), deployment);
            } else {
                return;
            }
        }

    }

    public void clear() {
        providerReaders.clear();
        extensionFactories.clear();
        applicationFactories.clear();
        expressionFactories.clear();
        messageFactories.clear();
        engineFactories.clear();
        bindingFactories.clear();
        schemaFactories.clear();
        connectorFactories.clear();
        idlFactories.clear();
        policyFactories.clear();
        managementContextFactories.clear();
        featureFactories.clear();
    }

}
