/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext;

import java.io.File;
import java.util.Set;

import org.bluestemsoftware.specification.eoa.ext.RootExtension.AbstractRootExtension;
import org.bluestemsoftware.specification.eoa.system.ManagementContext;

/**
 * Identifies a specific type of <code>Extension</code>, i.e. an <code>Extension</code> whose
 * 'life cycle' is managed by <code>System</code>.
 */
public abstract class ManageableExtension extends AbstractRootExtension {

    private File extensionEtcDir;
    private File extensionVarDir;

    public interface Provider extends RootExtension.Provider {
    }

    public ManageableExtension(ExtensionFactory factory, Provider provider, File extensionEtcDir,
            File extensionVarDir) {
        super(factory, provider);
        this.extensionEtcDir = extensionEtcDir;
        this.extensionVarDir = extensionVarDir;
    }
    
    /**
     * Returns directory which MUST be used by <code>ManageableExtension</code> to store
     * configuration, i.e. instance specific configuration, which is managed via it's
     * management API.
     * <p>
     * Note that directory MUST be separate and apart from directory used by other instances of
     * <code>System</code> which MAY be started within a separate jvm.
     * 
     * @return File directory
     */
    public File getExtensionEtcDir() {
        return extensionEtcDir;
    }

    /**
     * Returns directory which MUST be used by <code>ManageableExtension</code> to store
     * permanent files of a variable nature, e.g. database files, etc ...
     * <p>
     * Note that directory MUST be separate and apart from directory used by other instances of
     * <code>System</code> which MAY be started within a separate jvm.
     * 
     * @return File directory
     */
    public File getExtensionVarDir() {
        return extensionVarDir;
    }

    /**
     * Invoked by <code>System</code> before <code>Extension</code> is used. Allows
     * <code>Extension</code> to initialize resources which MAY be required to implement
     * extension behavior.
     * 
     * @param managementContexts
     *        A <code>Set</code> of objects which may be cast by
     *        <code>ManageableExtension</code> to a 'well-known' type to enable management via
     *        <code>System</code>.
     *        <p>
     *        Note that <code>Set</code> represents the 'available' means by which the
     *        <code>Extension</code> MAY expose its management API, i.e. the <code>Set</code>
     *        MAY be empty.
     * 
     * @throws ExtensionException
     *         if an error occurs during initialization
     */
    public abstract void init(Set<ManagementContext> managementContexts) throws ExtensionException;

    /**
     * Invoked by <code>System</code> when <code>Extension</code> is no longer required. Allows
     * <code>Extension</code> to destroy resources which MAY have been created within init
     * method.
     */
    public abstract void destroy();

}
