/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.feature.ws.appdata;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageModule;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.ActionContext.MessageModules;
import org.bluestemsoftware.specification.eoa.component.message.rt.ContentFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * Contains 'application data' transmitted 'out-of-band', i.e. outside of the normal data
 * channel, e.g. soap:Body.
 * <p>
 * Note that the <code>ApplicationDataModule</code> definition is built-in to this
 * specification and as such, an instance is provided by default within every instance of
 * {@link MessageModules}.
 */
public final class ApplicationDataModule extends MessageModule {

    public static final String TYPE = ApplicationDataModule.class.getName();

    public static final QName MODULE_ELEMENT = new QName(TYPE, "ApplicationData");

    private Map<QName, Set<ApplicationData>> applicationData;

    public ApplicationDataModule() {
    }

    public ApplicationDataModule(Map<QName, Set<ApplicationData>> applicationData) {
        this.applicationData = applicationData;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageModule#getNamespace()
     */
    @Override
    public String getNamespace() {
        throw new UnsupportedOperationException(); // unused
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageModule#getType()
     */
    @Override
    public String getType() {
        return TYPE;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.component.intrface.rt.MessageModule#toElement(org.w3c.dom.Document)
     */
    @Override
    public synchronized Element toElement(Document ownerDocument) {
        if (applicationData == null || applicationData.size() == 0) {
            return null;
        }
        Element wrapper = ownerDocument.createElementNS(TYPE, "eoa:ApplicationData");
        for (Set<ApplicationData> set : applicationData.values()) {
            for (Element element : set) {
                wrapper.appendChild(ownerDocument.importNode(element, true));
            }
        }
        return wrapper;
    }

    /**
     * Gets 'application data'. Note that returned <code>Map</code> is mutable, i.e. data MAY
     * be modified or removed by <code>Application</code> as required.
     * 
     * @param elementName
     *        qualifed name of data, i.e. elements, to return
     * @return Map or an empty map if no data
     */
    public synchronized Map<QName, Set<ApplicationData>> getApplicationData() {
        if (applicationData == null) {
            applicationData = new HashMap<QName, Set<ApplicationData>>();
        }
        return applicationData;
    }

    /**
     * Sets 'application data'. Note that existing data, if any, is overwritten.
     * 
     * @param elements
     * @see ContentFactory#createApplicationData(String, String)
     */
    public synchronized void setApplicationData(Map<QName, Set<ApplicationData>> applicationData) {
        if (applicationData == null) {
            throw new IllegalArgumentException("applicationData null");
        }
        this.applicationData = applicationData;
    }
    
    /**
     * Adds 'application data' to module.
     * @param applicationData
     */
    public synchronized void addApplicationData(ApplicationData applicationData) {
        if (applicationData == null) {
            throw new IllegalArgumentException("applicationData null");
        }
        Map<QName, Set<ApplicationData>> map = getApplicationData();
        QName name = new QName(applicationData.getNamespaceURI(), applicationData.getLocalName());
        Set<ApplicationData> set = map.get(name);
        if (set == null) {
            set = new HashSet<ApplicationData>();
            map.put(name, set);
        }
        set.add(applicationData);
    }
    
    /**
     * Returns size of 'application data' <code>Map</code>.
     * @return size
     */
    public int getSize() {
        if (applicationData == null) {
            return 0;
        } else {
            return applicationData.size();
        }
    }

    /**
     * Converts an <code>Element</code> serialized by this module to an instance of
     * <code>ApplicationDataModule</code>.
     * 
     * @param element
     *        element instance to convert
     * @return
     */
    public static ApplicationDataModule valueOf(Element element) {
        if (element == null || !element.getNamespaceURI().equals(TYPE)) {
            throw new IllegalArgumentException("Expected element of type " + new QName(TYPE, "module"));
        }
        Map<QName, Set<ApplicationData>> applicationData = null;
        NodeList nodeList = element.getChildNodes();
        for (int i = 0; i < nodeList.getLength(); i++) {
            if (nodeList.item(i).getNodeType() != Node.ELEMENT_NODE) {
                continue;
            }
            Element temp = (Element)nodeList.item(i);
            if (!(temp instanceof ApplicationData)) {
                throw new IllegalArgumentException("Wrapper element must contain instances of "
                        + ApplicationData.class.getName()
                        + " only.");
            }
            if (applicationData == null) {
                applicationData = new HashMap<QName, Set<ApplicationData>>();
            }
            QName elementName = new QName(temp.getNamespaceURI(), temp.getLocalName());
            Set<ApplicationData> set = applicationData.get(elementName);
            if (set == null) {
                set = new HashSet<ApplicationData>();
                applicationData.put(elementName, set);
            }
            set.add((ApplicationData)temp);
        }
        ApplicationDataModule result = new ApplicationDataModule();
        if (applicationData != null) {
            result.setApplicationData(applicationData);
        }
        return result;
    }

}
