/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.ext.policy;

import java.net.URI;
import java.net.URISyntaxException;

import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.component.ComponentContext;
import org.bluestemsoftware.specification.eoa.component.ComponentReader;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier.Scheme;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ElementPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.PolicyDocument;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactory;
import org.xml.sax.EntityResolver;

/**
 * A factory that produces a specific type of <code>PolicyExpression</code>, i.e. a specific
 * implementation of the WS-Policy specification, e.g. version 1.5, etc ...
 * <p>
 * Design Note: <code>PolicyFactory</code> is not a <code>ProviderReader</code>. Provider
 * definitions MUST be 'inlined' as 'top-level' elements within an IDL document and read via
 * the <code>ComponentReader</code> interface, i.e. such that 'deployed' expressions can acquire
 * the enclosing IDL document's targetNamespace.
 */

public abstract class PolicyFactory extends ExtensionFactory implements ComponentReader {

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends ExtensionFactory.Provider {
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ExtensionFactory#getType()
     */
    public final Type getType() {
        return Type.POLICY;
    }

    /**
     * Parses <code>Source</code> into a policy document.
     * 
     * @param entityResolver
     *        which MUST be used to resolve external entities, if any.
     * @param source
     *        or <code>null</code> in which case a compliant factory MUST return an empty
     *        <code>PolicyDocument</code>.
     *        <p>
     *        Note that if source is instanceof <code>SAXSource</code> which supplies an
     *        <code>XMLReader</code>, the reader MAY be ignored in preference to parser
     *        supplied by implementation.
     * @return
     * 
     * @throws DeploymentException
     */
    public abstract PolicyDocument readPolicy(EntityResolver entityResolver, Source source) throws PolicyException;

    /**
     * Reads attached element policy and returns as a <code>ElementPolicy</code> component.
     * 
     * @param componentContext
     *        Which MUST be used to resolve external entities which MAY be referenced by
     *        expression.
     * 
     * @param source
     *        which MUST contain an element information item instance, i.e. element to which
     *        policy is attached. To facilitate resolution of referenced expressions which MAY
     *        be located via a URI fragment, compliant implementations MUST populate systemId
     *        of <code>Source</code> with target namespace of parent IDL document.
     * 
     * @return ElementPolicy or <code>null</code> if no element policy found
     * 
     * @throws DeploymentException
     */
    public abstract ElementPolicy readElementPolicy(ComponentContext componentContext, DOMSource source) throws DeploymentException;

    /*
     * (non-Javadoc)
     */
    protected String normalizeLocation(String systemID, String location) throws PolicyException {

        if (location.startsWith("#")) {

            // decode fragment to remove any escaped octets before attempting to parse
            // into a fragmentid

            String fragment;
            try {
                fragment = new URI(location).getFragment();
            } catch (URISyntaxException ue) {
                throw new PolicyException("Error attempting to resolve referenced policy at URI '"
                        + location
                        + "'. "
                        + ue);
            }

            // location is a fragment, i.e. uri references a deployed policy expression

            FragmentIdentifier fragmentIdentifier = null;
            try {

                // attempt to parse fragment into a fragment id, i.e. if value fully
                // qualified. note that if parent expression is embedded within a
                // local engine definition, i.e. parent document is not wsdl, then
                // this is the only allowable form

                fragmentIdentifier = FragmentIdentifier.valueOf(fragment);

            } catch (IllegalArgumentException ie) {

                if (systemID == null) {
                    throw new PolicyException("Error attempting to resolve referenced policy at URI '"
                            + location
                            + "'. SystemID is null.");
                }

                // we assume fragment is just policy id and remainder of fragmentid is
                // implied using target namespace of parent document, i.e. wsdl doc

                fragmentIdentifier = new FragmentIdentifier(systemID, Scheme.EOA, "policy", fragment);

            }

            // create location with '#' prepended and escape any illegal chars within
            // framgent, e.g. the '#' character which occurs after document uri

            try {
                return new URI(null, null, fragmentIdentifier.toString()).toString();
            } catch (URISyntaxException ue) {
                throw new PolicyException("Error normalizing policy reference URI. " + ue.getMessage());
            }

        } else {

            // location is a url, i.e. policy is defined remotely. leave value
            // as is

            return location;

        }

    }

}
