/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */
package org.bluestemsoftware.specification.eoa.system;

import java.io.File;

import org.bluestemsoftware.specification.eoa.ext.RootExtension.AbstractRootExtension;
import org.bluestemsoftware.specification.eoa.ext.management.ManagementContextFactory;
import org.bluestemsoftware.specification.eoa.ext.management.ManagementException;

/**
 * Models an extensible set of management technologies which MAY be used to facilitate
 * <code>System</code> management. A concrete instance of <code>ManagementContext</code> is
 * cast by <code>ManageableExtensions</code> to a well-known type, e.g. Portal Context, JMX
 * Context, etc ...
 * 
 */
public abstract class ManagementContext extends AbstractRootExtension {

    private boolean initialized;
    private boolean finalized;
    private File extensionEtcDir;
    private File extensionVarDir;

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends AbstractRootExtension.Provider {

        public void spi_init() throws ManagementException;

        public void spi_finalize() throws ManagementException;
        
        public void spi_unFinalize() throws ManagementException;

        public void spi_destroy();

    }

    public ManagementContext(ManagementContextFactory factory, Provider provider, File extensionEtcDir,
            File extensionVarDir) {
        super(factory, provider);
        this.extensionEtcDir = extensionEtcDir;
        this.extensionVarDir = extensionVarDir;
    }

    /**
     * Invoked by <code>System</code> prior to passing context to instances of
     * <code>ManageableExtension.</code>
     * <p>
     * Note, because other <code>Extension</code> instances MAY rely upon this object in order
     * to initialize their state, concrete implementations of this class MUST NOT attempt to
     * access and use <code>Feature</code> instances, upon which this <code>Extension</code>
     * depends, if any, MUST be deferred until {@link #finalize().}
     * 
     * 
     * @throws ManagementException
     */
    public final void init() throws ManagementException {
        if (initialized) {
            return;
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_init();
        } finally {
            thread.setContextClassLoader(cl);
        }
        initialized = true;
    }

    /**
     * Invoked by <code>System</code> after all instances of <code>ManageableExtension</code>
     * have been initialized and after successful completion of {@link #init()}
     */
    public final void finalize() throws ManagementException {
        if (finalized) {
            return;
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_finalize();
        } finally {
            thread.setContextClassLoader(cl);
        }
        finalized = true;
    }

    /**
     * Life cycle method invoked by <code>System</code> prior to destroying instances of
     * <code>ManageableExtension.</code>
     * <p>
     * Allows <code>ManagementContext</code> instance to clean-up resources which MAY have been
     * created by <code>Feature</code> instances, upon which this <code>Extension</code>
     * depends, if any.
     * 
     */
    public final void unFinalize() throws ManagementException {
        if (finalized) {
            return;
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_unFinalize();
        } finally {
            thread.setContextClassLoader(cl);
        }
        finalized = true;
    }

    /**
     * Life cycle method invoked by <code>System</code> when context is no longer required
     * after {@link #unFinalize()} has successfully completed.
     */
    public final void destroy() {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_destroy();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Returns directory which MUST be used by <code>ManagementContext</code> to store
     * configuration, i.e. instance specific configuration, which is managed via it's
     * management API.
     * <p>
     * Note that directory MUST be separate and apart from directory used by other instances of
     * <code>System</code> which MAY be started within a separate jvm.
     * 
     * @return File directory
     */
    public File getExtensionEtcDir() {
        return extensionEtcDir;
    }

    /**
     * Returns directory which MUST be used by <code>ManagementContext</code> to store
     * permanent files of a variable nature, e.g. log files, etc ...
     * <p>
     * Note that directory MUST be separate and apart from directory used by other instances of
     * <code>System</code> which MAY be started within a separate jvm.
     * 
     * @return File directory
     */
    public File getExtensionVarDir() {
        return extensionVarDir;
    }

}
