/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.system;

import java.io.File;
import java.util.List;
import java.util.Set;

import javax.xml.namespace.QName;

import org.bluestemsoftware.specification.eoa.DeploymentException;
import org.bluestemsoftware.specification.eoa.SystemDependencies;
import org.bluestemsoftware.specification.eoa.component.ComponentDeployment;
import org.bluestemsoftware.specification.eoa.component.FragmentIdentifier;
import org.bluestemsoftware.specification.eoa.component.RootComponent;
import org.bluestemsoftware.specification.eoa.component.application.Application;
import org.bluestemsoftware.specification.eoa.component.binding.Binding;
import org.bluestemsoftware.specification.eoa.component.engine.Engine;
import org.bluestemsoftware.specification.eoa.component.intrface.Interface;
import org.bluestemsoftware.specification.eoa.component.message.InterfaceMessage;
import org.bluestemsoftware.specification.eoa.component.policy.Policy;
import org.bluestemsoftware.specification.eoa.component.rt.ProviderReader;
import org.bluestemsoftware.specification.eoa.component.service.Service;
import org.bluestemsoftware.specification.eoa.component.types.Schema;
import org.bluestemsoftware.specification.eoa.ext.Extension;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactory;
import org.bluestemsoftware.specification.eoa.ext.ExtensionFactoryDeployment;
import org.bluestemsoftware.specification.eoa.ext.ManageableExtension;
import org.bluestemsoftware.specification.eoa.ext.application.ApplicationFactory;
import org.bluestemsoftware.specification.eoa.ext.binding.BindingFactory;
import org.bluestemsoftware.specification.eoa.ext.connector.ConnectorFactory;
import org.bluestemsoftware.specification.eoa.ext.container.ContainerFactory;
import org.bluestemsoftware.specification.eoa.ext.engine.EngineFactory;
import org.bluestemsoftware.specification.eoa.ext.expression.ExpressionFactory;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureFactory;
import org.bluestemsoftware.specification.eoa.ext.idl.IDLFactory;
import org.bluestemsoftware.specification.eoa.ext.management.ManagementContextFactory;
import org.bluestemsoftware.specification.eoa.ext.message.MessageFactory;
import org.bluestemsoftware.specification.eoa.ext.policy.PolicyFactory;
import org.bluestemsoftware.specification.eoa.ext.schema.SchemaFactory;
import org.bluestemsoftware.specification.eoa.ext.server.ServerFactory;
import org.bluestemsoftware.specification.eoa.system.container.Container;
import org.bluestemsoftware.specification.eoa.system.server.Server;

/**
 * Models an EOA <code>System</code>. An EOA <code>System</code> is responsible for dependency
 * management, component/extension deployment, and the implementation of {@link Container} and
 * {@link Server} behavior.
 */
public interface System {

    /**
     * Gets ref URI which identifies <code>System</code> type, i.e. implementation.
     * 
     * @return
     */
    public String getRef();

    /**
     * Gets <code>System</code> 'managed' dependencies.
     * 
     * @return managed dependencies
     */
    public SystemDependencies getSystemDependencies();

    /**
     * Returns directory which SHOULD be utilized by {@link Extension} to store temporary
     * files. The returned directory MUST be separate and apart from directory used by other
     * instances of <code>System</code> which MAY be started within a separate jvm. Contents of
     * directory, if any, MUST be cleared by a compliant system following a re-start.
     * 
     * @return File
     */
    public abstract File getSystemTmpDir();

    /**
     * Returns base directory utilized by {@link ManageableExtension} objects to store
     * permanent files of a variable nature, e.g. database files, etc ...
     * <p>
     * Note that the returned directory MUST be separate and apart from directory used by other
     * instances of <code>System</code> which MAY be started within a separate jvm.
     * 
     * @return File
     */
    public abstract File getSystemVarDir();
    
    /**
     * Returns base directory utilized by {@link ManageableExtension} objects to store
     * instance specific, manageable configuration.
     * <p>
     * Note that the returned directory MUST be separate and apart from directory used by other
     * instances of <code>System</code> which MAY be started within a separate jvm.
     * 
     * @return File
     */
    public abstract File getSystemEtcDir();

    /**
     * Gets 'container' implemented by system. Provides connection and transaction support to
     * extensions.
     * <p>
     * Note that the means by which a specific type and implementation of container are chosen
     * is system specific and is outside the scope of this specification.
     * 
     * @return Container
     * @throws IllegalStateException
     *         if container is undefined, e.g. a 'lightweight' system with no additional
     *         behavior beyond that defined within the core specification.
     */
    public Container getContainer();

    /**
     * Gets 'server' implemented by system. Provides access to managed features and protocol
     * specific transport objects.
     * <p>
     * Note that the means by which a specific type and implementation of server are chosen is
     * system specific and is outside the scope of this specification.
     * 
     * @return Server
     * @throws IllegalStateException
     *         if server is undefined, e.g. a 'lightweight' system with no additional behavior
     *         beyond that defined within the core specification.
     */
    public Server getServer();

    /**
     * Returns <code>Set</code> of available management context objects.
     * 
     * @return <code>Set</code> of available management context objects or an empty
     *         <code>Set</code> if none defined.
     * 
     * @see ManageableExtension#init(Set)
     */
    public Set<ManagementContext> getManagementContexts();

    /**
     * Gets a <code>RootComponent</code> definition by fragment identifier.
     * 
     * @param fragmentID
     * @return RootComponent or <code>null</code> if undefined.
     * @throws DeploymentException
     *         if deployment within which component is defined is not yet deployed, and an
     *         error occurs deploying it.
     */
    public RootComponent getComponent(FragmentIdentifier fragmentID);

    /**
     * Gets an <code>Application</code> component by qualified name.
     * 
     * @param componentName
     * @return Component or <code>null</code> if undefined.
     * @throws DeploymentException
     *         if deployment within which component is defined is not yet deployed, and an
     *         error occurs deploying it.
     */
    public Application getApplication(QName componentName);

    /**
     * Gets a <code>Binding</code> component by qualified name.
     * 
     * @param componentName
     * @return Component or <code>null</code> if undefined.
     * @throws DeploymentException
     *         if deployment within which component is defined is not yet deployed, and an
     *         error occurs deploying it.
     */
    public Binding getBinding(QName componentName);

    /**
     * Gets an <code>Engine</code> component by qualified name.
     * 
     * @param componentName
     * @return Component or <code>null</code> if undefined.
     * @throws DeploymentException
     *         if deployment within which component is defined is not yet deployed, and an
     *         error occurs deploying it.
     */
    public Engine getEngine(QName componentName);

    /**
     * Gets an <code>Interface</code> component by qualified name.
     * 
     * @param componentName
     * @return Component or <code>null</code> if undefined.
     * @throws DeploymentException
     *         if deployment within which component is defined is not yet deployed, and an
     *         error occurs deploying it.
     */
    public Interface getInterface(QName componentName);

    /**
     * Gets a <code>InterfaceMessage</code> component by qualified name.
     * 
     * @param componentName
     * @return Component or <code>null</code> if undefined.
     * @throws DeploymentException
     *         if deployment within which component is defined is not yet deployed, and an
     *         error occurs deploying it.
     */
    public InterfaceMessage getMessage(QName componentName);

    /**
     * Gets a <code>Policy</code> component by qualified name.
     * 
     * @param componentName
     * @return Component or <code>null</code> if undefined.
     * @throws DeploymentException
     *         if deployment within which component is defined is not yet deployed, and an
     *         error occurs deploying it.
     */
    public Policy getPolicy(QName componentName);

    /**
     * Gets a <code>Service</code> component by qualified name.
     * 
     * @param componentName
     * @return Component or <code>null</code> if undefined.
     * @throws DeploymentException
     *         if deployment within which component is defined is not yet deployed, and an
     *         error occurs deploying it.
     */
    public Service getService(QName componentName);

    /**
     * Gets a <code>Schema</code> component by qualified name.
     * 
     * @param componentName
     * @return Component or <code>null</code> if undefined.
     * @throws DeploymentException
     *         if deployment within which component is defined is not yet deployed, and an
     *         error occurs deploying it.
     */
    public Schema getSchema(QName componentName);

    /**
     * Retrieves a <code>ComponentDeployment</code> by deployment ref.
     * 
     * @param deploymentRef
     * @return ComponentDeployment or <code>null</code> if undefined.
     */
    public ComponentDeployment getComponentDeployment(String deploymentRef);

    /**
     * Gets <code>ExtensionFactory</code> by deployment ref.
     * 
     * @see ExtensionFactoryDeployment#getRef()
     */
    public ExtensionFactory getExtensionFactory(String deploymentRef);

    /**
     * s Gets all <code>ExtensionFactory</code> definitions which MUST be listed in the order
     * in which they were deployed.
     * 
     * @return List of factory objects, or an empty List if none defined.
     */
    public List<ExtensionFactory> getExtensionFactories();

    /**
     * Gets <code>ProviderReader</code> which 'understands' indicated namespace and is capable
     * of reading runtime component definitions from element with indicated namespace.
     * 
     * @param providerNamespace
     */
    public ProviderReader<?> getProviderReader(String providerNamespace);

    /**
     * Gets factory that creates <code>Application</code> component extensions.
     * 
     * @param ref
     *        URI which identifies type of <code>Application</code> created by factory.
     * @return ApplicationFactory or <code>null</code> if undefined.
     */
    public ApplicationFactory getApplicationFactory(String ref);

    /**
     * Gets factory that creates runtime <code>Expression</code> extensions.
     * 
     * @param ref
     *        URI which identifies type of <code>Expression</code> created by factory or
     *        <code>null</code> to retrieve factory definition 'built-in' to this
     *        specification.
     * @return ExpressionFactory or <code>null</code> if undefined.
     */
    public ExpressionFactory getExpressionFactory(String ref);

    /**
     * Gets factory that creates <code>Message</code> extensions.
     * 
     * @param ref
     *        URI which identifies type of <code>Message</code> created by factory.
     * @return MessageFactory or <code>null</code> if undefined.
     */
    public MessageFactory getMessageFactory(String ref);

    /**
     * Returns an arbitrary message factory implementation used to generate 'default' message
     * instances.
     * 
     * @return
     */
    public MessageFactory getDefaultMessageFactory();

    /**
     * Gets factory that creates <code>Container</code> extensions.
     * 
     * @param ref
     *        URI which identifies type of <code>Container</code> created by factory.
     * @return ContainerFactory or <code>null</code> if undefined.
     */
    public ContainerFactory getContainerFactory(String ref);

    /**
     * Gets factory that creates <code>Server</code> extensions.
     * 
     * @param ref
     *        URI which identifies type of <code>Server</code> created by factory.
     * @return ServerFactory or <code>null</code> if undefined.
     */
    public ServerFactory getServerFactory(String ref);

    /**
     * Gets factory that creates <code>Engine</code> component extensions.
     * 
     * @param ref
     *        URI which identifies type of <code>Engine</code> created by factory.
     * @return EngineFactory or <code>null</code> if undefined.
     */
    public EngineFactory getEngineFactory(String ref);

    /**
     * Gets factory that creates <code>Binding</code> component extensions.
     * 
     * @param ref
     *        URI which identifies type of <code>Binding</code> created by factory. Note that
     *        this matches the value returned by {@link BindingFactory#getType()} and NOT the
     *        value returned by {@link BindingFactory#getBindingType()}.
     * @return BindingFactory or <code>null</code> if undefined.
     */
    public BindingFactory getBindingFactory(String ref);

    /**
     * Gets factories that create <code>Binding</code> component extensions.
     * 
     * @return Set of Binding factories defined on system or an empty set if none defined.
     */
    public Set<BindingFactory> getBindingFactories();

    /**
     * Gets factory that creates <code>Schema</code> component extensions.
     * 
     * @param ref
     *        URI which identifies type of <code>Schema</code> created by factory.
     * @return SchemaFactory or <code>null</code> if undefined
     */
    public SchemaFactory getSchemaFactory(String ref);

    /**
     * Gets factory that creates <code>Connector</code> component extensions.
     * 
     * @param ref
     *        URI which identifies type of <code>Connector</code> created by factory.
     * @param impl
     *        URI which uniquely identifies implementation. More than one implementation of a
     *        <code>ConnectorFactory</code> may exist, e.g. if <code>ConnectorFactory</code> is
     *        a factory for <code>javax.sql.DataSource</code> instances, an impl per supported
     *        DBMS vendor MAY be loaded.
     * @return ConnectorFactory or <code>null</code> if undefined.
     */
    public ConnectorFactory getConnectorFactory(String ref, String impl);

    /**
     * Gets factory that creates <code>IDLDocument</code> extensions.
     * 
     * @param ref
     *        URI which identifies type of <code>IDLDocument</code> created by factory.
     * @return IDLFactory or <code>null</code> if undefined.
     */
    public IDLFactory getIDLFactory(String ref);

    /**
     * Gets factory that creates <code>PolicyExpression</code> component extensions.
     * 
     * @param ref
     *        URI which identifies type of <code>PolicyExpression</code> created by factory.
     * 
     * @return PolicyFactory or <code>null</code> if undefined.
     */
    public PolicyFactory getPolicyFactory(String ref);

    /**
     * Gets factories that create <code>Policy</code> component extensions.
     * 
     * @return Set of Policy factories defined on system or an empty set if none defined.
     */
    public Set<PolicyFactory> getPolicyFactories();

    /**
     * Gets factories that create <code>Schema</code> component extensions.
     * 
     * @return Set of Schema factories defined on system or an empty set if none defined.
     */
    public Set<SchemaFactory> getSchemaFactories();

    /**
     * Gets factories that create <code>ManagementContext</code> extensions.
     * 
     * @return Set of ManagementContext factories defined on system or an empty set if none
     *         defined.
     */
    public Set<ManagementContextFactory> getManagementContextFactories();

    /**
     * Gets factory that creates <code>Feature</code> component extensions.
     * 
     * @param featureType
     *        Ref URI which identifies type of <code>Feature</code> created by factory.
     * @param impl
     *        URI which identifies a specific <code>Feature</code> implementation within
     *        context of indicated type or <code>null</code> to request an arbitrary
     *        implementation.
     * @return FeatureFactory or <code>null</code> if undefined.
     */
    public FeatureFactory getFeatureFactory(String featureType, String impl);

    /**
     * Gets <code>System</code> specific log implementation.
     * 
     * @param clazz
     *        Used to derive name which identifies <code>Log</code>.
     * 
     * @return log
     */
    public Log getLog(Class<?> clazz);

    /**
     * An interface which allows <code>System</code> to provide its own logging implementation
     * or optionally 'wrap' a third-party impl, the choice of which is <code>System</code>
     * specific and is outside the scope of this specification.
     */
    public interface Log {

        // here's an excerpt on why isWarnEnabled and isErrorEnabled methods
        // were removed from log4j api - from which this api was modeled

        // ... because you would usually want warning and fatal messages
        // being logged whenever they occur. They are usually encapsulated by a
        // conditional checking for null value or in a catch clause for when an
        // exception happens. These things shouldn't normally happen unless
        // there is some bad state within your program. This is much different
        // than debug and info messages which may be strewn throughout your code
        // just providing general information on the state of the program...good
        // or bad. Because debug and info messages might happen anywhere, it is
        // useful to have the conditional check to see if info or debug is
        // enabled. The info and debug messages still wouldn't be logged if
        // info and debug were not enabled, but everything within the log message
        // would be evaluated including string concatenation so the conditional
        // check provides an important performance increase for your application
        // when logging is not enabled and doesn't provide much of a hit when it
        // is enabled.

        public void fatal(String message);

        public void fatal(String message, Throwable t);

        public void debug(String message);

        public void debug(String message, Throwable t);

        public void trace(String message);

        public void trace(String message, Throwable cause);

        public void info(String message);

        public void info(String message, Throwable cause);

        public void error(String message);

        public void error(String message, Throwable cause);

        public void warn(String message);

        public void warn(String message, Throwable cause);

        public boolean isDebugEnabled();

        public boolean isInfoEnabled();

        public boolean isTraceEnabled();

    }

}
