/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.system;

import java.io.IOException;
import java.net.URL;
import java.util.Enumeration;
import java.util.Vector;

import org.bluestemsoftware.specification.eoa.AbstractClassLoader;
import org.bluestemsoftware.specification.eoa.ProvidedDependency;
import org.bluestemsoftware.specification.eoa.SharedClassLoader;

/**
 * Hides <code>System</code> implementation, i.e. loads classes required by
 * <code>System</code> to implement its behavior. Requests for classes/resources are first
 * delegated to parent classloader, i.e. classloader retrieved via
 * {@link #getParentClassLoader()}. If not found, repository scoped to system is checked. If
 * still not found, request is delegated to {@link SharedClassLoader}.
 * 
 * @see ProvidedDependency
 */
public class SystemClassLoader extends AbstractClassLoader {

    protected SharedClassLoader sharedClassLoader;

    public SystemClassLoader(URL[] urls) {
        super("SystemClassLoader", urls);
    }

    /*
     * (non-Javadoc)
     * 
     * Introduced for testing
     */
    public SystemClassLoader(ClassLoader parent, URL[] urls) {
        super("SystemClassLoader", urls, parent);
    }

    /**
     * Returns shared classloader.
     * 
     * @return shared classloader or <code>null</code> if not defined.
     */
    public SharedClassLoader getSharedClassLoader() {
        return sharedClassLoader;
    }

    /**
     * Used by system to set <code>SharedClassLoader</code> once system
     * dependencies have been modeled. Once set, the instance variable
     * is immutable.
     * 
     * @param sharedClassLoader
     */
    public void setSharedClassLoader(SharedClassLoader sharedClassLoader) {
        if (this.sharedClassLoader == null) {
            this.sharedClassLoader = sharedClassLoader;
        }
    }

    /*
     * (non-Javadoc)
     * @see java.lang.ClassLoader#loadClass(java.lang.String)
     */
    @Override
    public Class<?> loadClass(String name) throws ClassNotFoundException {
        return this.loadClass(name, false);
    }

    /*
     * (non-Javadoc)
     * @see java.lang.ClassLoader#loadClass(java.lang.String, boolean)
     */
    @Override
    public synchronized Class<?> loadClass(String name, boolean resolve) throws ClassNotFoundException {
        try {
            return super.loadClass(name, resolve);
        } catch (ClassNotFoundException cfe) {
            if (sharedClassLoader != null) {
                return sharedClassLoader.loadClass(name);
            }
            throw cfe;
        }
    }

    /*
     * (non-Javadoc)
     * @see java.lang.ClassLoader#getResource(java.lang.String)
     * 
     * NOTE: we synchronize to avoid race condition where two threads attempt to load same
     * class simultaneously (method on super is synchronized, but we need to lock method on
     * super AND on shared loader.) if class not loaded both threads will invoke findClass. it
     * is then a race in terms of which one will load and define class and which one will throw
     * a class linkage error:
     * 
     * "attempted duplicate class definition for name ..."
     */
    @Override
    public synchronized URL getResource(String name) {
        Enumeration<URL> answer = null;
        try {
            answer = getResources(name);
        } catch (IOException ie) {
            throw new RuntimeException(ie);
        }
        if (answer.hasMoreElements()) {
            return answer.nextElement();
        }
        return null;
    }

    /*
     * (non-Javadoc)
     * @see java.lang.ClassLoader#getResources(java.lang.String)
     * 
     * NOTE: we synchronize to avoid race condition where two threads attempt to load same
     * class simultaneously (method on super is synchronized, but we need to lock method on
     * super AND on shared loader.) if class not loaded both threads will invoke findClass. it
     * is then a race in terms of which one will load and define class and which one will throw
     * a class linkage error:
     * 
     * "attempted duplicate class definition for name ..."
     */
    @Override
    public synchronized Enumeration<URL> getResources(String name) throws IOException {

        Enumeration<URL> answer = super.getResources(name);
        if (!answer.hasMoreElements()) {
            if (sharedClassLoader != null) {
                answer = sharedClassLoader.getResources(name);
            }
        }

        if (!answer.hasMoreElements()) {
            return answer;
        }

        // fixes a bug within jar handler that returns a jar entry when searching
        // for meta-inf/services/foo and foo is a directory rather than a file. if
        // a dir entry, the returned stream is unusable. detect error by attempting
        // to close stream. as of jdk 1.5.0_08

        Vector<URL> validURLs = new Vector<URL>();
        while (answer.hasMoreElements()) {
            URL next = answer.nextElement();
            if (next.getProtocol().equals("jar")) {
                try {
                    next.openStream().close();
                    validURLs.add(next);
                } catch (Exception e) {
                }
            } else {
                validURLs.add(next);
            }
        }
        return validURLs.elements();

    }

    /*
     * (non-Javadoc)
     * @see java.net.URLClassLoader#getURLs()
     */
    @Override
    public URL[] getURLs() {
        if (sharedClassLoader != null) {
            URL[] a = sharedClassLoader.getURLs();
            if (a.length > 0) {
                URL[] b = super.getURLs();
                if (b.length > 0) {
                    URL[] combined = new URL[b.length + a.length];
                    java.lang.System.arraycopy(b, 0, combined, 0, b.length);
                    java.lang.System.arraycopy(a, 0, combined, b.length, a.length);
                    return combined;
                } else {
                    return a;
                }
            } else {
                return super.getURLs();
            }
        } else {
            return super.getURLs();
        }
    }

    @Override
    public String toString() {
        return id + "@" + Integer.toHexString(hashCode());
    }

}
