/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.system.server;

import java.io.File;
import java.util.Set;

import org.bluestemsoftware.specification.eoa.ext.ManageableExtension;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureException;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureFactory;
import org.bluestemsoftware.specification.eoa.system.ManagementContext;
import org.w3c.dom.Element;

/**
 * Encapsulates a specific type of abstract functionality as identified by
 * {@link #getExtensionType()}. Type specific methods are invoked by casting
 * <code>Feature</code> to a well-known implementation.
 * 
 * @see Server#getFeature(Class)
 */

public abstract class Feature extends ManageableExtension {

    protected String impl;

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends ManageableExtension.Provider {

        public String spi_getFeatureImpl();

        public void spi_init(Set<ManagementContext> managementContexts) throws FeatureException;

        public void spi_destroy();

    }

    public Feature(FeatureFactory factory, Provider provider, File extensionEtcDir, File extensionVarDir) {
        super(factory, provider, extensionEtcDir, extensionVarDir);
    }

    /**
     * Returns URI which uniquely identifies <code>Feature</code> implementation within
     * context of <code>Feature</code> type.
     * 
     * @return implementation URI
     */
    public String getFeatureImpl() {
        if (impl == null) {
            Thread thread = Thread.currentThread();
            ClassLoader cl = thread.getContextClassLoader();
            try {
                thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
                impl = ((Provider)provider).spi_getFeatureImpl();
            } finally {
                thread.setContextClassLoader(cl);
            }
        }
        return impl;
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ManageableExtension#init(java.util.Set)
     */
    public void init(Set<ManagementContext> managementContexts) throws FeatureException {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_init(managementContexts);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.ext.ManageableExtension#destroy()
     */
    public void destroy() {
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_destroy();
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    // *********************** Identity Methods ************************************

    @Override
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb.append("type: ");
        sb.append(getExtensionType());
        sb.append(System.getProperty("line.separator"));
        sb.append("impl: ");
        sb.append(getFeatureImpl());
        return sb.toString();
    }

    @Override
    public int hashCode() {
        return hashCode(getExtensionType()) + hashCode(getFeatureImpl());
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == this) {
            return true;
        }

        if (!(obj instanceof Feature)) {
            return false;
        }

        if ((getExtensionType() == ((Feature)obj).getExtensionType() || (getExtensionType() != null && getExtensionType()
                .equals(((Feature)obj).getExtensionType())))
                && (getFeatureImpl() == ((Feature)obj).getFeatureImpl() || (getFeatureImpl() != null && getFeatureImpl()
                        .equals(((Feature)obj).getFeatureImpl())))) {
            return true;
        }

        return false;
    }

    private int hashCode(Object object) {
        return (object == null ? 0 : object.hashCode());
    }

}