/**
 * Copyright 2008 Bluestem Software LLC
 *   
 * All Rights Reserved. Bluestem Software LLC licenses this file
 * to you under the EOA Specification License Agreement [1]. You
 * may not use this file except in compliance with the License. You
 * may obtain a copy of the License at:
 *   
 * [1] http://alakai.org/licensing/specification-eoa-license.html
 *   
 * Unless required by applicable law or agreed to in writing,
 * software distributed under the License is distributed on an
 * "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
 * KIND, either express or implied. See the License for the
 * specific language governing permissions and limitations
 * under the License.
 */

package org.bluestemsoftware.specification.eoa.system.server;

import java.io.File;
import java.util.Set;

import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointActionReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.EndpointReference;
import org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule;
import org.bluestemsoftware.specification.eoa.component.engine.rt.TransportModule;
import org.bluestemsoftware.specification.eoa.component.intrface.rt.SystemFault;
import org.bluestemsoftware.specification.eoa.component.policy.rt.ActionPolicy;
import org.bluestemsoftware.specification.eoa.component.policy.rt.UnsupportedPolicyException;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureException;
import org.bluestemsoftware.specification.eoa.ext.feature.FeatureFactory;
import org.bluestemsoftware.specification.eoa.ext.feature.ws.transport.rt.TransportProtocolRT;

/**
 * Identifies a specific type of <code>Feature</code>, i.e. a feature which transports
 * messages into/out of 'binding layer'.
 */
public abstract class TransportFeature extends Feature implements ModulableFeature {

    protected Set<String> policyVocabularyNamespaces;

    /*
     * (non-Javadoc)
     * 
     * SPI
     */
    public interface Provider extends Feature.Provider {

        public TransportModule spi_createTransportModule(EndpointActionReference endpointActionReference, ActionPolicy privatePolicy, ActionPolicy publicPolicy) throws FeatureException, UnsupportedPolicyException;

        public TransportModule spi_createDefaultTransportModule(EndpointReference endpointReference) throws FeatureException;

        public Set<String> spi_getPolicyVocabularyNamespaces();
        
        public void spi_destroyFeatureModule(FeatureModule featureModule);

    }

    public TransportFeature(FeatureFactory factory, Provider provider, File extensionEtcDir, File extensionVarDir) {
        super(factory, provider, extensionEtcDir, extensionVarDir);
    }

    /**
     * Identifies transport 'type' employed by feature to transport messages into/out of
     * binding layer.
     * 
     * @return transport type
     */
    public abstract Class<? extends TransportProtocolRT> getTransportProtocol();

    /**
     * Processes domain specific policy which enables/constrains <code>Feature</code> and
     * creates an instance of <code>TransportModule</code> which implements
     * <code>Feature</code> on a per message basis by processing message instances associated
     * with action identified by {@link EndpointActionReference#getAction()}.
     * 
     * @param endpointActionReference
     *        targeted component upon which module will be installed. References action to
     *        which 'private' and 'public' policy applies.
     * 
     * @param privatePolicy
     *        Aggregates 'private' effective policy expressions from all four policy subjects
     *        which MAY affect processing of messages associated with referenced action. The
     *        contained expressions have been filtered such that they contain only assertions
     *        understood by <code>ModulableFeature</code>, i.e. assertions which match a
     *        namespace returned by {@link #getPolicyVocabularyNamespaces()}.
     *        <p>
     *        Note that 'private' policy MAY be used to negotiate alternatives defined on
     *        referenced action, i.e. referenced action may advertise that it supports more
     *        than one capability, in which case 'private' policy would be intersected with
     *        policy defined on referenced action to 'negotiate' a preferred alternative.
     * 
     * @param publicPolicy
     *        Aggregates 'public' effective policy expressions from all four policy subjects
     *        which MAY affect processing of messages associated with referenced action. The
     *        contained expressions have been filtered such that they contain only assertions
     *        understood by <code>ModulableFeature</code>, i.e. assertions which match a
     *        namespace returned by {@link #getPolicyVocabularyNamespaces()}.
     * 
     * @return TransportModule instance
     * @throws FeatureException
     *         if an error occurs creating module(s).
     * @throws UnsupportedPolicyException
     *         Which MUST be thrown by Feature if it is unable to comply with policy.
     */
    public TransportModule createTransportModule(EndpointActionReference endpointActionReference, ActionPolicy privatePolicy, ActionPolicy publicPolicy) throws FeatureException, UnsupportedPolicyException {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_createTransportModule(endpointActionReference, privatePolicy,
                    publicPolicy);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /**
     * Returns <code>TransportModule</code> configured using a default configuration which is
     * used by <code>Binding</code> to process undeclared action {@link SystemFault#ACTION},
     * i.e. where implied message direction is 'Out' and which will provide transport behavior
     * for referenced <code>Endpoint</code> in the absence of a policy enabled/configured
     * module.
     * 
     * @param endpointReference
     *        upon which module will be installed
     * @return default module
     * @throws FeatureException
     *         if an error occurs creating module
     * 
     * @see EndpointReference#getDefaultTransportModule()
     */
    public TransportModule createDefaultTransportModule(EndpointReference endpointReference) throws FeatureException {
        if (provider == null) {
            throw new IllegalStateException("provider not set");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            return ((Provider)provider).spi_createDefaultTransportModule(endpointReference);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.system.server.ModulableFeature#getPolicyVocabularyNamespaces()
     */
    public Set<String> getPolicyVocabularyNamespaces() {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        if (policyVocabularyNamespaces == null) {
            Thread thread = Thread.currentThread();
            ClassLoader cl = thread.getContextClassLoader();
            try {
                thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
                policyVocabularyNamespaces = ((Provider)provider).spi_getPolicyVocabularyNamespaces();
            } finally {
                thread.setContextClassLoader(cl);
            }
        }
        return policyVocabularyNamespaces;
    }
    
    /*
     * (non-Javadoc)
     * @see org.bluestemsoftware.specification.eoa.system.server.ModulableFeature#destroyFeatureModule(org.bluestemsoftware.specification.eoa.component.engine.rt.FeatureModule)
     */
    public void destroyFeatureModule(FeatureModule featureModule) {
        if (provider == null) {
            throw new IllegalStateException("provider not set.");
        }
        Thread thread = Thread.currentThread();
        ClassLoader cl = thread.getContextClassLoader();
        try {
            thread.setContextClassLoader(factory.getFactoryContext().getClassLoader());
            ((Provider)provider).spi_destroyFeatureModule(featureModule);
        } finally {
            thread.setContextClassLoader(cl);
        }
    }

}
